/*
** Module   : Keyboard.java
** Abstract : A collection of keyboard event mapping methods.
**
** Copyright (c) 2005-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051201   @23590 Created initial version.
** 002 NVS 20060120   @24003 Added mapping of HELP to F2.
** 003 SIY 20060209   @24427 Safe handling of null input.
** 004 NVS 20060222   @24666 Fixed a bug in the remapKey() method.
** 005 NVS 20060227   @24785 Defined key actions and keyAction() method
**                           that converts a key code into a key action if
**                           possible.
** 006 NVS 20060302   @24860 Added KA_LEFT_END and KA_RIGHT_END constants
**                           and keyCodeForAction() method.
** 007 NVS 20060315   @25078 KA_TAB was not encoded properly.
** 008 NVS 20060322   @25182 Added new method resetKeyboard() and ported
**                           the locate() method to the new context local
**                           framework.
** 009 GES 20060323   @25209 Code cleanup and addition of a simple test
**                           harness.
** 010 NVS 20060323   @25225 Changed the default key function mappings to 
**                           match those in Progress.
** 011 SIY 20060327   @25280 Added codes for the DEFAULT_ACTION,  
**                           VALUE_CHANGED actions and others.
** 012 NVS 20060616   @27429 Added definitions for CHOOSE synthetic event.
** 013 NVS 20060630   @27743 High level (synthetic) event SE_ENTRY was
**                           improperly spelled as SE_ENTER. Fixed.
** 014 GES 20060704   @27785 remapKey() was modified to add a new key
**                           function if the given key function is one of
**                           the well known function names and it does not
**                           have a default mapping (only some of the
**                           key functions have default mappings).
**                           Otherwise functions (like STOP) could never
**                           be remapped.
** 015 GES 20060714   @28064 Added missing ANY-KEY, ROW-DISPLAY and
**                           WINDOW-CLOSE events.  Added ENTER as a
**                           synonym for RETURN.
** 016 GES 20060714   @28075 Honor alternate key labels in keyCode().
** 017 NVS 20060816   @28622 Added key function definition for ESC key
**                           (integer code 27) as "\u001B".
** 018 SIY 20060816   @28646 Added definitions missing key action codes
**                           used by Editor widget.  
** 019 NVS 20060906   @29234 Page Up and Page Down keys are assigned key
**                           functions PAGE-UO and PAGE-DOWN by default.
**                           These definitions are in addition to the
**                           existing primary definitions of PAGE-UP and
**                           PAGE-DOWN as ESC+up and ESC+down respectively
** 020 NVS 20070328   @32637 Refactored the code to make it fully
**                           compliant with the ContextLocal guidelines.
** 021 NVS 20070507   @33422 Added event names by code printout from the
**                           command line using -e option.
** 022 EVL 20071217   @36400 Adding key mapping for BACK-TAB and other
**                           missing keys.
** 023 GES 20080314   @37450 Safer I18N-aware initialization of the
**                           "extended ASCII" region to allow other
**                           character sets to be supported other than
**                           just ISO-8859-1. Added isPrintable() method.
** 024 GES 20080414   @38010 Moved I18N logic into external helpers that
**                           can be reused.
** 025 CA  20081006   @40037 Fix for the KBLABEL function, when 1-char
**                           function parameter is received - it must not
**                           be returned in uppercase.
** 026 CA  20081202   @40817 Added ROW-DISPLAY event to the event list.
** 027 GES 20090422   @41900 Converted to standard string formatting.
** 028 CA  20090422   @41837 Added event ids for NEXT-FRAME and PREV-FRAME.
** 029 SIY 20100619          Switching to own implementation of character UI.
** 030 SIY 20100804          Changed imports.
** 031 SIY 20101114          Changes related to switching from KeyEvent to 
**                           KeyInput. Some minor cleanups.
** 032 SIY 20111002          Updated javadoc.
** 033 GES 20111005          Updated import to match refactoring.
** 034 SIY 20111101          Updated imports.
** 035 EVL 20130725          Removed CHARVA package, using javadoc KeyProcessor from console
**                           driver.
** 036 CA  20130822          Added support of appserver async requests.  Added support for events
**                           generated on server-side: PROCEDURE-COMPLETE.
** 037 CA  20130908          Added constants related to socket events (CONNECT and READ-RESPONSE).
*/

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.client.chui.driver.console.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.util.*;

/**
 * This class represents the keyboard event codes, key labels and Progress
 * functions and all their interrelations. Beside that this class contains
 * default mapping between Progress 4GL key codes and names assigned to these
 * code.
 * <p><h4>Key Codes and Key Names</h4><br>
 * All key codes recognized by Progress 4GL (and P2J) are in range 0-4095. 
 * This range is divided into several subranges and decoding it follows one 
 * of the rules listed below, depending on the range to which code belongs<br>
 * <ul>
 * <li>Ranges [0-511], [520-521], [525], [539], [656-1535], [1544-1545], [1549], 
 * [1563], [1663-2559], [2568-2569], [2573], [2587], [2687-3583], [3592-3593], 
 * [3597], [3611] and [3711-4095] are handled using <code>Rule 1</code>.</li>
 * <li>Ranges [1614-1662], [2638-2686] and [3662-3710] are always recognized as
 * keys with no name and are ignored (see {@link KeyProcessor#map(int)}).</li>
 * <li>Codes which are not fall into ranges listed above follow <code>Rule 2</code>.
 * </ul>
 * Both rules decode key into base key name first and then add optional prefix
 * consisting of "SHIFT-", "CTRL-", "ESC-" or combination of these strings.<br> 
 * Decoding rules are following:
 * <ul>
 *   <li>Rule 1:<br>
 *   Take name of the key by retrieving from the name table (see below) using 
 *   key code modulo 512 as index. Then to basic name can be added prefix 
 *   depending on the key code:<br>
 *   <ul>
 *     <li>Codes greater or equal 3584 get prefix <code>CTRL-SHIFT-ESC-</code> </li>
 *     <li>Codes between 3584 and 3072 get prefix <code>CTRL-ESC-</code> </li>
 *     <li>Codes between 3072 and 2560 get prefix <code>CTRL-SHIFT-</code> </li>
 *     <li>Codes between 2560 and 2048 get prefix <code>CTRL-</code> </li>
 *     <li>Codes between 2048 and 1536 get prefix <code>SHIFT-ESC-</code> </li>
 *     <li>Codes between 1536 and 1024 get prefix <code>ESC-</code> </li>
 *     <li>Codes between 1024 and 512 get prefix <code>SHIFT-</code> </li>
 *     <li>Codes below 512 get no additional prefix.</li>
 *   </ul>
 *   </li>
 *   <li>Rule 2:<br>
 *       Take name of the key by retrieving its name from the table using key 
 *       code modulo 512 as index (for keys with codes modulo 512 greater than 143) 
 *       or key code modulo 512 + 512 (for key with codes modulo 512 less or equal
 *       to 143). Then to basic name can be added prefix depending on the key 
 *       code:<br>
 *   <ul>
 *     <li>Codes greater or equal 3584 get prefix <code>CTRL-ESC-</code> </li>
 *     <li>Codes between 3584 and 2560 get prefix <code>CTRL-</code> </li>
 *     <li>Codes between 1536 and 2560 get prefix <code>ESC-</code> </li>
 *     <li>Codes below 1536 get no additional prefix.</li>
 *   </ul>
 *   </li>
 * </ul>
 * <br>
 * The name table for keys consist of 656 entries. This range can be divided
 * into several subranges for convenience.<br>
 * <ul>
 *    <li>Range 0-127 represents ASCII table, where 0-31 and 127 are control 
 *    codes and remaining part consists of ASCII printable characters - 
 *    punctuation,  *    upper- and lower-case letters, digits, etc.<br>
 *    Control codes in Progress have following names:
 *    <table>
 *       <tr><th>Key Code</th><th>Key Name</th></tr>
 *       <tr><td> 0</td><td>   CTRL-@</td></tr>
 *       <tr><td> 1</td><td>   CTRL-A</td></tr>
 *       <tr><td> 2</td><td>   CTRL-B</td></tr>
 *       <tr><td> 3</td><td>   CTRL-C</td></tr>
 *       <tr><td> 4</td><td>   CTRL-D</td></tr>
 *       <tr><td> 5</td><td>   CTRL-E</td></tr>
 *       <tr><td> 6</td><td>   CTRL-F</td></tr>
 *       <tr><td> 7</td><td>   CTRL-G</td></tr>
 *       <tr><td> 8</td><td>BACKSPACE</td></tr>
 *       <tr><td> 9</td><td>      TAB</td></tr>
 *       <tr><td>10</td><td>   CTRL-J</td></tr>
 *       <tr><td>11</td><td>   CTRL-K</td></tr>
 *       <tr><td>12</td><td>   CTRL-L</td></tr>
 *       <tr><td>13</td><td>   RETURN</td></tr>
 *       <tr><td>14</td><td>   CTRL-N</td></tr>
 *       <tr><td>15</td><td>   CTRL-O</td></tr>
 *       <tr><td>16</td><td>   CTRL-P</td></tr>
 *       <tr><td>17</td><td>   CTRL-Q</td></tr>
 *       <tr><td>18</td><td>   CTRL-R</td></tr>
 *       <tr><td>19</td><td>   CTRL-S</td></tr>
 *       <tr><td>20</td><td>   CTRL-T</td></tr>
 *       <tr><td>21</td><td>   CTRL-U</td></tr>
 *       <tr><td>22</td><td>   CTRL-V</td></tr>
 *       <tr><td>23</td><td>   CTRL-W</td></tr>
 *       <tr><td>24</td><td>   CTRL-X</td></tr>
 *       <tr><td>25</td><td>   CTRL-Y</td></tr>
 *       <tr><td>26</td><td>   CTRL-Z</td></tr>
 *       <tr><td>27</td><td>      ESC</td></tr>
 *       <tr><td>28</td><td>   CTRL-\</td></tr>
 *       <tr><td>29</td><td>   CTRL-]</td></tr>
 *       <tr><td>30</td><td>   CTRL-^</td></tr>
 *       <tr><td>31</td><td>   CTRL-_</td></tr>
 *       <tr><td>127</td><td>     DEL</td></tr>
 *    </table>
 *    </li>
 *    <li>Range 128-255 holds locale-specific characters.</li>
 *    <li>Range 256-299 is not used.</li>
 *    <li>Range 300-499 contains names of function keys - F0-F99 followed by
 *    PF0-PF99.</li>
 *    <li>Code 500 is not used.</li>
 *    <li>Range 501-655 contains cursor movement keys and mouse events.
 *    Not all positions in this range are filled. Filled positions are listed 
 *    below:<br>
 *    <table>
 *       <tr><th>Key Code</th><th>Key Name</th></tr>
 *       <tr><td>501</td><td>            CURSOR-UP</td></tr>
 *       <tr><td>502</td><td>          CURSOR-DOWN</td></tr>
 *       <tr><td>503</td><td>         CURSOR-RIGHT</td></tr>
 *       <tr><td>504</td><td>          CURSOR-LEFT</td></tr>
 *       <tr><td>505</td><td>                 HOME</td></tr>
 *       <tr><td>506</td><td>                  END</td></tr>
 *       <tr><td>507</td><td>              PAGE-UP</td></tr>
 *       <tr><td>508</td><td>            PAGE-DOWN</td></tr>
 *       <tr><td>509</td><td>             BACK-TAB</td></tr>
 *       <tr><td>510</td><td>                  INS</td></tr>
 *       <tr><td>511</td><td>             HELP-KEY</td></tr>
 *       <tr><td>512</td><td>             DEL-CHAR</td></tr>
 *       <tr><td>513</td><td>              EXECUTE</td></tr>
 *       <tr><td>514</td><td>                 PAGE</td></tr>
 *       <tr><td>515</td><td>                 FIND</td></tr>
 *       <tr><td>516</td><td>             INS-LINE</td></tr>
 *       <tr><td>517</td><td>             DEL-LINE</td></tr>
 *       <tr><td>518</td><td>           LINE-ERASE</td></tr>
 *       <tr><td>519</td><td>           PAGE-ERASE</td></tr>
 *       <tr><td>520</td><td>      SHIFT-BACKSPACE</td></tr>
 *       <tr><td>521</td><td>            SHIFT-TAB</td></tr>
 *       <tr><td>522</td><td>                 EXIT</td></tr>
 *       <tr><td>525</td><td>         SHIFT-RETURN</td></tr>
 *       <tr><td>535</td><td>                ERASE</td></tr>
 *       <tr><td>536</td><td>                WHITE</td></tr>
 *       <tr><td>537</td><td>                 BLUE</td></tr>
 *       <tr><td>538</td><td>                  RED</td></tr>
 *       <tr><td>539</td><td>            SHIFT-ESC</td></tr>
 *       <tr><td>541</td><td>           CTRL-BREAK</td></tr>
 *       <tr><td>551</td><td>                   U1</td></tr>
 *       <tr><td>552</td><td>                   U2</td></tr>
 *       <tr><td>553</td><td>                   U3</td></tr>
 *       <tr><td>554</td><td>                   U4</td></tr>
 *       <tr><td>555</td><td>                   U5</td></tr>
 *       <tr><td>556</td><td>                   U6</td></tr>
 *       <tr><td>557</td><td>                   U7</td></tr>
 *       <tr><td>558</td><td>                   U8</td></tr>
 *       <tr><td>559</td><td>                   U9</td></tr>
 *       <tr><td>560</td><td>                  U10</td></tr>
 *       <tr><td>586</td><td>                   DO</td></tr>
 *       <tr><td>587</td><td>               SELECT</td></tr>
 *       <tr><td>588</td><td>               REMOVE</td></tr>
 *       <tr><td>589</td><td>          INSERT-HERE</td></tr>
 *       <tr><td>609</td><td>      MOUSE-SELECT-UP</td></tr>
 *       <tr><td>610</td><td>        MOUSE-MOVE-UP</td></tr>
 *       <tr><td>611</td><td>        MOUSE-MENU-UP</td></tr>
 *       <tr><td>612</td><td>      MOUSE-EXTEND-UP</td></tr>
 *       <tr><td>617</td><td>    MOUSE-SELECT-DOWN</td></tr>
 *       <tr><td>618</td><td>      MOUSE-MOVE-DOWN</td></tr>
 *       <tr><td>619</td><td>      MOUSE-MENU-DOWN</td></tr>
 *       <tr><td>620</td><td>    MOUSE-EXTEND-DOWN</td></tr>
 *       <tr><td>625</td><td>   MOUSE-SELECT-CLICK</td></tr>
 *       <tr><td>626</td><td>     MOUSE-MOVE-CLICK</td></tr>
 *       <tr><td>627</td><td>     MOUSE-MENU-CLICK</td></tr>
 *       <tr><td>628</td><td>   MOUSE-EXTEND-CLICK</td></tr>
 *       <tr><td>633</td><td>    MOUSE-SELECT-DRAG</td></tr>
 *       <tr><td>634</td><td>      MOUSE-MOVE-DRAG</td></tr>
 *       <tr><td>635</td><td>      MOUSE-MENU-DRAG</td></tr>
 *       <tr><td>636</td><td>    MOUSE-EXTEND-DRAG</td></tr>
 *       <tr><td>640</td><td>           MOUSE-MOVE</td></tr>
 *       <tr><td>649</td><td>MOUSE-SELECT-DBLCLICK</td></tr>
 *       <tr><td>650</td><td>  MOUSE-MOVE-DBLCLICK</td></tr>
 *       <tr><td>651</td><td>  MOUSE-MENU-DBLCLICK</td></tr>
 *       <tr><td>652</td><td>MOUSE-EXTEND-DBLCLICK</td></tr>
 *       </table><br>
 *    Note that code 606 is special - the key with this code has no defined
 *    name (key label), but has assigned key function 
 *    (see below more details about mapping).
 *    </li>
 * </ul>
 * 
 * <p><h4>Mapping Between Key Function and Key Label</h4><br>
 * The mapping between key functions and key labels are 
 * maintained in two maps, f2l and l2f.
 * <p>
 * First of all, the f2l map serves as a table of all valid key function
 * names.
 * <p>
 * f2l map has the key function (String) as a key. Values are sets 
 * (LinkedHashSet) of key labels which have been assigned this key function.  
 * Therefore, a single key function can be mapped to multiple key labels.
 * <p>
 * l2f map has the key label (String) as a key. Values are key functions
 * (String) assigned to those key labels. A single key label maps into a 
 * single key function, although more than one key label can be mapped into 
 * the same key function.
 * <p>
 * When a key label has to be remapped into a different key function, both 
 * maps are modified. The old mapping of this key label is queried from the
 * l2f map. If there was some mapping found in the l2f map, the following
 * cleanup is done:
 * <ul>
 *   <li>the f2l map is queried and the key label is deleted from the
 *       resulting set (the empty set is deleted, but the mapping itself);
 *   <li>the l2f mapping is deleted;
 * </ul>
 * Next, the new mapping is created as follows:
 * <ul>
 *   <li>a new mapping is added to l2f pointing to the new key function;
 *   <li>a set is retrieved from the f2l map for the new key function and then
 *       this key label is added to the set.
 * </ul>
 */
public class Keyboard
{
   // symbolic names for key functions
   public static final String ABORT              = "ABORT";
   public static final String APPEND_LINE        = "APPEND-LINE";
   public static final String BACK_TAB           = "BACK-TAB";
   public static final String BACKSPACE          = "BACKSPACE";
   public static final String BELL               = "BELL";
   public static final String BLOCK              = "BLOCK";
   public static final String BOTTOM_COLUMN      = "BOTTOM-COLUMN";
   public static final String BREAK_LINE         = "BREAK-LINE";
   public static final String CANCEL_PICK        = "CANCEL-PICK";
   public static final String CHOICES            = "CHOICES";
   public static final String CLEAR              = "CLEAR";
   public static final String CLOSE              = "CLOSE";
   public static final String COMPILE            = "COMPILE";
   public static final String COPY               = "COPY";
   public static final String CURSOR_DOWN        = "CURSOR-DOWN";
   public static final String CURSOR_LEFT        = "CURSOR-LEFT";
   public static final String CURSOR_RIGHT       = "CURSOR-RIGHT";
   public static final String CURSOR_UP          = "CURSOR-UP";
   public static final String CUT                = "CUT";
   public static final String DEFAULT_POP_UP     = "DEFAULT-POP-UP";
   public static final String DELETE_CHARACTER   = "DELETE-CHARACTER";
   public static final String DELETE_COLUMN      = "DELETE-COLUMN";
   public static final String DELETE_END_LINE    = "DELETE-END-LINE";
   public static final String DELETE_FIELD       = "DELETE-FIELD";
   public static final String DELETE_LINE        = "DELETE-LINE";
   public static final String DELETE_WORD        = "DELETE-WORD";
   public static final String EDITOR_BACKTAB     = "EDITOR-BACKTAB";
   public static final String EDITOR_TAB         = "EDITOR-TAB";
   public static final String END                = "END";
   public static final String END_ERROR          = "END-ERROR";
   public static final String ENTER_MENUBAR      = "ENTER-MENUBAR";
   public static final String ESC_FUNC           = "\u001b";
   public static final String EXIT               = "EXIT";
   public static final String FIND               = "FIND";
   public static final String FIND_NEXT          = "FIND-NEXT";
   public static final String FIND_PREVIOUS      = "FIND-PREVIOUS";
   public static final String GET                = "GET";
   public static final String GO                 = "GO";
   public static final String GOTO               = "GOTO";
   public static final String HELP               = "HELP";
   public static final String HOME               = "HOME";
   public static final String INSERT_COLUMN      = "INSERT-COLUMN";
   public static final String INSERT_FIELD       = "INSERT-FIELD";
   public static final String INSERT_FIELD_DATA  = "INSERT-FIELD-DATA";
   public static final String INSERT_FIELD_LABEL = "INSERT-FIELD-LABEL";
   public static final String INSERT_MODE        = "INSERT-MODE";
   public static final String LEFT_END           = "LEFT-END";
   public static final String MAIN_MENU          = "MAIN-MENU";
   public static final String MOVE               = "MOVE";
   public static final String NEW                = "NEW";
   public static final String NEW_LINE           = "NEW-LINE";
   public static final String NEXT_ERROR         = "NEXT-ERROR";
   public static final String NEXT_FRAME         = "NEXT-FRAME";
   public static final String NEXT_WORD          = "NEXT-WORD";
   public static final String OPEN_LINE_ABOVE    = "OPEN-LINE-ABOVE";
   public static final String OPTIONS            = "OPTIONS";
   public static final String PAGE_DOWN          = "PAGE-DOWN";
   public static final String PAGE_LEFT          = "PAGE-LEFT";
   public static final String PAGE_RIGHT         = "PAGE-RIGHT";
   public static final String PAGE_UP            = "PAGE-UP";
   public static final String PASTE              = "PASTE";
   public static final String PICK               = "PICK";
   public static final String PICK_AREA          = "PICK-AREA";
   public static final String PICK_BOTH          = "PICK-BOTH";
   public static final String PREV_FRAME         = "PREV-FRAME";
   public static final String PREV_WORD          = "PREV-WORD";
   public static final String PUT                = "PUT";
   public static final String RECALL             = "RECALL";
   public static final String REPLACE            = "REPLACE";
   public static final String REPORTS            = "REPORTS";
   public static final String RESUME_DISPLAY     = "RESUME-DISPLAY";
   public static final String RETURN             = "RETURN";
   public static final String RIGHT_END          = "RIGHT-END";
   public static final String SAVE_AS            = "SAVE-AS";
   public static final String SCROLL_LEFT        = "SCROLL-LEFT";
   public static final String SCROLL_MODE        = "SCROLL-MODE";
   public static final String SCROLL_RIGHT       = "SCROLL-RIGHT";
   public static final String SETTINGS           = "SETTINGS";
   public static final String START_RESIZE       = "START-RESIZE";
   public static final String STOP               = "STOP";
   public static final String STOP_DISPLAY       = "STOP-DISPLAY";
   public static final String TAB                = "TAB";
   public static final String TOP_COLUMN         = "TOP-COLUMN";
   public static final String UNIX_END           = "UNIX-END";

   // symbolic names for key actions, which are not also key functions
   public static final String ENDKEY             = "ENDKEY";
   public static final String ERROR              = "ERROR";

   // symbolic names for events
   public static final String ANY_KEY            = "ANY-KEY";
   public static final String ANY_PRINTABLE      = "ANY-PRINTABLE";
   public static final String CHOOSE             = "CHOOSE";
   public static final String DEFAULT_ACTION     = "DEFAULT-ACTION";
   public static final String ENTRY              = "ENTRY";
   public static final String LEAVE              = "LEAVE";
   public static final String ROW_DISPLAY        = "ROW-DISPLAY";
   public static final String VALUE_CHANGED      = "VALUE-CHANGED";
   public static final String WINDOW_CLOSE       = "WINDOW-CLOSE";

   // server events
   public static final String PROCEDURE_COMPLETE = "PROCEDURE-COMPLETE";
   public static final String CONNECT            = "CONNECT";
   public static final String READ_RESPONSE      = "READ-RESPONSE";

   /** Event/action names. */
   private static String[] evNames1 = new String[] 
   {
      ABORT             , APPEND_LINE       , BACKSPACE         ,
      BACK_TAB          , BELL              , BLOCK             ,
      BOTTOM_COLUMN     , BREAK_LINE        , CANCEL_PICK       ,
      CHOICES           , CHOOSE            , CLEAR             ,
      CLOSE             , COMPILE           , COPY              ,
      CURSOR_DOWN       , CURSOR_LEFT       , CURSOR_RIGHT      ,
      CURSOR_UP         , CUT               , DEFAULT_ACTION    ,
      DEFAULT_POP_UP    , DELETE_CHARACTER  , DELETE_COLUMN     ,
      DELETE_END_LINE   , DELETE_FIELD      , DELETE_LINE       ,
      DELETE_WORD       , EDITOR_BACKTAB    , EDITOR_TAB        ,
      END               , ENDKEY            , END_ERROR         ,
      ENTER_MENUBAR     , ENTRY             , ERROR             ,
      EXIT              , FIND              , FIND_NEXT         ,
      FIND_PREVIOUS     , GET               , GO                ,
      GOTO              , HELP              , HOME              ,
      INSERT_COLUMN     , INSERT_FIELD      , INSERT_FIELD_DATA ,
      INSERT_FIELD_LABEL, INSERT_MODE       , LEAVE             ,
      LEFT_END          , MAIN_MENU         , MOVE              ,
      NEW               , NEW_LINE          , NEXT_ERROR        ,
      NEXT_FRAME        , NEXT_WORD         , OPEN_LINE_ABOVE   ,
      OPTIONS           , PAGE_DOWN         , PAGE_LEFT         ,
      PAGE_RIGHT        , PAGE_UP           , PASTE             ,
      PICK              , PICK_AREA         , PICK_BOTH         ,
      PREV_FRAME        , PREV_WORD         , PUT               ,
      RECALL            , REPLACE           , REPORTS           ,
      RESUME_DISPLAY    , RETURN            , RIGHT_END         ,
      SAVE_AS           , SCROLL_LEFT       , SCROLL_MODE       ,
      SCROLL_RIGHT      , SETTINGS          , START_RESIZE      ,
      STOP              , STOP_DISPLAY      , TAB               ,
      TOP_COLUMN        , UNIX_END          , VALUE_CHANGED     ,
      WINDOW_CLOSE      , ROW_DISPLAY       , PROCEDURE_COMPLETE,
      CONNECT           , READ_RESPONSE
   };

   private static final int KA_FIRST_CODE = -101;
   private static final int KA_LAST_CODE = KA_FIRST_CODE - evNames1.length;

   /* key action codes */
   public static final int KA_BACKSPACE = eventCode(BACKSPACE);
   
   /* key action code */
   public static final int KA_BACK_TAB = eventCode(BACK_TAB);

   /* key action code */
   public static final int KA_BELL = eventCode(BELL);

   /* key action code */
   public static final int KA_CLEAR = eventCode(CLEAR);

   /* key action code */
   public static final int KA_CURSOR_DOWN = eventCode(CURSOR_DOWN);

   /* key action code */
   public static final int KA_CURSOR_LEFT = eventCode(CURSOR_LEFT);

   /* key action code */
   public static final int KA_CURSOR_RIGHT = eventCode(CURSOR_RIGHT);

   /* key action code */
   public static final int KA_CURSOR_UP = eventCode(CURSOR_UP);

   /* key action code */
   public static final int KA_DELETE_CHARACTER = eventCode(DELETE_CHARACTER);

   /* key action code */
   public static final int KA_ENDKEY = eventCode(ENDKEY);

   /* key action code */
   public static final int KA_END_ERROR = eventCode(END_ERROR);

   /* key action code */
   public static final int KA_ERROR = eventCode(ERROR);

   /* key action code */
   public static final int KA_GO = eventCode(GO);

   /* key action code */
   public static final int KA_HELP = eventCode(HELP);

   /* key action code */
   public static final int KA_HOME = eventCode(HOME);

   /* key action code */
   public static final int KA_INSERT_MODE = eventCode(INSERT_MODE);

   /* key action code */
   public static final int KA_LEFT_END = eventCode(LEFT_END);

   /* key action code */
   public static final int KA_RECALL = eventCode(RECALL);

   /* key action code */
   public static final int KA_RETURN = eventCode(RETURN);

   /* key action code */
   public static final int KA_RIGHT_END = eventCode(RIGHT_END);

   /* key action code */
   public static final int KA_STOP = eventCode(STOP);

   /* key action code */
   public static final int KA_TAB = eventCode(TAB);

   /* key action code */
   public static final int KA_PAGE_UP = eventCode(PAGE_UP);

   /* key action code */
   public static final int KA_PAGE_DOWN = eventCode(PAGE_DOWN);

   /* key action code */
   public static final int KA_REPLACE = eventCode(REPLACE);

   /* key action code */
   public static final int KA_EDITOR_TAB = eventCode(EDITOR_TAB);

   /* key action code */
   public static final int KA_APPEND_LINE = eventCode(APPEND_LINE);

   /* key action code */
   public static final int KA_DELETE_END_LINE = eventCode(DELETE_END_LINE);

   /* key action code */
   public static final int KA_DELETE_LINE = eventCode(DELETE_LINE);

   /* key action code */
   public static final int KA_DELETE_WORD = eventCode(DELETE_WORD);

   /* key action code */
   public static final int KA_EDITOR_BACKTAB = eventCode(EDITOR_BACKTAB);

   /* key action code */
   public static final int KA_END = eventCode(END);

   /* key action code */
   public static final int KA_NEW_LINE = eventCode(NEW_LINE);

   /* key action code */
   public static final int KA_NEXT_WORD = eventCode(NEXT_WORD);

   /* key action code */
   public static final int KA_OPEN_LINE_ABOVE = eventCode(OPEN_LINE_ABOVE);

   /* key action code */
   public static final int KA_PAGE_LEFT = eventCode(PAGE_LEFT);

   /* key action code */
   public static final int KA_PAGE_RIGHT = eventCode(PAGE_RIGHT);

   /* key action code */
   public static final int KA_PREV_WORD = eventCode(PREV_WORD);
   
   /* synthetic event code */
   public static final int SE_NEXT_FRAME = eventCode(NEXT_FRAME);

   /* synthetic event code */
   public static final int SE_PREV_FRAME = eventCode(PREV_FRAME);

   /* synthetic event code */
   public static final int SE_ENTRY = eventCode(ENTRY);

   /* synthetic event code */
   public static final int SE_LEAVE = eventCode(LEAVE);

   /* synthetic event code */
   public static final int SE_CHOOSE = eventCode(CHOOSE);

   /* synthetic event code */
   public static final int SE_VALUE_CHANGED = eventCode(VALUE_CHANGED);
   
   /* synthetic event code */
   public static final int SE_DEFAULT_ACTION = eventCode(DEFAULT_ACTION);
   
   /* synthetic event code */
   public static final int SE_PROCEDURE_COMPLETE = eventCode(PROCEDURE_COMPLETE);
   
   /* synthetic event code */
   public static final int SE_CONNECT = eventCode(CONNECT);
   
   /* synthetic event code */
   public static final int SE_READ_RESPONSE = eventCode(READ_RESPONSE);
   
   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer(); 

   /**
    * Translates the event code for the high level events and returns the
    * symbolic name of the event.
    * <p>
    * This method has no corresponding function in Progress.
    * 
    * @param   code
    *          event code to translate
    * @return  high level event name or key label string
    *  
    */
   public static String eventName(int code)
   {
      if (code <= KA_FIRST_CODE && code >= KA_LAST_CODE)
         return evNames1[-code + KA_FIRST_CODE]; 
      
      return keyLabel(code);
   }

   /**
    * Find the action code corresponding to a given event name.
    * 
    * @param    name
    *           Event name.
    *
    * @return   The action code or -1 if no action code exists for the name.
    */
   public static int eventCode(String name)
   {
      if (name == null)
         return -1;
      
      name = name.toUpperCase();
      
      for (int i = 0; i < evNames1.length; i++)
      {
         if (name.equals(evNames1[i]))
            return KA_FIRST_CODE - i;
      }

      return -1;
   }
   
   /**
    * Translates a key label into a numeric key code.
    * <p>
    * The rules of this translation are:
    * <ul>
    *    <li> if the label is a single character, it should match one of 
    *         the basic key names as is;
    *    <li> the label is uppercased;
    *    <li> the label is looked up in the basic key names for an exact match;
    *    <li> the label is looked up in the basic key names for ending match;
    *         if this is the case, the prefix should be one of:
    *         <ul>
    *            <li>CTRL-
    *            <li>SHIFT-
    *            <li>ESC-
    *            <li>SHIFT-ESC-  
    *            <li>CTRL-SHIFT-  
    *            <li>CTRL-ESC-  
    *            <li>CTRL-SHIFT-ESC-
    *         </ul>
    *    <li> if the result is a number within the range of 3137-3162, it is
    *         translated into the range 1025-1050, because Progress thinks
    *         names like CTRL-ESC-A should be treated as ESC-CTRL-A.
    *    <li> lookup the label in the alternate key label list
    * </ul>
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>KEYCODE </b>
    * 
    * @param   label
    *          Label to evaluate.
    * @return  key code or -1 if the label is invalid.
    */
   public static int keyCode(String label)
   {
      if (label.length() == 0)
         return -1;

      WorkArea kb = work.obtain();

      if (label.length() == 1)
         return kb.findKey(label);

      String lab = label.toUpperCase();
      int code = kb.findKey(lab);
      if (code != -1)
         return code;

      code = kb.findKeyEnd(lab);
      if (code >= 3137 && code <= 3162)
         return code - 2112;
      
      // check alternate key labels for a match
      if (code == -1)
      {
         Integer kc = (Integer) kb.altLabel.get(label);
         
         if (kc != null)
         {
            code = kc.intValue();
         }
      }

      return code;
   }

   /**
    * Translates a numeric key code into a key label.
    * <p>
    * Negative codes produce empty label. Codes above 4095 produce a label
    * of the code modulo 4096.
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>KEYLABEL </b>
    * 
    * @param   code
    *          key code to translate
    * @return  key label string or "" if the key code is invalid
    *  
    */
   public static String keyLabel(int code)
   {
      if (code < 0)
         return "";

      code = code % 4096;

      // locate the rule region
      WorkArea kb = work.obtain();

      // rule 1: [0-511], [520-521], [525], [539], [656-1535],
      //         [1544-1545], [1549], [1563], [1663-2559],
      //         [2568-2569], [2573], [2587], [2687-3583], [3592-3593], 
      //         [3597], [3611], [3711-4095].
      if ((code >= 0    && code <= 511 )  ||
          (code >= 520  && code <= 521 )  ||
          (code == 525)                   ||
          (code == 539)                   ||
          (code >= 656  && code <= 1535)  ||
          (code >= 1544 && code <= 1545)  ||
          (code == 1549)                  ||
          (code == 1563)                  ||
          (code >= 1663 && code <= 2559)  ||
          (code >= 2568 && code <= 2569)  ||
          (code == 2573)                  ||
          (code == 2587)                  ||
          (code >= 2687 && code <= 3583)  ||
          (code >= 3592 && code <= 3593)  ||
          (code == 3597)                  ||
          (code == 3611)                  ||
          (code >= 3711 && code <= 4095))
         return kb.applyRule1(code);
      
      // rule 3: [1614-1662], [2638-2686], [3662-3710]
      if ((code >= 1614 && code <= 1662) ||
          (code >= 2638 && code <= 2686) ||
          (code >= 3662 && code <= 3710))
         return kb.applyRule3(code);

      // the rest is rule 2
      return kb.applyRule2(code);
   }

   /**
    * Translates a string of text, which is a key function, into the key label
    * of a keyboard key, currently assigned for this function.
    * Since key functions can be assigned to more than one key labels, the
    * first found assignment is used in translation.
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>KBLABEL </b>
    * <p>
    * The main data structure
    * is a Map where the function name is the key and the value is an ordered
    * Set, containing key labels this key function is assigned to. This Map
    * corresponds to the Table 6-4 of the Progress Programming Handbook, 
    * p.6-19. This map is used for the translation. The first key label in the
    * Set for this function is returned. If the set is empty (no assignment)
    * or no such key function is found in the map (unrecognizable function),
    * the original function name is returned in upper case.
    *
    * @param   function
    *          text, describing the key function. case doesn't matter
    * @return  text, describing the key label. If the key function is
    *          unrecognizable, or it is a valid key function but currently 
    *          unassigned, it is returned translated to upper case.
    */
   public static String kbLabel(String function)
   {
      // normalize the function name
      String key = function.toUpperCase();

      // search the map
      WorkArea kb = work.obtain();

      Set s = (Set)kb.f2l.get(key);
      if (s == null || s.isEmpty())
         return function.length() == 1 ? function : key;

      // return the first key label if any
      Iterator iter = s.iterator();
      return (String)iter.next();
   }

   /**
    * Evaluates key code and returns event name assigned to specified key.
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>KEYFUNCTION </b>
    * 
    * @param   code
    *          Code to evaluate.
    *          
    * @return  key function or "" if the key code is invalid.
    */
   public static String keyFunction(int code)
   {
      // special case: code 606 - it has no label, so there is no way to remap
      // it; this is the only known case 
      if (code == 606)
         return NEXT_FRAME;
      
      WorkArea kb = work.obtain();

      // get key label for this code
      String label = keyLabel(code);
      
      // if there is no label, we are done
      if (label.length() == 0)
         return "";
      
      // range of 32-255: the key label is the key function 
      if (code >= 32 && code <= 255 && code != 127)
         return label;

      // get key function for this key label
      String function = (String)kb.l2f.get(label);

      // return the function if it exists otherwise return the empty string
      return (function == null) ? "" : function;
   }

   /**
    * Returns the key function assigned to this key label.
    * <p>
    * 
    * @param   label
    *          key label
    * @return  key function or "" if the key label is invalid.
    */
   public static String keyFunction(String label)
   {
      WorkArea kb = work.obtain();

      // normalize the key label
      String lab = null;
      if (label.length() == 1)
         lab = label;
      else
         lab = label.toUpperCase();

      // validate the key label
      int code = keyCode(lab);
      if (code == -1)
         return "";

      // range of 32-255: the key label is the key function 
      if (code >= 32 && code <= 255 && code != 127)
         return label;

      // get key function for this key label
      String function = (String)kb.l2f.get(lab);

      return (function == null) ? "" : function;
   }

   /**
    * Defines a specific action for the key.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>ON key-label key-function </b>
    * 
    * @param   keyLabel
    *          the label of the key
    * @param   keyFunction
    *          the action to take when the user presses the key
    */
   public static void remapKey(String keyLabel, String keyFunction)
   {
      WorkArea kb = work.obtain();

      // normalize the function name and key label
      String func = keyFunction.toUpperCase();
      String lab  = keyLabel.toUpperCase();

      // make sure the key label exists
      if (keyCode(lab) == -1)
         return;
      
      // does the key function exist?
      if (kb.f2l.get(func) == null)
      {
         // test if this is a known key function
         if (eventCode(func) != -1)
         {
            // this is a key function that has no default and/or has not yet
            // been previously added, create it now
            kb.addFunc(func, new String[] {lab});
         }
         else
         {
            // invalid key function
            return;
         }
      }

      // remap the key
      String oldFunc = (String)kb.l2f.get(lab);
      if (oldFunc != null)
      {
         Set s = (Set)kb.f2l.get(oldFunc);
         s.remove(lab);
      }

      Set s = (Set)kb.f2l.get(func);
      s.add(lab);

      kb.l2f.put(lab, func);
   }

   /**
    * Evaluates key code and returns a generic action code.
    * 
    * @param   code
    *          key code to evaluate.
    * @return  generic action code or the key code unchanged
    */
   public static int keyAction(int code)
   {
      // get key function for this key code
      String function = keyFunction(code);
      if (function.length() == 0)
         return code;

      // get key action code for this key function
      int action = eventCode(function);
      if (action == -1)
         return code;

      return action;
   }

   /**
    * Translates a key action into a typical key code.
    * This translation is done only to those actions having a defined KA_
    * constant and a meaningful standard code. Any other action returns
    * <code>KeyEvent.CHAR_UNDEFINED</code>.
    * 
    * @param   action
    *          key action code to translate
    * @return  a key code that is normally associated with the action for use
    *          in compatibility mode in CHARVA.
    */
   public static int keyCodeForAction(int action)
   {
      if (action == KA_BACKSPACE)
         return Key.VK_BACKSPACE;
      if (action == KA_BACK_TAB)
         return Key.VK_BACK_TAB;
      if (action == KA_BELL)
         return Key.VK_BELL;
      if (action == KA_CLEAR)
         return KeyInput.CHAR_UNDEFINED;
      if (action == KA_CURSOR_DOWN)
         return Key.VK_CURSOR_DOWN;
      if (action == KA_CURSOR_LEFT)
         return Key.VK_CURSOR_LEFT;
      if (action == KA_CURSOR_RIGHT)
         return Key.VK_CURSOR_RIGHT;
      if (action == KA_CURSOR_UP)
         return Key.VK_CURSOR_UP;
      if (action == KA_DELETE_CHARACTER)
         return Key.VK_DELETE;
      if (action == KA_ENDKEY)
         return KeyInput.CHAR_UNDEFINED;
      if (action == KA_END_ERROR)
         return KeyInput.CHAR_UNDEFINED;
      if (action == KA_ERROR)
         return KeyInput.CHAR_UNDEFINED;
      if (action == KA_GO)
         return KeyInput.CHAR_UNDEFINED;
      if (action == KA_HELP)
         return KeyInput.CHAR_UNDEFINED;
      if (action == KA_HOME)
         return Key.VK_HOME;
      if (action == KA_INSERT_MODE)
         return Key.VK_INSERT;
      if (action == KA_LEFT_END)
         return Key.VK_HOME;
      if (action == KA_RECALL)
         return KeyInput.CHAR_UNDEFINED;
      if (action == KA_RETURN)
         return Key.VK_ENTER;
      if (action == KA_RIGHT_END)
         return Key.VK_END;
      if (action == KA_STOP)
         return KeyInput.CHAR_UNDEFINED;
      if (action == KA_TAB)
         return Key.VK_TAB;
   
      return KeyInput.CHAR_UNDEFINED;
   }

   /**
    * Resets the Keyboard class. The existing context local instance of the
    * class is released and a new one is created with all default key
    * mappings.
    */
   public static void resetKeyboard()
   {
      work.reset();
   }
   
   /**         
    * Detects if the event is one of those that can be described as
    * "printable" in Progress terms. If the event name is more than a
    * single character, it is assumed to be non-printable. Any single
    * character event names have their character checked to determine if
    * it is a control character.  If so, it is considered non-printable.
    * <p>
    * This method has not been well tested in comparison to Progress as
    * of 03/14/2008.
    *
    * @param    event
    *           The name of the event to check.  Must not be
    *           <code>null</code>.
    *
    * @return   <code>true</code> if this is a single character name whose
    *           character is not an ISO control code (based on a query to
    *           the <code>Character</code> class.
    */
   public static boolean isPrintable(String event)
   {
      return event.length() == 1 && !Character.isISOControl(event.charAt(0));
   }

   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {
      /** mapping of key functions into key labels the function is mapped to*/
      private Map f2l = new HashMap();
      
      /** mapping of key labels into key functions that are assigned to them*/
      private Map l2f = new HashMap();
      
      /** alternate key labels */
      private Map altLabel = new HashMap();
      
      /** basic key names */
      private String[] basicKeys = 
      {
         "CTRL-@", "CTRL-A", "CTRL-B", "CTRL-C",              
         "CTRL-D", "CTRL-E", "CTRL-F", "CTRL-G",              
         "BACKSPACE", "TAB", "CTRL-J", "CTRL-K",              
         "CTRL-L", "RETURN", "CTRL-N", "CTRL-O",              
         "CTRL-P", "CTRL-Q", "CTRL-R", "CTRL-S",              
         "CTRL-T", "CTRL-U", "CTRL-V", "CTRL-W",              
         "CTRL-X", "CTRL-Y", "CTRL-Z", "ESC",                 
         "CTRL-\\", "CTRL-]", "CTRL-^", "CTRL-_",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "DEL",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "", "", "", "",
         "F0", "F1", "F2", "F3",    
         "F4", "F5", "F6", "F7",                  
         "F8", "F9", "F10", "F11",                  
         "F12", "F13", "F14", "F15",                 
         "F16", "F17", "F18", "F19",                 
         "F20", "F21", "F22", "F23",                 
         "F24", "F25", "F26", "F27",                 
         "F28", "F29", "F30", "F31",                 
         "F32", "F33", "F34", "F35",                 
         "F36", "F37", "F38", "F39",                 
         "F40", "F41", "F42", "F43",                 
         "F44", "F45", "F46", "F47",                 
         "F48", "F49", "F50", "F51",                 
         "F52", "F53", "F54", "F55",                 
         "F56", "F57", "F58", "F59",                 
         "F60", "F61", "F62", "F63",                 
         "F64", "F65", "F66", "F67",                 
         "F68", "F69", "F70", "F71",                 
         "F72", "F73", "F74", "F75",                 
         "F76", "F77", "F78", "F79",                 
         "F80", "F81", "F82", "F83",                 
         "F84", "F85", "F86", "F87",                 
         "F88", "F89", "F90", "F91",                 
         "F92", "F93", "F94", "F95",                 
         "F96", "F97", "F98", "F99",                 
         "PF0", "PF1", "PF2", "PF3",    
         "PF4", "PF5", "PF6", "PF7",                  
         "PF8", "PF9", "PF10", "PF11",                  
         "PF12", "PF13", "PF14", "PF15",                 
         "PF16", "PF17", "PF18", "PF19",                 
         "PF20", "PF21", "PF22", "PF23",                 
         "PF24", "PF25", "PF26", "PF27",                 
         "PF28", "PF29", "PF30", "PF31",                 
         "PF32", "PF33", "PF34", "PF35",                 
         "PF36", "PF37", "PF38", "PF39",                 
         "PF40", "PF41", "PF42", "PF43",                 
         "PF44", "PF45", "PF46", "PF47",                 
         "PF48", "PF49", "PF50", "PF51",                 
         "PF52", "PF53", "PF54", "PF55",                 
         "PF56", "PF57", "PF58", "PF59",                 
         "PF60", "PF61", "PF62", "PF63",                 
         "PF64", "PF65", "PF66", "PF67",                 
         "PF68", "PF69", "PF70", "PF71",                 
         "PF72", "PF73", "PF74", "PF75",                 
         "PF76", "PF77", "PF78", "PF79",                 
         "PF80", "PF81", "PF82", "PF83",                 
         "PF84", "PF85", "PF86", "PF87",                 
         "PF88", "PF89", "PF90", "PF91",                 
         "PF92", "PF93", "PF94", "PF95",                 
         "PF96", "PF97", "PF98", "PF99",                 
         "", "CURSOR-UP", "CURSOR-DOWN", "CURSOR-RIGHT",        
         "CURSOR-LEFT", "HOME", "END", "PAGE-UP",             
         "PAGE-DOWN", "BACK-TAB", "INS", "HELP-KEY",      
         "DEL-CHAR", "EXECUTE", "PAGE", "FIND",                 
         "INS-LINE", "DEL-LINE", "LINE-ERASE", "PAGE-ERASE",           
         "BACKSPACE", "TAB", "EXIT", "",                     
         "", "RETURN", "", "",                     
         "", "", "", "",                     
         "", "", "", "ERASE",                
         "WHITE", "BLUE", "RED", "ESC",            
         "", "CTRL-BREAK", "", "",                     
         "", "", "", "",                     
         "", "", "", "U1",                   
         "U2", "U3", "U4", "U5",                   
         "U6", "U7", "U8", "U9",                   
         "U10", "", "", "",                     
         "", "", "", "",                     
         "", "", "", "",                     
         "", "", "", "",                     
         "", "", "", "",                     
         "", "", "", "",                     
         "", "", "DO", "SELECT",                    
         "REMOVE", "INSERT-HERE", "", "",                     
         "", "", "", "",                     
         "", "", "", "",                     
         "", "", "", "",                     
         "", "", "", "",                     
         "", "MOUSE-SELECT-UP", "MOUSE-MOVE-UP", "MOUSE-MENU-UP",         
         "MOUSE-EXTEND-UP", "", "", "",                     
         "", "MOUSE-SELECT-DOWN", "MOUSE-MOVE-DOWN", "MOUSE-MENU-DOWN",       
         "MOUSE-EXTEND-DOWN", "", "", "",                     
         "", "MOUSE-SELECT-CLICK", "MOUSE-MOVE-CLICK", "MOUSE-MENU-CLICK",      
         "MOUSE-EXTEND-CLICK", "", "", "",                     
         "", "MOUSE-SELECT-DRAG", "MOUSE-MOVE-DRAG", "MOUSE-MENU-DRAG",       
         "MOUSE-EXTEND-DRAG", "", "", "",                     
         "MOUSE-MOVE", "", "", "",                     
         "", "", "", "",                     
         "", "MOUSE-SELECT-DBLCLICK", "MOUSE-MOVE-DBLCLICK",
         "MOUSE-MENU-DBLCLICK", "MOUSE-EXTEND-DBLCLICK", "", "", ""                     
      };
      
      /**
       * Constructor.
       */
      private WorkArea()
      {
         // filling the map of key functions with standard mappings
   
         addFunc(APPEND_LINE, new String[] {"CTRL-A"});
   
         addFunc(BACK_TAB,      new String[] {"BACK-TAB", "CTRL-U"});
         addFunc(BACKSPACE,     new String[] {"BACKSPACE", "CTRL-H", "DEL-CHAR"});
         addFunc(BELL,          new String[] {"BELL"});
         addFunc(BLOCK,         new String[] {"CTRL-V"});
         addFunc(BOTTOM_COLUMN, new String[] {"ESC-CTRL-B"});
         addFunc(BREAK_LINE,    new String[] {"ESC-B"});
   
         addFunc(CANCEL_PICK,  new String[] {"ESC-CTRL-X"});
         addFunc(CHOICES,      new String[] {"ESC-BACKSPACE", "ESC-CTRL-H"});
         addFunc(CLEAR,        new String[] {"F8", "CTRL-Z"});
         addFunc(CLOSE,        new String[] {"ESC-Z"});
         addFunc(COMPILE,      new String[] {"ESC-P"});
         addFunc(COPY,         new String[] {"F11", "ESC-C"});
         addFunc(CURSOR_DOWN,  new String[] {"CURSOR-DOWN", "CTRL-J"});
         addFunc(CURSOR_LEFT,  new String[] {"CURSOR-LEFT", "CTRL-O"});
         addFunc(CURSOR_RIGHT, new String[] {"CURSOR-RIGHT", "CTRL-L"});
         addFunc(CURSOR_UP,    new String[] {"CURSOR-UP", "CTRL-K"});
         addFunc(CUT,          new String[] {"F10", "ESC-X"});
   
         addFunc(DEFAULT_POP_UP,   new String[] {"ESC-U"});
         addFunc(DELETE_CHARACTER, new String[] {"DEL", "DELETE"});
         addFunc(DELETE_COLUMN,    new String[] {"ESC-CTRL-Z"});
         addFunc(DELETE_END_LINE,  new String[] {"ESC-K"});
         addFunc(DELETE_FIELD,     new String[] {"ESC-CTRL-D"});
         addFunc(DELETE_LINE,      new String[] {"CTRL-D"});
         addFunc(DELETE_WORD,      new String[] {"ESC-D"});
   
         addFunc(EDITOR_BACKTAB, new String[] {"CTRL-B"});
         addFunc(EDITOR_TAB,     new String[] {"CTRL-G", "TAB"});
         addFunc(END,            new String[] {"END", "ESC->"});
         addFunc(END_ERROR,      new String[] {"F4", "CTRL-E", "PF4"});
         addFunc(ENTER_MENUBAR,  new String[] {"F3", "PF3", "ESC-M"});
         addFunc(ESC_FUNC,       new String[] {"ESC"});
         addFunc(EXIT,           new String[] {"ESC-Q"});
   
         addFunc(FIND,          new String[] {"CTRL-F"});
         addFunc(FIND_NEXT,     new String[] {"ESC-F"});
         addFunc(FIND_PREVIOUS, new String[] {"ESC-I"});
   
         addFunc(GET,  new String[] {"F5", "ESC-O"});
         addFunc(GO,   new String[] {"F1", "CTRL-X", "PF1"});
         addFunc(GOTO, new String[] {"ESC-G"});
   
         addFunc(HELP,  new String[] {"F2", "ESC-?", "PF2"});
         addFunc(HOME,  new String[] {"HOME", "ESC-<", "ESC-H"});
   
         addFunc(INSERT_COLUMN,      new String[] {"ESC-CTRL-N"});
         addFunc(INSERT_FIELD,       new String[] {"ESC-CTRL-G"});
         addFunc(INSERT_FIELD_DATA,  new String[] {"ESC-CTRL-F"});
         addFunc(INSERT_FIELD_LABEL, new String[] {"ESC-CTRL-E"});
         addFunc(INSERT_MODE,        new String[] {"F9", "CTRL-T", "INS"});
   
         addFunc(LEFT_END, new String[] {"ESC-CURSOR-LEFT"});
   
         addFunc(MAIN_MENU, new String[] {"ESC-RETURN", "ESC-CTRL-M"});
         addFunc(MOVE,      new String[] {"ESC-CTRL-V"});
   
         addFunc(NEW,        new String[] {"ESC-N"});
         addFunc(NEW_LINE,   new String[] {"CTRL-N"});
         addFunc(NEXT_ERROR, new String[] {"ESC-E"});
         addFunc(NEXT_FRAME, new String[] {"ESC-TAB", "ESC-CTRL-I"});
         addFunc(NEXT_WORD,  new String[] {"CTRL-W"});
   
         addFunc(OPEN_LINE_ABOVE, new String[] {"ESC-L"});
         addFunc(OPTIONS,         new String[] {"ESC-CTRL-O"});
   
         addFunc(PAGE_DOWN,  new String[] {"ESC-CURSOR-DOWN", "PAGE-DOWN"});
         addFunc(PAGE_LEFT,  new String[] {"ESC-W"});
         addFunc(PAGE_RIGHT, new String[] {"ESC-Y"});
         addFunc(PAGE_UP,    new String[] {"ESC-CURSOR-UP", "PAGE-UP"});
         addFunc(PASTE,      new String[] {"F12", "ESC-V"});
         addFunc(PICK,       new String[] {"ESC-CTRL-P"});
         addFunc(PICK_AREA,  new String[] {"ESC-CTRL-W"});
         addFunc(PICK_BOTH,  new String[] {"ESC-CTRL-Q"});
         addFunc(PREV_FRAME, new String[] {"ESC-CTRL-U"});
         addFunc(PREV_WORD,  new String[] {"CTRL-P"});
         addFunc(PUT,        new String[] {"F6", "ESC-S"});
   
         addFunc(RECALL,  new String[] {"F7", "CTRL-R"});
         addFunc(REPLACE, new String[] {"ESC-R"});
         addFunc(REPORTS, new String[] {"ESC-CTRL-A"});
         addFunc(RESUME_DISPLAY, new String[] {"CTRL-Q"});
         addFunc(RETURN,         new String[] {"RETURN", "CTRL-M", "ENTER"});
         addFunc(RIGHT_END,      new String[] {"ESC-CURSOR-RIGHT"});
   
         addFunc(SAVE_AS,  new String[] {"ESC-A"});
         addFunc(SCROLL_LEFT,  new String[] {"ESC-CTRL-L"});
         addFunc(SCROLL_MODE,  new String[] {"ESC-T"});
         addFunc(SCROLL_RIGHT, new String[] {"ESC-CTRL-R"});
         addFunc(SETTINGS,     new String[] {"ESC-CTRL-@"});
         addFunc(START_RESIZE, new String[] {"ESC-1"});
         addFunc(STOP_DISPLAY, new String[] {"CTRL-S"});
   
         addFunc(TAB,        new String[] {"TAB", "CTRL-I"});
         addFunc(TOP_COLUMN, new String[] {"ESC-CTRL-T"});
   
         // special entries for key actions
         addFunc(ENDKEY, new String[] {});
         addFunc(ERROR, new String[] {});
   
         // filling the map of alternate key labels
         addAltLabel("BELL", 7);
         addAltLabel("BS",   8);
         addAltLabel("LINEFEED", 10);
         addAltLabel("LF", 10);
         addAltLabel("FORMFEED", 12);
         addAltLabel("FF", 12);
         addAltLabel("ENTER", 13);
         addAltLabel("ESCAPE", 27);
         addAltLabel("CANCEL", 127);
         addAltLabel("UP", 501);
         addAltLabel("UP-ARROW", 501);
         addAltLabel("DOWN", 502);
         addAltLabel("DOWN-ARROW", 502);
         addAltLabel("RIGHT", 503);
         addAltLabel("RIGHT-ARROW", 503);
         addAltLabel("LEFT", 504);
         addAltLabel("LEFT-ARROW", 504);
         addAltLabel("ESC-H", 505);
         addAltLabel("PGUP", 507);
         addAltLabel("PREV-PAGE", 507);
         addAltLabel("PREV-SCREEN", 507);
         addAltLabel("PGDN", 508);
         addAltLabel("NEXT-PAGE", 508);
         addAltLabel("NEXT-SCREEN", 508);
         addAltLabel("SHIFT-TAB", 509);
         addAltLabel("INSERT", 510);
         addAltLabel("INS-CHAR", 510);
         addAltLabel("INS-C", 510);
         addAltLabel("INSERT-HERE", 510);
         addAltLabel("DEL-CHAR", 512);
         addAltLabel("DEL-C", 512);
         addAltLabel("DELETE-CHAR", 512);
         addAltLabel("INS-L", 516);
         addAltLabel("LINE-INS", 516);
         addAltLabel("DEL-L", 517);
         addAltLabel("LINE-DEL", 517);
   
         // WARNING: I18N support is not complete here --> there may be
         //          cases such as DBCS support where this code will fail
         //          or will yield different results from Progress
         
         // completing the table of basic key names with the characters
         for (char i = 32; i <= 126; i ++)
         {
            basicKeys[i] = new String(new char[] {i});
         }
   
         String cset = Utils.getCharsetOverride();
         char[] ch   = new char[1];
         
         for (int i = 128; i <= 255; i ++)
         {
            // if this fails, the extended ASCII range of characters will
            // all be the empty string
            ch[0] = Utils.toChar(i, cset);
            
            // some kind of encoding issue, bail out
            if (ch[0] == -1)
               break;
            
            basicKeys[i] = new String(ch);
         }
      }
      
      /**
       * Creates an entry in the key function map.
       * 
       * @param  func
       *         key function
       * @param  labels
       *         array of key labels this function is assigned to
       */
      private void addFunc(String func, String[] labels)
      {
         Set s = new LinkedHashSet();
         
         for (int i = 0; i < labels.length; i ++)
         {
            s.add(labels[i]);
            l2f.put(labels[i], func);
         }
            
         f2l.put(func, s);
      }

      /**
       * Creates an entry in the alternate key label map.
       * 
       * @param  label
       *         alternate key label in upper case
       * @param  code
       *         key code
       */
      private void addAltLabel(String label, int code)
      {
         altLabel.put(label, new Integer(code));
      }

      /**
       * Looks up the basic key names table for an exact match and returns the
       * index.
       * <p>
       * 
       * @param   label
       *          label to find the match for
       * @return  index or -1 if no match found
       */
      private int findKey(String label)
      {
         for (int i = 0; i < basicKeys.length; i ++)
            if (label.equals(basicKeys[i]))
                return i;
   
         return -1;
      }
   
      /**
       * Looks up the basic key names table for an ending match and returns the
       * code that takes into account all valid prefixes.
       * <p>
       * 
       * @param   label
       *          label to find the match for
       * @return  index or -1 if no match found
       */
      private int findKeyEnd(String label)
      {
         int l = label.length();
   
         for (int i = 0; i < basicKeys.length; i ++)
         {
            if (basicKeys[i].length() == 0)
               continue;
            if (label.endsWith(basicKeys[i]))
            {
               // this is a candidate; check prefix validity
               String prefix = label.substring(0, l - basicKeys[i].length());
               if (prefix.equals("CTRL-"))
                  return i + 2048;
               if (prefix.equals("SHIFT-"))
                  return i + 512;
               if (prefix.equals("ESC-"))
                  return i + 1024;
               if (prefix.equals("SHIFT-ESC-"))
                  return i + 1536;
               if (prefix.equals("CTRL-SHIFT-"))
                  return i + 2560;
               if (prefix.equals("CTRL-ESC-"))
                  return i + 3072;
               if (prefix.equals("CTRL-SHIFT-ESC-"))
                  return i + 3584;
            }
         }
   
         return -1;
      }
      
      /**
       * Applies rule 1 translation to the code and returns a label.
       * <p>
       * Rule 1 means the code is within a range where:
       * <ul>
       *   <li>codes 0-511 index the table of basic key names;
       *   <li>codes outside of 0-511 range produce regular labels with CTRL-,
       *       SHIFT-, ESC- prefixes and their combinations.
       * </ul>
       * 
       * @param   code
       *          key code to translate
       * @return  key label string
       *  
       */
      private String applyRule1(int code)
      {
         String basic = basicName1(code);
         if (basic.length() == 0)
            return basic;
   
         if (code >= 3584)
            return new String("CTRL-SHIFT-ESC-" + basic);
   
         if (code >= 3072)
            return new String("CTRL-ESC-" + basic);
   
         if (code >= 2560)
            return new String("CTRL-SHIFT-" + basic);
   
         if (code >= 2048)
            return new String("CTRL-" + basic);
   
         if (code >= 1536)
            return new String("SHIFT-ESC-" + basic);
   
         if (code >= 1024)
            return new String("ESC-" + basic);
   
         if (code >= 512)
            return new String("SHIFT-" + basic);
   
         return basic;
      }
   
      /**
       * Applies rule 2 translation to the code and returns a label.
       * <p>
       * Rule 2 means the code is within a range where:
       * <ul>
       *   <li>codes 512-655 index the table of basic key names;
       *   <li>codes outside of 512-655 range produce regular labels with 
       *       SHIFT-ESC-,  CTRL-SHIFT- and CTRL-SHIFT-ESC- prefixes.
       * </ul>
       * 
       * @param   code
       *          key code to translate
       * @return  key label string
       *  
       */
      private String applyRule2(int code)
      {
         String basic = basicName2(code);
         if (basic.length() == 0)
            return basic;
   
         if (code >= 3584)
            return new String("CTRL-ESC-" + basic);
   
         if (code >= 2560)
            return new String("CTRL-" + basic);
   
         if (code >= 1536)
            return new String("ESC-" + basic);
   
         return basic;
      }
   
      /**
       * Applies rule 3 translation to the code and returns a label.
       * <p>
       * Rule 3 means the code is within a range where the result is always
       * an empty string.
       * 
       * @param   code
       *          key code to translate
       * @return  key label string
       *  
       */
      private String applyRule3(int code)
      {
         return "";
      }
   
      /**
       * Returns the basic name for keys in the rage 0-511.
       * 
       * @param   code
       *          key code to translate
       * @return  key label string
       *  
       */
      private String basicName1(int code)
      {
         int idx = code & 0x01ff;
         
         return basicKeys[idx];
      }
   
      /**
       * Returns the basic name for keys in the rage 512-655.
       * 
       * @param   code
       *          key code to translate
       * @return  key label string
       *  
       */
      private String basicName2(int code)
      {
         int idx = code & 0x01ff;
         if (idx > 143)
            return basicKeys[idx];
         
         return basicKeys[idx + 512];
      }
   
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the outer class.
    */
   private static class ContextContainer
   extends ContextLocal
   {
      /**
       * Locates the context-local instance of the outer class.
       *
       * @return   The instance of the outer class.
       */
      public WorkArea obtain() 
      {
         return (WorkArea) this.get();
      }   
      
      /**
       * Resets the context-local instance of the outer class.
       */
      public void reset() 
      {
         this.set(this.initialValue());
      }   

      /**
       * Instantiates the context local copy of the class.
       *
       * @return   The newly instantiated class.
       */
      protected synchronized Object initialValue()
      {
         WorkArea wa = new WorkArea();
         return wa;
      }   
   }   
   
   /**
    * Simple command line test driver which prints out a report of our
    * key mappings.
    * <p>
    * Optionally, the starting and ending key code as integer values can be
    * passed as arguments.  If not passed a report from 0 through 255 will
    * be displayed.
    * <p>
    * Syntax: Keyboard [startCode endCode]
    *
    * @param    args
    *           Command line arguments array.
    */
   public static void main(String[] args)
   {
      int start = 0;
      int end   = 4096;
      String[] header = new String[]
      {
         "Key Code Key Label                      Key Function",
         "-------- ------------------------------ " +
         "------------------------------"
      };
      String spec = "%8d %30s %30s";
      
      try
      {
         if (args.length > 1)
         {
            // event name by code?
            if (args[0].equals("-e"))
            {
               for (int i = 1; i < args.length; i ++)
               {
                  int    code = Integer.parseInt(args[i]);
                  String name = Keyboard.eventName(code);
                  System.out.println("code " + args[i] + " is " + name);
               }
               
               return;
            }
            
            // table
            start = Integer.parseInt(args[0]);
            end = Integer.parseInt(args[1]);
         }
         
         System.out.println(header[0]);
         System.out.println(header[1]);
         
         for (int i = start; i < end; i++)
         {
            Object[] parms = new Object[]
            {
               new Integer(i),
               Keyboard.keyLabel(i),
               Keyboard.keyFunction(i)
            };
            
            System.out.println(String.format(spec, parms)); 
         }
      }
      
      catch (Exception exp)
      {
         exp.printStackTrace();
      }
   }
}
