/*
** Module   : ServerExports.java
** Abstract : defines all UI-related API's exported by the server for use
**            by the client  
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 2005xxxx   @23594 Created initial version.
** 002 NVS 20060130   @24149 Server exports taking screen buffer, now 
**                           return screen buffer as well. These are: 
**                           trigger(), validate().
** 003 SIY 20060130   @24156 Minor cleanups.
** 004 NVS 20060221   @24622 Added help() export that locates and runs a
**                           preconfigured help application.
** 005 GES 20060226   @24744 Added a query allowing the client to request
**                           the latest screen buffer from the server.
**                           This is useful in cases where the client
**                           initiates the view() of a frame without an
**                           explicit call from the server (e.g. page
**                           headers and footers in redirected output
**                           mode).
** 006 SIY 20060315   @25139 Changed signatures of BROWSE-related methods.
** 007 NVS 20060320   @25158 Simplified validation protocol by using new
**                           class ValidationResult.
** 008 GES 20060808   @28492 Use int instead of long for the browse 
**                           methods.
** 009 GES 20061017   @30452 Added a new option to trigger().
** 010 GES 20061030   @30788 Changed the return value from trigger() to
**                           eliminate the extra call to the client
**                           (to mark an event as "consume") and to make
**                           the trigger processing synchronous. This
**                           also makes it safer for nested triggers.
** 011 SIY 20061105   @30985 Added getCurrentRow() method.
** 012 NVS 20070118   @31889 EventList class is used instead of WaitList.
** 013 NVS 20070328   @32631 help() request from the client should report
**                           the lastKey variable.
** 014 SIY 20070504   @33426 Updated getRowCount().
** 015 SIY 20111101          Updated imports.
** 016 CA  20130822          Added support of appserver async requests: posting and executing of
**                           server-side events on the client side.
** 017 CA  20130908          Added support for socket runtime.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.util.*;

/**
 * Defines all UI-related API's exported by the server for use by the
 * client.
 */
public interface ServerExports
{
   /**
    * Called from the client-side, it executes the server-side event by calling 
    * {@link ServerEvent#invoke(int)}.
    * 
    * @param    id
    *           The id of the server event.
    */
   public void invokeServerEvent(int id);
   
   /**
    * Check if the resource with the given id is still valid to be used in a WAIT-FOR statement.
    * It will return <code>false</code> if:
    * <ul>
    *    <li>there is no resource associated with the given id or the resource is unknown.</li>
    *    <li>the SENSITIVE attribute for the resource is <code>false</false>.</li>
    *    <li>the resource is not valid</li>.
    * </ul>
    * In all other cases, it will return <code>true</code>
    * 
    * @param    id
    *           The resource id.
    * 
    * @return   See above.
    */
   public boolean validResource(int id);
   
   /**
    * Signals a trigger to the application.  If a trigger is registered for
    * this event and source widget combination, then it will be executed.
    * 
    * @param    triggerId
    *           The trigger to be called.
    * @param    eventId
    *           The event which was generated.
    * @param    sourceId
    *           The widget which generated the event.
    * @param    sb
    *           The associated frame's current <code>ScreenBuffer</code>
    *
    * @return   A container including the <code>ScreenBuffer</code> 
    *           representing the latest screen values of all widgets and the
    *           consume flag.
    */
   public TriggerReturnValue trigger(int          triggerId,
                                     int          eventId,
                                     int          sourceId,
                                     ScreenBuffer sb);
   
   /**
    * Signals a validation event to the application.
    * <p>
    * This signal is used to trigger validation execution on the server.
    *
    * @param   sourceId
    *          widget ID of the widget generated the trigger.
    * @param   sb
    *          current <code>ScreenBuffer</code>
    * 
    * @return  <code>ValidationResult</code> representing the validation  
    *          result, error message and screen value changes.
    */
   public ValidationResult validate(int sourceId, ScreenBuffer sb);
   
   /**
    * Locates and executes a preconfigured HELP application.
    *
    * @param   lastKey
    *          The most recent key event code 
    */
   public void help(int lastKey);

   /**
    * Obtains the latest screen buffer associated with the given frame.
    *
    * @param   frameId
    *          The frame for which the screen buffer should be returned.
    *
    * @return  <code>ScreenBuffer</code> representing the current screen 
    *          values of all widgets in the given frame.
    */
   public ScreenBuffer refresh(int frameId);   
   
   //------------------------------------------------------------------------
   // BrowseModel calls
   //------------------------------------------------------------------------
   /**
    * Return information about columns.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * 
    * @return  A meta information about columns.
    */
   public BrowseColumnConfig[] getColumns(int browseId);

   /**
    * Get number of available rows starting from the given one.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   startRow
    *          Initial row to start count rows from. 
    * 
    * @return  Number of available rows.
    */
   public int getRowCount(int browseId, int startRow);
   
   /**
    * Retrieves several subsequent rows from the database.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   startRowIndex
    *          Index of the first row to retrieve.
    * @param   count
    *          Maximum count of rows to return.
    * 
    * @return  Array of rows. Each row is array of cell values.
    */
   public BaseDataType[][] getRows(int browseId,
                                   int startRowIndex,
                                   int count);
   
   /**
    * Updates selected cells in the given row.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   rowIndex
    *          Index of the row to update.
    * @param   cellIndices
    *          Cells for update indexes.
    * @param   cellValues
    *          Cells for update new values.
    * 
    * @throws  DataValidationException
    *          New values don't passed database validation rules.
    */
   public void updateRow(int            browseId,
                         int            rowIndex,
                         int[]          cellIndices,
                         BaseDataType[] cellValues)
   throws DataValidationException;

   /**
    * Returns <code>true</code> if specified row is new and is not present
    * in the data base.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   rowIndex
    *          Index of the row to return information about.
    * @return  <code>true</code> if row is new and <code>false</code>
    *          otherwise.
    */
   public boolean isNewRow(int browseId, int rowIndex);

   /**
    * Inform server about current row in the browse.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   rowIndex
    *          Index of the current row.
    *          
    * @return  New current row.
    */
   public int setCurrentRow(int browseId, int rowIndex);

   /**
    * Get current row in the query.
    *
    * @param   browseId
    *          ID of the browse which should handle the request.
    *          
    * @return  The current row index.
    */
   public int getCurrentRow(int browseId);
   
   /**
    * Called by the P2J client side, it will inform the server that a server-event was generated.
    * Although the event originates from the P2J Client Side, socket events are server-side 
    * events.
    * 
    * @param    resourceId
    *           The resource ID for the socket to which the READ-RESPONSE event belongs.
    *           
    * @return   The ID of the generated server event or <code>-1</code> if a event could not be
    *           posted (i.e. the resource is not SENSITIVE).
    */
   public int readResponseEvent(int resourceId);
   
   /**
    * Called by the P2J client side, it will inform the server that a server event was generated.
    * Although the event originates from the P2J Client Side, socket events are server-side 
    * events.
    * 
    * @param    resourceId
    *           The resource ID for the server socket to which the CONNECT event belongs.
    * @param    clientSocketId
    *           The resource ID of the socket created for the new connected accepted by the
    *           server socket.
    *           
    * @return   The ID of the generated server event or <code>-1</code> if a event could not be
    *           posted (i.e. the resource is not SENSITIVE).
    */
   public int connectEvent(int resourceId, int clientSocketId);
   
   /**
    * Create a socket and associated with the given ID.  The received ID is not a resource ID; 
    * this represents the ID of the associated {@link LowLevelSocketImpl} instance.
    *   
    * @param    id
    *           The ID of the associated {@link LowLevelSocketImpl} instance.
    *           
    * @return   The ID of the created {@link SocketImpl} resource.
    */
   public int createSocket(int id);
}
