/*
** Module   : EventManager.java
** Abstract : manages events and event queues
**
** Copyright (c) 2010-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20101122 Cleaned up imports after switching from KeyEvent 
**                  to KeyInput.
** 004 SIY 20101202 Removed unnecessary handling of mouse events.
** 005 SIY 20110112 Refactoring and simplification.
** 006 GES 20111012 Moved static data to context-local. This allows multiple
**                  client sessions to be hosted in the server.
** 007 CA  20130822 Added queue where to save server-side events.
** 008 CA  20130908          Added API to remove certain server events from the queue.
*/

package com.goldencode.p2j.ui.client.event;

import java.util.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Manages a stack of event queues, allows posting events to the current queue and getting events 
 * from the current queue.  This includes both UI events and events generated by the server side
 * (like the {@link Keyboard#PROCEDURE_COMPLETE} event).
 * <p>
 * Documentation or testing showed no info which would suggest that keys/server events are 
 * processed in a certain way or chained in a certain way. When a server-event is posted, it will 
 * be chained in the event queue, but there is no guarantee where will be chained relative to  
 * other non-server events.  4GL enforces the server-event order only relative to other  
 * server-events and this only in case the events are not generated by truly asynchronous 
 * requests (like in the case of an appserver connected in Session-managed operating mode); 
 * for truly asynchronous requests (like in the case of an appserver connected in Session-free 
 * operating mode), the order of the server-events is undefined.
 * <p>
 * For this reason and for the fact that server-side events can be posted even when a WAIT-FOR 
 * loop is not in effect, the server-events are kept in a global {@link WorkArea#serverEvents 
 * event queue}.
 */
public class EventManager
{
   /** Stores context-local state variables. */
   private static final ContextLocal<WorkArea> work = 
   new ContextLocal<WorkArea>()
   {
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }
   };
   
   /**
    * Hidden constructor, all usage of this class will be static.
    */
   private EventManager()
   {
   }
      
   /**
    * Return next available event or <code>null</code> if queue is empty.
    * 
    * @return  next available event.
    */
   public static Event getEvent()
   {
      WorkArea wa = work.get();
      
      Event evt = null;
      
      synchronized (wa)
      {
         evt = wa.queue.get();
      }
      
      return evt;
   }

   /**
    * Install new event queue.
    */
   public synchronized static void pushEventQueue()
   {
      WorkArea wa = work.get();
      
      synchronized (wa)
      {
         wa.stack.push(wa.queue);
         wa.queue = new EventQueue();
      }
   }

   /**
    * Restore old event queue.
    */
   public synchronized static void popEventQueue()
   {
      WorkArea wa = work.get();
      
      synchronized (wa)
      {
         if (wa.stack.isEmpty())
            throw new RuntimeException("Unbalanced PUSH/POP Event queue!!!");
      
         EventQueue save = wa.queue;
         
         wa.queue = wa.stack.pop();
         
         wa.queue.insert(save);
      }
   }

   /**
    * Convenient shortcut for posting events.
    * 
    * @param   event
    *          Event to post.
    */
   public static void postEvent(Event event)
   {
      WorkArea wa = work.get();
      
      synchronized (wa)
      {
         // do not post events to the main queue because they are just wasted;
         // there is no method in P2J that would process those events, ever,
         // because P2J always brackets events it cares about with push and pop
         if (wa.stack.size() == 0)
         {
            return;
         }
      
         wa.queue.post(event);
      }
   }

   /**
    * Saves the specified server event in the {@link WorkArea#serverEvents} queue, as server-side
    * events can be created if if a UI-blocking statement is not in effect.
    * 
    * @param    event
    *           The server event.
    */
   public static void postServerEvent(ServerEvent event)
   {
      WorkArea wa = work.get();
      
      synchronized (wa.serverEvents)
      {
         wa.serverEvents.post(event);
      }
   }

   /**
    * Remove from the server event queue all the events in the given list, identified by their id.
    * 
    * @param    eventIds
    *           The list of event IDs to be removed.
    */
   public static void removeServerEvents(int[] eventIds)
   {
      WorkArea wa = work.get();
      
      // build a set so we can ease lookup
      Set<Integer> ids = new HashSet<Integer>();
      for (int id : eventIds)
      {
         ids.add(id);
      }
      
      synchronized (wa.serverEvents)
      {
         // this will be a queue with the "clean" events in reverse order
         Deque<Event> cleanQueue = new ArrayDeque<Event>();
         
         // collect all events, excluding the unwanted events
         while (!wa.serverEvents.isEmpty())
         {
            ServerEvent event = (ServerEvent) wa.serverEvents.get();
            
            if (!ids.contains(event.id()))
            {
               cleanQueue.push(event);
            }
         }
         
         // add the clean events back to the queue, in the proper order
         while (!cleanQueue.isEmpty())
         {
            Event event = cleanQueue.pop();

            wa.serverEvents.post(event);
         }
      }
   }
   
   /**
    * Get the next available server event from the {@link WorkArea#serverEvents} queue.
    * 
    * @return   The next available server event or <code>null</code> if the queue is empty.
    */
   public static ServerEvent getServerEvent()
   {
      WorkArea wa = work.get();
      
      ServerEvent evt = null;
      
      synchronized (wa.serverEvents)
      {
         evt = (ServerEvent) wa.serverEvents.get();
      }
      
      return evt;
   }

   /**
    * Check if there are pending server events in the {@link WorkArea#serverEvents} queue.
    * 
    * @return   <code>true</code> if there are pending server events, <code>false</code> otherwise.
    */
   public static boolean hasServerEvents()
   {
      WorkArea wa = work.get();
      
      synchronized (wa.serverEvents)
      {
         return !wa.serverEvents.isEmpty();
      }
   }
   
   /**
    * Convenient shortcut for posting key events.
    * 
    * @param   key
    *          Key to post.
    */
   public static void postKey(int key)
   {
      postEvent(getKeyEvent(key, WindowManager.findFocus(), false));
   }

   /**
    * Convert a key code into a complete event (which includes the event's
    * source widget) by using the current widget in focus.
    *
    * @param    key
    *           The key code of the event which was read from the terminal.
    *
    * @return   The complete event including a valid source widget or
    *           <code>null</code> in the case of any error.
    */
   public static KeyInput eventFromKey(int key)
   {
      if (key == -1)
         return null;

      return getKeyEvent(key, WindowManager.findFocus(), true);
   }
   
   /**
    * Creates an instance of <code>keyEvent</code> for specified key.
    *
    * @param   key
    *          The key code.
    * @param   src
    *          The source widget for this event.
    * @param   isReal
    *          Flag which marks real keys pressed by user.
    *          
    * @return  New event instance generated for specified key and widget.
    */
   private static KeyInput getKeyEvent(int key, Widget src, boolean isReal)
   {
      //TODO: fix character check here, at present it handles only ASCII
      //Perhaps Character.isValidCodePoint(key) + some additional check may
      //help us filter out correct keys.
      int id = (key > 255 || key < ' ') ? EventType.KEY_PRESSED :
                                          EventType.KEY_TYPED;
      
      return new KeyInput(src, id, key, false, isReal);
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Queue stack (saved queues). */
      private Stack<EventQueue> stack = new Stack<EventQueue>();
   
      /** Event queue currently in use. */
      private EventQueue queue = new EventQueue();
      
      /** Queue dedicated for registered events originating from the server-side. */
      private EventQueue serverEvents = new EventQueue();
   }
}
