/*
** Module   : ServerEvent.java
** Abstract : defines an event which originates from the server-side 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130813 Created initial version.
** 002 CA  20130908          Added API to remove certain server events from the queue.
*/

package com.goldencode.p2j.ui.client.event;

import java.io.*;
import java.util.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.*;

/**
 * Instances of this class represent a definition of an event which originates from the server 
 * side. Currently, the following server-side events are supported:
 * <ul>
 *    <li>{@link Keyboard#PROCEDURE_COMPLETE}: raised when an async appserver request is finished
 *    </li> 
 * </ul>
 * 
 * This class also keeps a registry of server-side events posted to the client-side; when the
 * client-side processes the event, the associated {@link ServerEvent} instance will be resolved
 * and the event will be executed, using the registered {@link #task}.
 * <p>
 * In 4GL, is possible to add user-defined triggers or use the server-events in APPLY statements;
 * testing showed that doing so results in abnormal behavior (as the i.e. parameters, procedure 
 * context or the SELF handle can't be resolved). 
 * <p>
 * In current P2J implementation, any server-event names used when defining triggers or in APPLY
 * statements will be treated and will behave as any other UI event; their server-event "flavor" 
 * will be hidden and is exposed only when the {@link ServerEvent} instances are posted to the
 * event queue via {@link LogicalTerminal#postServerEvent}.
 */
public class ServerEvent
extends Event
implements Externalizable
{
   /** A registry of server-side events which were posted to the client. */
   private static ContextLocal<Map<Integer, ServerEvent>> events =
   new ContextLocal<Map<Integer, ServerEvent>>()
   {
      protected Map<Integer, ServerEvent> initialValue()
      {
         return new HashMap<Integer, ServerEvent>();
      };
   };
   
   /** The id of the event. */
   private int id;
   
   /** The resource ID which created the event. */
   private int resourceId;
   
   /** The name of this event. */
   private String name;
   
   /**
    * The associated task which will be executed when the event is processed by the client-side.
    * May be <code>null</code> if the instance is created on the server-side. 
    */
   private Runnable task = null;

   /**
    * Public c'tor added for deserialization purposes only. Not for public use. This will be
    * invoked only on client-side, when a new server event instance is received. In this case, it
    * is OK for the {@link #task} field to be null.
    */
   public ServerEvent()
   {
      super(null, -1);

      // no-op
   }
   
   /**
    * Create a new server-side event.
    * 
    * @param    resourceId
    *           The id of the resource which created the event.
    * @param    name
    *           The name of this event.
    * @param    task
    *           The task associated with this event.
    */
   public ServerEvent(int resourceId, String name, Runnable task)
   {
      super(null, Utils.uniqueId());

      // inherit the overwritten id from the super implementation
      this.id = super.id();
      this.name = name;
      this.task = task;
      this.resourceId = resourceId;
      
      Map<Integer, ServerEvent> evts = events.get();
      synchronized (evts)
      {
         evts.put(id(), this);
      }
   }

   /**
    * Resolve and invoke the event with the given ID.
    * 
    * @param    id
    *           The event ID.
    */
   public static void invoke(int id)
   {
      Map<Integer, ServerEvent> evts = events.get();
      ServerEvent event = null;
      
      synchronized (evts)
      {
         event = events.get().remove(id);
      }
      
      // invoke only if the event was not already purged
      if (event != null)
      {
         event.task.run();
      }
   }
   
   /**
    * Remove from the server event queue all the events in the given set, identified by their id.
    * 
    * @param    eventIds
    *           The set of event IDs to be removed.
    */
   public static void removeServerEvents(Set<Integer> eventIds)
   {
      Map<Integer, ServerEvent> evts = events.get();
      synchronized (evts)
      {
         evts.keySet().removeAll(eventIds);

         LogicalTerminal.removeServerEvents(eventIds);
      }
   }
   
   /**
    * Get the ID of this server-side event.
    * 
    * @return   The {@link #id} of this server-side event.
    */
   @Override
   public int id()
   {
      return id;
   }
   
   /**
    * Get the name of this server-event.
    * 
    * @return   See above.
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Get the resource ID of this server-event.
    * 
    * @return   See above.
    */
   public int resourceId()
   {
      return resourceId;
   }

   /**
    * Send this server-event to the destination. Only the {@link #id}, {@link #resourceId} and
    * {@link #name} fields will be sent.
    * 
    * @param    out
    *           The output destination to which the server event will be sent.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(id);
      out.writeInt(resourceId);
      out.writeUTF(name);
   }

   /**
    * Initialize this server event instance with the data received from the input source.
    * 
    * @param    in
    *           The input source from which server event will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      id = in.readInt();
      resourceId = in.readInt();
      name = in.readUTF();
   };
}
