/* 
** Module   : ConnectHelper.java
** Abstract : Helper methods for parsing the options in a connect-like method.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130908 Created initial version.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;
import java.util.logging.*;

/**
 * This is a helper class to bring to central point the options which can be passed to a connect
 * like methods. The options are passed as a string to {@link #parseOptions}; using the list of
 * valid options, this will be able to identify no-value options (which act like flags) and also
 * options which must receive an int value.
 * <p>
 * The <code>-pf file</code> option can be used to read a set of options from a file, and override
 * any existing ones.
 */
public class ConnectHelper
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(ConnectHelper.class);

   /**
    * Parse and validate the given options; a map is returned with each parameters's value. If a
    * parameter doesn't have a value, the map will contain <code>null</code>.
    *  
    * @param    options
    *           The options from which the parameters will be extracted.
    * @param    knownOptions
    *           Set of valid options.
    * @param    noValueOptions
    *           Set of options which require no argument, i.e. "flags".
    * @param    intOptions
    *           Set of options which requre an int argument.
    *
    * @return   The parameter-to-value map.
    */
   public static Map<String, String> parseOptions(character            options,
                                                  Set<String>          knownOptions,
                                                  Set<String>          noValueOptions,
                                                  Map<String, Integer> intOptions)
   {
      Map<String, String> parms = new HashMap<String, String>();
      String[] opts = options.toStringMessage().split(" ");
   
      String arg = null;
      for (int i = 0; i < opts.length; i++)
      {
         String txt = opts[i];
         if (txt.length() == 0)
            continue;
         
         boolean err = false;
         if (txt.startsWith("-"))
         {
            if (!knownOptions.contains(txt))
            {
               err = true;
            }
            else
            {
               arg = txt;
            }
            
            // some options do not have a corresponding value
            if (noValueOptions.contains(arg))
            {
               parms.put(arg, null);
               arg = null;
            }
         }
         else
         {
            if (arg != null)
            {
               if (intOptions.containsKey(arg))
               {
                  // ensure that int values can be parsed
                  int value = 0;
                  try
                  {
                     value = Integer.parseInt(txt);
                  }
                  catch (NumberFormatException e)
                  {
                     // ignore
                  }
                  int def = intOptions.get(arg);
                  if (value <= def)
                  {
                     final String msg1 = "The %s parameter requires an argument greater than %d";
                     final String msg2 = "Unable to process parameters";
                     ErrorManager.displayError(1404, String.format(msg1, arg, def), false);
                     ErrorManager.displayError(5509, msg2, false);
                     return null;
                  }
               }
   
               parms.put(arg, txt);
            }
            else
            {
               err = true;
            }
            
            arg = null;
         }
         
         if (err)
         {
            final String format = "Could not recognise argument: %s";
            final String msg1 = String.format(format, txt);
            final String msg2 = "Unable to process parameters";
            ErrorManager.displayError(301, msg1);
            ErrorManager.displayError(5509, msg2, false);
            return null;
         }
      }
      
      if (parms.containsKey("-pf"))
      {
         String pf = parms.get("-pf");
         Map<String, String> pfParms = readPfFile(pf);
         if (pfParms == null)
         {
            final String msg = "Unable to open parameter file %s, errno 2";
            ErrorManager.recordOrShowError(1247, String.format(msg, pf), false);
   
            return null;
         }
         
         // the parameters from the file are not validated
         parms.putAll(pfParms);
      }
      
      return parms;
   }

   /**
    * Parse the value for the specified key as an int, or default to the specified value if the
    * key does not exist.
    * 
    * @param    parms
    *           The options map.
    * @param    key
    *           The key to search.
    * @param    def
    *           The default value to return if the key does not exist.
    *           
    * @return   See above.
    */
   public static int getInt(Map<String, String> parms, String key, int def)
   {
      String sval = parms.get(key);
      if (sval == null)
      {
         return def;
      }
   
      int val = 0;
      try
      {
         return Integer.parseInt(sval);
      }
      catch (NumberFormatException e)
      {
         val = 0;
      }
      
      return val;
   }

   /**
    * Get a string from the parameters map, using the specified key and defaulting to the 
    * specified value, if it does not exist.
    * 
    * @param    parms
    *           The options map.
    * @param    key
    *           The key to search.
    * @param    def
    *           The default value to return if the key does not exist.
    *           
    * @return   See above.
    */
   public static String getString(Map<String, String> parms, String key, String def)
   {
      String val = parms.get(key);
      return val == null ? def : val;
   }

   /**
    * Read the parameters configured in the given file.
    * 
    * @param    file
    *           The file containing the parameters.
    *           
    * @return   A map containing the value for each parameter or <code>null</code> if there were
    *           problems reading the file.
    */
   public static Map<String, String> readPfFile(String file)
   {
      if (FileSystemOps.searchPath(file).isUnknown())
      {
         // file does not exist.
         return null;
      }
      
      StreamWrapper stream = new StreamWrapper();
      stream.assign(StreamFactory.openFileStream(file, false, false));
      Map<String, String> result = new HashMap<String, String>();
      try
      {
         String line = stream.readLn();
         while (line != null)
         {
            line = line.trim();
            if (line.startsWith("-"))
            {
               int sharpIdx = line.indexOf('#');
               if (sharpIdx >= 0)
               {
                  // drop trailing comment
                  line = line.substring(0, sharpIdx).trim();
               }
               
               // this is a configuration line
               String param = line;
               int spaceIdx = param.indexOf(' ');
               if (spaceIdx >= 0)
               {
                  param = param.substring(0, spaceIdx);
                  // drop the parameter, to leave only the value (if any);
                  line = line.substring(spaceIdx).trim();
               }
               
               String val = (line.length() == 0 ? null : line);
               
               result.put(param,  val);
            }
         }
      }
      catch (EOFException e)
      {
         // ignore
      }
      catch (IOException e)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Problems reading parameter file " + file, e);
         }

         // read error
         return null;
      }
      catch (InterruptedException e)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Problems reading parameter file " + file, e);
         }

         // other error
         return null;
      }

      return result;
   }
}
