/*
** Module : ErrorManager.java
** Public : provides ERROR-STATUS, NO-ERROR and error message support
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20050823   @21210 Created initial version which provides
**                           ERROR-STATUS and NO-ERROR support.
** 002 GES 20050829   @22362 Minor mods to more closely duplicate how
**                           Progress handles error text. There are
**                           cases where error text is added but no
**                           error condition is raised (even in a
**                           non-silent mode situation).
** 003 ECF 20051104   @23262 Added capability to extract error data from
**                           Java exceptions. This includes both formatted
**                           error texts as well as recording ErrorEntry
**                           objects. Overloaded recordOrThrowError for
**                           this purpose and added a set of overloaded
**                           errorTexts() methods.
** 004 ECF 20051205   @23698 Added recordOrShowError method variants.
**                           Displays an error message in an alert box and
**                           logs the error.
** 005 ECF 20051219   @23752 Added warning mode. When in warning mode,
**                           errors are reported to the UI using an alert
**                           box, rather than an exception being thrown.
** 006 ECF 20060117   @23959 Enabled show error code. Removed temporary
**                           code which used stderr and replaced with
**                           calls to LogicalTerminal.message().
** 007 GES 20060123   @24021 Moved to context-local storage.
** 008 GES 20060208   @24389 Errors are now displayed via message when
**                           not in silent error mode.
** 009 NVS 20060306   @24903 Errors occurring on the client are now passed
**                           to the server for centralized logging.
** 010 NVS 20060313   @24994 Error messages are displayed using Logical
**                           Terminal methods if on the server, and Client
**                           Exports directly otherwise.
** 011 GES 20060313   @24995 Use new logging infrastructure.
** 012 GES 20060626   @27569 Implemented scoped context local data which
**                           makes it safe to use silent error mode in
**                           a nested manner.
** 013 ECF 20060719   @28122 Changed logging level for silent mode errors
**                           from WARNING to FINE.
** 014 GES 20060807   @28454 Match new signature in ClientExports.
** 015 GES 20060927   @29966 Exposed displayError() as public.
** 016 GES 20061004   @30134 Added support for nested errors (errors
**                           generated during processing of another
**                           error).
** 017 NVS 20061017   @30470 Implemented ability to restart the session
**                           without exiting the process.
** 018 ECF 20071217   @36499 Added headless mode. Intended to support
**                           environments with no UI and limited logging.
**                           Implementation at this time is limited.
** 019 CA  20080530   @38463 Added methods related to displaying an error
**                           message before the application stops.
** 020 GES 20080619   @38886 Added a convenience method to format and
**                           display an error without throwing any
**                           exception, w/o being affected by silent
**                           error mode and without recording anything.
** 021 GES 20080624   @38912 Added an ignore flag to NOP all display,
**                           recording or exception generation. Cleaned
**                           up the API to meet coding standards.
** 022 CA  20080815   @39447 Added displayAbend - it displays an abend
**                           message, but no actual abend is triggered.
** 023 SVL 20081208   @40826 Match client exports API change.
** 024 ECF 20081215   @40913 Added throwError() variants. Unconditionally
**                           displays an error message and throws an error
**                           condition exception, regardless of silent
**                           error mode. Headless mode is still honored.
** 025 GES 20090422   @41908 Converted to standard string formatting.
** 026 SVL 20090612   @42669 showErrorAndAbend() can be executed on the client
**                           side.
** 027 ECF 20090622   @42987 Added new display option for errors. New variants
**                           of displayError() and buildErrorText() allow the
**                           standard message prefix to be omitted.
** 028 SIY 20090701   @43012 Made ErrorEntry public for use by other classes.
**                           Inferred generics. Minor cleanups.
** 029 SVL 20090803   @43476 showErrorAndAbend(int[],String[]) has been added.
** 030 GES 20090831   @43802 Removed stacking from the error-status data
**                           members. The 4GL doesn't stack the error list or
**                           the error flag. The previous stacking was an
**                           over-compensation for saving state across top-
**                           level block boundaries. The transaction mgr now
**                           handles saving/restoring the silent error mode
**                           flag in the necessary cases. This class now can
**                           be greatly simplified which also fixes some issues
**                           with the error-status system handle.
** 031 SVL 20090908   @43847 recordOrThrowError() saves error code in the
**                           ErrorConditionException.
** 032 GES 20090924   @44010 Completed headless support. Both the global flag
**                           and a context-specific flag are honored.
** 033 SIY 20110112          Minor refactorings and cleanups.
** 034 CA  20130112          Added asHandle, getType, isValid and isUnknown 
**                           methods, accessed by the ERROR-STATUS handle.
**                           Added new recordOrThrowError APIs.
** 035 CA  20130216          Added support for ERROR-OBJECT-DETAIL attribute.
** 036 CA  20130221          getType getter for TYPE attribute was renamed to getResourceType.
** 037 CS  20130221          Added conversion support for _msg() function
** 038 CA  20130222          Added missing setter for ERROR attribute.
** 039 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 040 VMN 20130411          Added boolean parameter modal to recordOrShowError() method.
** 041 EVK 20130413          Implemented method getErrorNumberAtIndex
** 042 CA  20130612          ErrorConditionException's generated in recordOrThrowError will carry
**                           as cause a NumberedException identifying the error number and the
**                           associated message.
** 043 OM  20130606          Handle for soapFault set to null for the moment. This is a bug that
**                           manifests with pl/java and should be fixed (with lazy initialization ?)
** 044 EVL 20130718          Adding new method to display message and abend. Also it silently
**                           returns if NO-ERROPR option is specified.
** 045 VMN 20130731          Adding asMsg parameter to recordOrThrowError() method.
** 046 CA  20130909          Added "prefix" parameter to recordOrShowError, to allow error 
**                           messages without double-asterix prefix.
*/

package com.goldencode.p2j.util;

import java.util.*;
import java.util.logging.*;
import com.goldencode.p2j.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.proxy.StaticProxy;

/**
 * A context-local class providing static methods to support the 
 * <code>ERROR-STATUS</code> and <code>NO-ERROR</code> Progress features.
 * <p>
 * The following is the mapping of Progress features to Java:
 * <p>
 * <pre>
 * Progress Feature             Type          Java Feature
 * ---------------------------  ------------  ----------------------------
 * error-status:error           attribute     {@link #isError}         
 * error-status:num-messages    attribute     {@link #numErrors}
 * error-status:get-message     method        {@link #getErrorText}
 * error-status:get-number      method        {@link #getErrorNumber}
 * </pre>
 * <p>
 * In order to properly use this class, all processing which must be handled
 * on a <code>NO-ERROR</code> basis must be enclosed within a matching pair
 * of calls to {@link #silentErrorEnable} and {@link #silentErrorDisable}
 * silent error mode respectively.  These calls must be handled from the
 * client application and all called code that honors the
 * <code>NO-ERROR</code> must honor the silent mode.  This can be done via
 * the following methods (in a library):
 * <p>
 * <ol>
 *    <li> If a simple return from a library is easily possible, delegate
 *         processing to this class using {@link #recordOrThrowError}. After
 *         calling this method, return silently from the library.
 *    <li> Do it yourself.
 *    <ul>
 *       <li> Use {@link #isSilentError} to determine if the silent mode is  
 *            enabled.
 *       <li> Use {@link #addError} to add an error to the list if silent
 *            mode is enabled.
 *       <li> Return silently after the error is added to the list.
 *       <li> If silent mode is not enabled, the
 *            {@link ErrorConditionException} can be thrown directly.
 *       <li> This may be useful if a simple silent return is not easy to do
 *            as one can use the {@link ErrorUnwindException} to unwind the
 *            stack to the point in the library where a silent return is
 *            possible.  This exception must be caught in the library code
 *            and will not be caught in the client application.  Use of
 *            this exception should be completely hidden from the client
 *            application.
 *    </ul>
 * </ol>
 * <p>
 * On the server, the silent error mode flag is saved at every top level
 * scope (see {@link TransactionManager#pushScope}) and it is restored at
 * the end of that scope (see {@link TransactionManager#popScope}).
 * This is done so that "nested" NO-ERROR conditions (e.g. during a RUN
 * statement on an internal or external procedure or an assignment statement
 * with a call to a user defined function) are maintained properly.  In
 * particular, the called procedure or function does NOT execute as if in
 * silent mode unless the code there is so bracketed.  However, the one
 * exception to this is that the RETURN ERROR behavior implemented by the
 * <code>TransactionManager</code> is aware of whether or not silent error
 * mode was enabled in the caller.
 * <p>
 * It must be noted that Progress seems to provide a single global instance of
 * ERROR-STATUS handle and all its data. Thus this data can be seen across
 * call/return boundaries and is only modified at very specific points. for
 * example, the error flag is only changed at the end of a NO-ERROR statement.
 * Likewise, the previous list of errors can be read until the first error
 * occurs after a NO-ERROR statement starts. This deferred clearing of the
 * error list is implemented using the pending error flag. When that flag is
 * set, the error list is cleared. If no errors occur in the NO-ERROR statement
 * then the error list is cleared at the same time the error flag is cleared
 * at the end of the NO-ERROR statement.
 */
public class ErrorManager
{
   /** Logger (safe as a JVM-wide value rather than as context-local). */
   private static final Logger LOG = LogHelper.getLogger(ErrorManager.class);
   
   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer(); 
   
   /** Error data access interface between client and server. */
   private static RemoteErrorData da = new ServerDataAccess(); 
   
   /** Client exports access. */
   private static ClientExports client = null; 
   
   /** Headless mode */
   private static boolean headless = false;

   /**
    * Get a the instance for the ERROR-STATUS system handle.  Is obtained using
    * a call to the {@link StaticProxy#obtain(Class, Class[])}, using the
    * {@link CommonErrorStatus} interface and its methods implemented by these
    * classes: {@link ErrorManager}.
    * 
    * @return   See above.
    */
   public static handle asHandle()
   {
      Class[] classes = { ErrorManager.class };
      CommonErrorStatus proxy = StaticProxy.obtain(CommonErrorStatus.class,
                                                   classes);
      return new handle(proxy);
   }   
   
   /**
    * Implementation for the {@link WrappedResource#valid()} API.
    * 
    * @return   Always true.
    */
   public static boolean valid()
   {
      return true;
   }
   
   /**
    * Implementation for the {@link WrappedResource#unknown()} API.
    * 
    * @return   Always false.
    */
   public static boolean unknown()
   {
      return false;
   }
   
   /**
    * API needed to implement read-only attribute assignment (a 4GL 
    * "feature").
    * 
    * @param    attribute
    *           The attribute's name.
    *           
    * @see      handle#readOnlyError(handle, String)
    */
   public static void readOnlyError(String attribute)
   {
      handle.readOnlyError(asHandle(), attribute);
   }
   
   /**
    * Initialize <code>ErrorManager</code> for headless mode.
    */
   public static void setHeadless()
   {
      headless = true;
   }
   
   /**
    * Sets silent mode during CAN-FIND calls. The semaphore keep tracks of the nested CAN-FIND
    * calls. 
    * If the counter is positive <code>ErrorManager</code> will run similar to headless mode 
    * and will silently discard error messages.
    * <p/> 
    * <strong>Note</strong>Always call this method in pairs (one with enter = true before 
    * executing the query and, one with enter = false, immediately after) to assure the 
    * correct management of the headless context flag. 
    * 
    * @param   enter
    *          If <code>true</code> will increment the nested CAN-FIND count.
    */
   public static void setCanFindBracket(boolean enter)
   {
      if (enter)
      {
         // push a CAN-FIND level
         work.obtain().canFindLevel++;
      }
      else
      {
         // pop put CAN-FIND level:
         work.obtain().canFindLevel--;
      }
   }
   
   /**
    * Check whether this context is in a CAN-FIND bracket. In a CAN-FIND bracket all 4GL errors
    * are not printed 
    * 
    * @return  <code>true</code> if currently a CAN-FIND is executed.
    */
   public static boolean isInCanFindBracket()
   {
      return work.obtain().canFindLevel > 0;
   }
   
   /**
    * Initializes the proper error data access on the client.
    * 
    * @param    errorServer
    */
   public static void initClient(RemoteErrorData errorServer)
   {
      da = errorServer;
      client = ThinClient.getInstance();
   }
      
   /**
    * Reverts the error data access on the client.
    */
   public static void resetClient()
   {
      da = new ServerDataAccess();
   }
      
   /**
    * Reports if silent error mode has been enabled.
    *
    * @return   <code>true</code> if silent mode is enabled, otherwise
    *           <code>false</code>.
    */
   public static boolean isSilentError()
   {
      return work.obtain().silent;
   }
   
   /**
    * Reports if warning mode has been enabled.  When in warning mode, error
    * messages are reported to the UI using an alert box.
    *
    * @return   <code>true</code> if warning mode is enabled, otherwise
    *           <code>false</code>.
    */
   public static boolean isWarningMode()
   {
      return work.obtain().warning;
   }
   
   /**
    * Enables silent error mode, clears the current error list but leaves the
    * previous error status alone.  The previous error status is assigned
    * in {@link #silentErrorDisable} based on whether any errors were added
    * in the interim.
    * <p>
    * Each call to this method must be matched with a corresponding call  
    * to {@link #silentErrorDisable}.
    */
   public static void silentErrorEnable()
   {
      // enable silent error processing
      da.setSilent(true);
   }
   
   /**
    * Disables silent error mode, assigns the error status based on whether
    * any errors were added to the error list since the last call to
    * {@link #silentErrorDisable}.
    * <p>
    * Each call to this method must be matched with a corresponding previous  
    * call to {@link #silentErrorEnable}.
    */
   public static void silentErrorDisable()
   {
      // disable further silent error processing
      da.setSilent(false);
      
      // was there an error?
      boolean pending = da.isPending();
      
      // set or reset the error flag based on whether a pending error occurred
      da.setError(pending);
      
      // if there was no error, now is the time to clear the error list; note
      // that if there was a pending error, the first error to be logged has
      // already forced the clearing of the list previously and we must not
      // do it again here
      if (!pending)
         da.clearList();
      
      // clear the pending flag
      da.setPending(false);
   }
   
   /**
    * Enables warning mode, during which errors are non-fatal, but are
    * reported to the UI of the current context via a call to
    * <code>LogicalTerminal.message()</code>.
    * <p>
    * Each call to this method must be matched with a corresponding call  
    * to {@link #warningModeDisable}.
    */
   public static void warningModeEnable()
   {
      // enable warning processing
      work.obtain().warning = true;
   }
   
   /**
    * Disables warning mode.
    * <p>
    * Each call to this method must be matched with a corresponding, previous
    * call to {@link #warningModeEnable}.
    */
   public static void warningModeDisable()
   {
      // disable further warning processing
      work.obtain().warning = false;
   }
   
   /**
    * Manufacture error text based on the given parameters, if silent error mode is enabled,
    * then this error number and the generated error text will be stored in the error list,
    * otherwise this generated error text will be used to display modal or modeless box in
    * the UI.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    modal
    *           <code>true</code> if message will be shown in modal dialog.
    */
   public static void recordOrShowError(int num, String text, boolean modal)
   {
      recordOrShowError(new int[] { num }, new String[] { text }, modal, true);
   }
   
   /**
    * Manufacture error text based on the given parameters, if silent error mode is enabled,
    * then this error number and the generated error text will be stored in the error list,
    * otherwise this generated error text will be used to display modal or modeless box in
    * the UI.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    modal
    *           <code>true</code> if message will be shown in modal dialog.
    */
   public static void recordOrShowError(int[] num, String[] text, boolean modal)
   {
      recordOrShowError(num, text, modal, true);
   }
    
   /**
    * Manufacture error text based on the given parameters, if silent error mode is enabled,
    * then this error number and the generated error text will be stored in the error list,
    * otherwise this generated error text will be used to display modal or modeless box in
    * the UI.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    modal
    *           <code>true</code> if message will be shown in modal dialog.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the formatted error 
    *           message, <code>false</code> to omit this text.
    */
   public static void recordOrShowError(int num, String text, boolean modal, boolean prefix)
   {
      recordOrShowError(new int[] { num }, new String[] { text }, modal, prefix);
   }

   /**
    * Manufacture one or more error texts based on the given parameters, if silent error mode is
    * enabled, then these error numbers and the respective generated error text messages will be
    * stored in the error list, otherwise this generated error texts will be used to display
    * modal or modeless box in the UI.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    modal
    *           <code>true</code> if message will be shown in modal dialog.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the formatted error 
    *           message, <code>false</code> to omit this text.
    */
   public static void recordOrShowError(int[] num, String[] text, boolean modal, boolean prefix)
   {
      if (da.isIgnore())
         return;
      
      if (num.length != text.length)
      {
         throw new IllegalArgumentException(
            "Unbalanced error number and text message pairs");
      }
      
      StringBuffer buf = new StringBuffer();
      String sep = System.getProperty("line.separator");
      boolean silent = isSilentError();
      
      for (int i = 0; i < num.length; i++)
      {
         String errmsg = buildErrorText(num[i], text[i], prefix);
         
         if (i > 0)
         {
            buf.append(sep);
         }
         buf.append(errmsg);
         
         if (silent)
         {
            if (i == 0)
            {
               // this may reset the error list so it must be done BEFORE we
               // add the new error(s) to the list; this only needs to be done
               // the first time through this loop
               da.setPending(true);
            }
            
            da.addRecord(new ErrorEntry(num[i], errmsg));
         }
      }
      
      String masterMsg = buf.toString();
      
      log(masterMsg, silent);
      
      if (!silent && !isHeadless())
      {
         if (modal)
         {
            if (client != null)
            {
               client.messageBox(new Object[] {masterMsg},
                                 null,
                                 LogicalTerminal.ALERT_ERROR,
                                 LogicalTerminal.BTN_OK,
                                 "Error",
                                 (new ColorSpec(ColorSpec.COLOR_NORMAL)).convert()
                                 );
            }
            else
            {
               LogicalTerminal.messageBox(masterMsg,
                                          LogicalTerminal.ALERT_ERROR,
                                          LogicalTerminal.BTN_OK,
                                          "Error");
            }
         }
         else
         {
            LogicalTerminal.message(masterMsg);
         }
      }
   }

   /**
    * Manufacture error text based on the given parameters; after displaying
    * the error text on the client, a pause is forced before stopping the
    * application.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * 
    * @throws   StopConditionException
    *           always, unless the error manager is in ignore mode.
    */
   public static void showErrorAndAbend(int num, String text)
   throws StopConditionException
   {
      if (da.isIgnore())
         return;
      
      String errmsg = buildAbendText(num, text);
      try
      {
         if (!isHeadless())
         {
            // display the error message (because of pause processing, an
            // endkey, error... even stop condition can be generated here)
            displayError(errmsg);
   
            if (client != null)
            {                                  
               client.pause(-1, null, false);
            }
            else
            {
               LogicalTerminal.pause();
            }
         }
      }
      finally
      {
         postprocessOnAbend(errmsg);
      }
   }

   /**
    * Manufacture error text based on the given parameters; after displaying the error text on
    * the client, the application is stopping.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * 
    * @throws   StopConditionException
    *           always, unless the error manager is in ignore mode.
    */
   public static void conditionalShowErrorAndAbend(int num, String text)
   throws StopConditionException
   {
      // If the NO-ERROR option was specified this method just silently retuns doing nothing
      if (isSilentError())
      {
         return;
      }
      
      if (da.isIgnore())
      {
         return;
      }
      
      String errmsg = buildAbendText(num, text);
      try
      {
         if (!isHeadless())
         {
            // display the error message (because of pause processing, an
            // endkey, error... even stop condition can be generated here)
            displayError(errmsg);
         }
      }
      finally
      {
         postprocessOnAbend(errmsg);
      }
   }

   /**
    * Perform postprocessing for {@link #showErrorAndAbend(int, String)}.
    *
    * @param   errmsg
    *          Message text.
    */
   private static void postprocessOnAbend(String errmsg)
   {
      throw new StopConditionException(errmsg);
   }

   /**
    * Manufacture one or more error texts based on the given parameters, if
    * silent error mode is enabled, then these error numbers and the
    * respective generated error text messages will be stored in the error
    * list, otherwise this generated error texts will be used to display a
    * modal, alert (error) box in the UI.<p>
    * After error messages has been processed, a STOP condition is raised.
    *
    * @param    num
    *           The error numbers.
    * @param    text
    *           The text of the error messages.
    *
    * @throws   StopConditionException
    *           always, unless the error manager is in ignore mode.
    */
   public static void showErrorAndAbend(int[] num, String[] text)
   throws StopConditionException
   {
      if (da.isIgnore())
         return;

      try
      {
         recordOrShowError(num, text, true);
      }
      finally
      {
         postprocessOnAbend(text);
      }
   }

   /**
    * Perform postprocessing for {@link #showErrorAndAbend(int[], String[])}.
    *
    * @param   text
    *          Message text.
    */
   private static void postprocessOnAbend(String[] text)
   {
      StringBuilder sb = new StringBuilder();
      String sep = System.getProperty("line.separator");

      for (int i = 0; i < text.length; i++)
      {
         sb.append(text);

         if (i != text.length - 1)
         {
            sb.append(sep);
         }
      }

      throw new StopConditionException(sb.toString());
   }

   /**
    * Manufacture error text based on the given parameters, if silent error
    * mode is enabled, then this error number and the generated error text
    * will be stored in the error list, otherwise this generated error text
    * will be used to throw a new instance of {@link ErrorConditionException}.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    *           
    * @throws   ErrorConditionException 
    */
   public static void recordOrThrowError(int num, String text)
   throws ErrorConditionException
   {
      recordOrThrowError(num, text, true);
   }
   
   /**
    * Manufacture error text based on the given parameters, if silent error
    * mode is enabled, then this error number and the generated error text
    * will be stored in the error list, otherwise this generated error text
    * will be used to throw a new instance of {@link ErrorConditionException}.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    *           
    * @throws   ErrorConditionException 
    */
   public static void recordOrThrowError(int num, String text, boolean prefix)
   {
      recordOrThrowError(new int[] { num }, new String[] { text }, prefix, false);
   }

   /**
    * Manufacture error text based on the given parameters, if silent error
    * mode is enabled, then this error number and the generated error text
    * will be stored in the error list, otherwise this generated error text
    * will be used to throw a new instance of {@link ErrorConditionException}.
    *
    * @param    nums
    *           The error numbers.
    * @param    texts
    *           The text of the error messages.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    *
    * @throws   ErrorConditionException
    */
   public static void recordOrThrowError(int[] nums,
                                         String[] texts,
                                         boolean prefix)
   throws ErrorConditionException
   {
      recordOrThrowError(nums, texts, prefix, false);
   }

   /**
    * Manufacture error text based on the given parameters, if silent error
    * mode is enabled, then this error number and the generated error text
    * will be stored in the error list, otherwise this generated error text
    * will be used to throw a new instance of {@link ErrorConditionException}.
    *
    * @param    nums
    *           The error numbers.
    * @param    texts
    *           The text of the error messages.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    * @param    asMsg
    *           <code>true</code> to use the given text as the message text else
    *           <code>false</code> to use buildErrorText.
    *
    * @throws   ErrorConditionException 
    */
   public static void recordOrThrowError(int[] nums,
                                         String[] texts, 
                                         boolean prefix,
                                         boolean asMsg)
   throws ErrorConditionException
   {
      if (da.isIgnore())
         return;
      
      boolean silent = isSilentError();

      for (int i = 0; i < nums.length; i++)
      {
         int num = nums[i];
         String text = texts[i];
         boolean last = i == nums.length - 1;

         String errmsg = asMsg ? text : buildErrorText(num, text, prefix);
         
         log(errmsg, silent);
         
         if (silent)
         {
            // this may reset the error list so it must be done BEFORE we add
            // the new error to the list
            da.setPending(true);
            
            da.addRecord(new ErrorEntry(num, errmsg));
         }
         else if (isHeadless())
         {
            if (last)
            {
               throw new ErrorConditionException(num, errmsg, new NumberedException(text, num));
            }
         }
         else
         {
            try
            {
               // display the error message (because of pause processing, an
               // endkey, error... even stop condition can be generated here)
               displayError(errmsg);
            }
            
            catch (ConditionException err)
            {
               if (client == null && TransactionManager.isNestedMode())
               {
                  // honor nested mode here
                  TransactionManager.saveNestedError(err);
               }
               else
               {
                  throw err;
               }
            }
            
            if (last && !isWarningMode())
            {
               // raise the error condition
               throw new ErrorConditionException(num, errmsg, new NumberedException(text, num));
            }
         }
      }
   }

   /**
    * Manufacture error text based on the given text, if silent error
    * mode is enabled, then this error number and the generated error text
    * will be stored in the error list, otherwise this error text
    * will be used to throw a new instance of {@link ErrorConditionException}.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    * @param    asMsg
    *           <code>true</code> to use the given text as the message text else
    *           <code>false</code> to use buildErrorText.
    *
    * @throws   ErrorConditionException
    */
   public static void recordOrThrowError(int num, String text, boolean prefix, boolean asMsg)
   {
      recordOrThrowError(new int[] { num }, new String[] { text }, prefix, asMsg);
   }

   /**
    * Record a chain of errors or throw an exception based upon the given
    * exception (which of these two is done depends upon the state of the
    * silent mode feature).  If the former, the error messages and numbers
    * are extracted from <code>error</code> and any other instances of
    * <code>NumberedException</code> present in the chain of root cause
    * exceptions.  The first error entry recorded is derived from the given
    * error number and text, however.
    *
    * @param   num
    *          The error number.
    * @param   text
    *          The text of the error message.
    * @param   error
    *          Exception containing an error number, message, and optionally
    *          a root cause chain.
    */
   public static void recordOrThrowError(int num,
                                         String text,
                                         NumberedException error)
   {
      recordOrThrowError(new NumberedException(text, num, error));
   }
   
   /**
    * Record a chain of errors or throw an exception based upon the given
    * exception (which of these two is done depends upon the state of the
    * silent mode feature).  If the former, the error messages and numbers
    * are extracted from <code>error</code> and any other instances of
    * <code>NumberedException</code> present in the chain of root cause
    * exceptions.
    *
    * @param   error
    *          Exception containing an error number, message, and optionally
    *          a root cause chain.
    */
   public static void recordOrThrowError(NumberedException error)
   {
      if (da.isIgnore())
         return;
      
      boolean silent = isSilentError();
      
      log("stack trace follows", error, silent);
      
      if (silent)
      {
         // this may reset the error list so it must be done BEFORE we add
         // the new error(s) to the list
         da.setPending(true);
         
         Iterator<ErrorEntry> iter = compileErrorEntries(error).iterator();
         
         while (iter.hasNext())
         {
            da.addRecord(iter.next());
         }
      }
      else
      {
         throwError(error);
      }
   }
   
   /**
    * Display error text and throw an exception based upon the given
    * exception.  The error messages and numbers are extracted from
    * <code>error</code> and any other instances of
    * <code>NumberedException</code> present in the chain of root cause
    * exceptions.  The first error entry recorded is derived from the given
    * error number and text, however.
    *
    * @param   num
    *          The error number.
    * @param   text
    *          The text of the error message.
    * @param   error
    *          Exception containing an error number, message, and optionally
    *          a root cause chain.
    */
   public static void throwError(int num,
                                 String text,
                                 NumberedException error)
   {
      throwError(new NumberedException(text, num, error));
   }
   
   /**
    * Display error text and throw an exception based upon the given
    * exception.  The error messages and numbers are extracted from
    * <code>error</code> and any other instances of
    * <code>NumberedException</code> present in the chain of root cause
    * exceptions.
    *
    * @param   error
    *          Exception containing an error number, message, and optionally
    *          a root cause chain.
    */
   public static void throwError(NumberedException error)
   {
      if (!isHeadless())
      {
         // display the messages
         Iterator<String> iter = errorTexts(error);
         while (iter.hasNext())
         {
            try
            {
               // because of possible pause processing, an endkey, error...
               // even stop condition can be generated here
               displayError(iter.next());
            }
            
            catch (ConditionException err)
            {
               if (client == null && TransactionManager.isNestedMode())
               {
                  // honor nested mode here
                  TransactionManager.saveNestedError(err);
               }
               else
               {
                  throw err;
               }
            }
         }
      }
      
      if (!isWarningMode())
      {
         // raise the error condition
         throw new ErrorConditionException(error);
      }
   }
   
   /**
    * Manufactures Progress compatible formatted error text based on the
    * given error data.  The given data is substituted into the following
    * string "** &#37;s. (&#37;i)".
    *
    * @param    num
    *           The error number substituted for &#37;i in the format string.
    * @param    text
    *           The text of the error message substituted for &#37;s in the
    *           format string.
    *
    * @return   The formatted text.
    */
   public static String buildErrorText(int num, String text)
   {
      return buildErrorText(num, text, true);
   }
   
   /**
    * Manufactures Progress compatible formatted error text based on the
    * given error data.  The given data is substituted into the following
    * string "** &#37;s. (&#37;i)".
    *
    * @param    num
    *           The error number substituted for &#37;i in the format string.
    * @param    text
    *           The text of the error message substituted for &#37;s in the
    *           format string.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    *
    * @return   The formatted text.
    */
   public static String buildErrorText(int num, String text, boolean prefix)
   {
      String spec = (prefix ? "** %s. (%d)" : "%s. (%d)");
      
      return String.format(spec, text, num);      
   }
   
   /**
    * Manufactures Progress compatible formatted abend error text based on the
    * given error data.  The given data is substituted into the following
    * string "&#37;s. (&#37;i)".
    *
    * @param    num
    *           The error number substituted for &#37;i in the format string.
    * @param    text
    *           The text of the error message substituted for &#37;s in the
    *           format string.
    *
    * @return   The formatted text.
    */
   public static String buildAbendText(int num, String text)
   {
      return String.format("%s. (%d)", text, num);
   }

   /**
    * Adds an error entry to the error list.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    errFlag
    *           If <code>false</code> this should add the number and text to
    *           the error list but not set the error flag.  If
    *           <code>true</code> the error flag should also be set.
    */
   public static void addError(int num, String text, boolean errFlag)
   {
      if (errFlag)
      {
         // this may reset the error list so it must be done BEFORE we add
         // the new error to the list
         da.setPending(true);
      }
      
      String errmsg = buildErrorText(num, text);
      
      // manually using the work area is OK in this case since this method
      // is only used on the server
      work.obtain().errList.add(new ErrorEntry(num, errmsg));
   }
   
   /**
    * Returns the state of the current error flag which denotes if any error
    * has occurred between the current bracketed calls to the methods
    * {@link #silentErrorEnable} and {@link #silentErrorDisable}. If silent
    * error mode is not active, then this method always will return
    * <code>false</code>.
    *
    * @return    State of the current error flag.
    */
   public static boolean isPendingError()
   {
      return da.isSilent() ? da.isPending() : false;
   }
   
   /**
    * Returns the state of the error flag which was set at the last call to
    * {@link #silentErrorDisable} or if {@link #silentErrorEnable} has never
    * been called, this will return <code>false</code>.
    *
    * @return    Current state of the error flag.
    */
   public static logical isError()
   {
      return new logical(work.obtain().error);
   }
   
   /**
    * Returns the number of the errors (recorded in the error list) which
    * have occurred since the last call to {@link #silentErrorEnable}.
    *
    * @return    Current number of errors in the error list if silent mode
    *            has been enabled. This will be 0 if no errors have occurred  
    *            since the last <code>silentErrorEnable</code> call OR if  
    *            <code>silentErrorEnable</code> has never been called.
    */
   public static integer numErrors()
   {
      return new integer(work.obtain().errList.size());
   }
   
   /**
    * Returns the Progress compatible error number associated with the error
    * entry at the given index in the error list.  If the list is empty or
    * if the index is invalid, 0 is returned.
    *
    * @param    index
    *           Position in the error list from which to return the entry.
    *           This is 1-based list index, where 1 is the oldest error number
    *           and highest number is the most recent error number.
    *
    * @return   The error number at that index or 0 if the list is empty or
    *           if the index is invalid.
    */
   public static integer getErrorNumber(integer index)
   {
      if (index.isUnknown())
      {
         return new integer();
      }
      
      return getErrorNumber(index.intValue());
   }
   
   /**
    * Returns the Progress compatible error number associated with the error
    * entry at the given index in the error list.  If the list is empty or
    * if the index is invalid, 0 is returned.
    *
    * @param    index
    *           Position in the error list from which to return the entry.
    *           This is 1-based list index, where 1 is the oldest error number
    *           and highest number is the most recent error number.
    *
    * @return   The error number at that index or 0 if the list is empty or
    *           if the index is invalid.
    */
   public static integer getErrorNumber(int index)
   {
      int result = 0;
      
      if (index > 0 && index <= da.getSize())
      {
         ErrorEntry ee = (ErrorEntry) da.getRecord(--index);
         
         result = ee.num;
      }
      
      return new integer(result);
   }
   
   /**
    * Java implementation for the _msg() function.
    * <p>
    * This will return the error number found on the error stack at the given position, 
    * or 0 if there isn't any error in the stack at the given position.
    * <p>
    * This implementation always will return 0 for any non-positive number (0 or negative). 
    * <p>
    * In the 4GL, some negative numbers will return what appears to be a random number. 
    * This is most likely a flaw in the 4GL implementation where there is improper boundary 
    * testing and an invalid index value is used to dereference memory that is not associated 
    * with the error stack.
    * <p> 
    * As such the behavior is non-deterministic and also is a likely security flaw.
    * <p> 
    * This implementation is deliberately not duplicating this bad behavior and it seems unlikely
    * that an application would be coded to rely upon it.
    * 
    * @param  index
    *         The position of the stack from which we want to get the value error number.
    *         This should be greater than 0 for valid results.
    *         This is 1-based stack offset, where 1 is the most recent error number
    *         and highest number is oldest error number in the stack.
    * 
    * @return The error number found at the given position in the error stack, or 0 if there
    *         isn't any error number at the specified position
    */
   public static integer getErrorNumberAtIndex(long index)
   {
      return getErrorNumberAtIndex(new int64(index));
   }
   
   /**
    * Java implementation for the _msg() function.
    * <p>
    * This will return the error number found on the error stack at the given position, 
    * or 0 if there isn't any error in the stack at the given position.
    * <p>
    * This implementation always will return 0 for any non-positive number (0 or negative). 
    * <p>
    * In the 4GL, some negative numbers will return what appears to be a random number. 
    * This is most likely a flaw in the 4GL implementation where there is improper boundary 
    * testing and an invalid index value is used to dereference memory that is not associated 
    * with the error stack.
    * <p> 
    * As such the behavior is non-deterministic and also is a likely security flaw.
    * <p> 
    * This implementation is deliberately not duplicating this bad behavior and it seems unlikely
    * that an application would be coded to rely upon it.
    * 
    * @param  index
    *         The position of the stack from which we want to get the value error number.
    *         This should be greater than 0 for valid results.
    *         This is 1-based stack offset, where 1 is the most recent error number
    *         and highest number is oldest error number in the stack.
    * 
    * @return The error number found at the given position in the error stack, or 0 if there
    *         isn't any error number at the specified position
    */
   public static integer getErrorNumberAtIndex(integer index)
   {
      return getErrorNumberAtIndex(new int64(index));
   }
   
   /**
    * Java implementation for the _msg() function.
    * <p>
    * This will return the error number found on the error stack at the given position, 
    * or 0 if there isn't any error in the stack at the given position.
    * <p>
    * This implementation always will return 0 for any non-positive number (0 or negative). 
    * <p>
    * In the 4GL, some negative numbers will return what appears to be a random number. 
    * This is most likely a flaw in the 4GL implementation where there is improper boundary 
    * testing and an invalid index value is used to dereference memory that is not associated 
    * with the error stack.
    * <p> 
    * As such the behavior is non-deterministic and also is a likely security flaw.
    * <p> 
    * This implementation is deliberately not duplicating this bad behavior and it seems unlikely
    * that an application would be coded to rely upon it.
    * 
    * @param  index
    *         The position of the stack from which we want to get the value error number.
    *         This should be greater than 0 for valid results.
    *         This is 1-based stack offset, where 1 is the most recent error number
    *         and highest number is oldest error number in the stack.
    * 
    * @return The error number found at the given position in the error stack, or 0 if there
    *         isn't any error number at the specified position
    */
   public static integer getErrorNumberAtIndex(int64 index)
   {
      final integer result;

      if (index.longValue() > 0 && index.longValue() <= Integer.MAX_VALUE)
      {
         result = getErrorNumber(da.getSize() - index.intValue() + 1);
      }
      else
      {
         result = new integer(0);
      }

      return result;
   }
   
   /**
    * Returns the Progress compatible error text associated with the error
    * entry at the given index in the error list.  If the list is empty or
    * if the index is invalid, an empty string "" is returned.
    *
    * @param    index
    *           Position in the error list from which to return the entry.
    *
    * @return   The error text at that index or the empty string "" if the
    *           list is empty or if the index is invalid.
    */
   public static character getErrorText(integer index)
   {
      if (index.isUnknown())
      {
         return new character();
      }
      
      return getErrorText(index.intValue());
   }
   
   /**
    * Returns the Progress compatible error text associated with the error
    * entry at the given index in the error list.  If the list is empty or
    * if the index is invalid, an empty string "" is returned.
    *
    * @param    index
    *           Position in the error list from which to return the entry.
    *
    * @return   The error text at that index or the empty string "" if the
    *           list is empty or if the index is invalid.
    */
   public static character getErrorText(int index)
   {
      String result = "";
      
      if (index > 0 && index <= da.getSize())
      {
         ErrorEntry ee = (ErrorEntry) da.getRecord(--index);
         
         result = ee.text;
      }
      
      return new character(result);
   }   
   
   /**
    * Queries the state of <code>silent</code> variable.
    *
    * @return   <code>true</code> if silent mode is enabled, otherwise
    *           <code>false</code>.
    */
   public static boolean isSilent()
   {
      return da.isSilent();
   }

   /**
    * Sets the <code>silent</code> variable.
    *
    * @param    value
    *           The new state of the variable.
    */
   public static void setSilent(boolean value)
   {
      da.setSilent(value);
   }

   /**
    * Queries the state of <code>ignore</code> variable.
    *
    * @return   <code>true</code> if ignore mode is enabled, otherwise
    *           <code>false</code>.
    */
   public static boolean isIgnore()
   {
      return da.isIgnore();
   }

   /**
    * Sets the <code>ignore</code> variable.
    *
    * @param    value
    *           The new state of the variable.
    */
   public static void setIgnore(boolean value)
   {
      da.setIgnore(value);
   }

   /**
    * Queries the state of <code>error</code> variable.
    *
    * @return   <code>true</code> if error flag is turned on, otherwise
    *           <code>false</code>.
    */
   public static boolean isErrorFlag()
   {
      return da.isError();
   }
   
   /**
    * Sets the <code>error</code> variable.
    *
    * @param    value
    *           The new state of the variable.
    */
   public static void setError(boolean value)    
   {
      da.setError(value);
   }
   
   /**
    * Sets the <code>error</code> variable.
    *
    * @param    value
    *           The new state of the variable.
    */
   public static void setError(logical value)    
   {
      setError(value.booleanValue());
   }

   /**
    * Queries the state of <code>pendingError</code> variable.
    *
    * @return   <code>true</code> if pending error flag is turned on, otherwise
    *           <code>false</code>.
    */
   public static boolean isPending()    
   {
      return da.isPending();
   }

   /**
    * Sets the <code>pendingError</code> variable.
    *
    * @param    value
    *           The new state of the variable.
    */
   public static void setPending(boolean value)    
   {
      da.setPending(value);
   }

   /**
    * Determines if headless mode should be supported. In headless mode, no
    * client-side output will occur.
    *
    * @return   <code>true</code> if the headless flag is turned on, otherwise <code>false</code>.
    */
   public static boolean isHeadless()
   {
      return headless || da.isHeadless() || isInCanFindBracket();
   }
   
   /**
    * Clears the list of error records.
    */
   public static void clearList()    
   {
      da.clearList();
   }

   /**
    * Gets the size of the list of error records.
    *
    * @return   The size of the list.
    */
   public static int getSize()    
   {
      return da.getSize();
   }

   /**
    * Adds a record to the list of error records.
    *
    * @param    record
    *           The object to add to the list of error records.
    */
   public static void addRecord(Object record)    
   {
      da.addRecord(record);
   }

   /**
    * Gets a record from the list of error records.
    *
    * @param    index
    *           The record index.
    *
    * @return   The object associated with the given index in the list.
    */
   public static Object getRecord(int index)    
   {
      return da.getRecord(index);
   }

   /**
    * Format the text of the error message and then display that text on
    * on the terminal.
    *
    * @param    num
    *           Error number.
    * @param    msg
    *           The text to display in the message line.
    */
   public static void displayError(int num, String msg)
   {
      displayError(num, msg, true);
   }
   
   /**
    * Format the text of the error message and then display that text on
    * on the terminal.
    *
    * @param    num
    *           Error number.
    * @param    msg
    *           The text to display in the message line.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    */
   public static void displayError(int num, String msg, boolean prefix)
   {
      displayError(buildErrorText(num, msg, prefix));
   }
   
   /**
    * Format the text of the abend message and then display that text on
    * on the terminal.
    *
    * @param    num
    *           Error number.
    * @param    msg
    *           The text to display in the message line.
    */
   public static void displayAbend(int num, String msg)
   {
      displayError(buildAbendText(num, msg));
   }

   /**
    * Display the given error message on the terminal.
    *
    * @param    errmsg
    *           The text to display in the message line.
    */
   public static void displayError(String errmsg)
   {
      if (da.isIgnore() || isHeadless())
         return;
      
      if (client != null)
      {
         // we are running on the client
         ColorSpec cs = new ColorSpec(ColorSpec.COLOR_MESSAGES);
         
         client.message(errmsg, false, null, false, null, cs.convert());
      }
      else
      {
         // we are running on the server
         LogicalTerminal.message(errmsg);
      }
   }
   
   /**
    * Get the type of its associated handle.
    * 
    * @return   Always return the PSEUDO-WIDGET value.
    */
   public static character getResourceType()
   {
      return new character("PSEUDO-WIDGET");
   }
   
   /**
    * Returns a handle for the ERROR-OBJECT-DETAIL attribute, set in cases of SOAP-related faults.
    * 
    * @return   See above.
    */
   public static handle getErrorObjectDetail()
   {
      return new handle(work.get().soapFault);
   }
   
   /**
    * Set the ERROR-OBJECT-DETAIL attribute, a handle of type SOAP-FAULT.
    * TODO: confirm that the type of this handle must be SOAP-FAULT.
    * 
    * @param    h
    *           The handle representing the soap-fault
    */
   public static void setErrorObjectDetail(handle h)
   {
      if (!(h.get() instanceof SOAPFault))
      {
         final String msg = 
            "Incorrect handle set for ERROR-OBJECT-DETAIL - expecting SOAP-FAULT, got %s";
         String type = handle.unwrap(h).getResourceType().toStringMessage();

         throw new RuntimeException(String.format(msg, type));
      }

      work.get().soapFault.assign(h);
   }

   /**
    * Get an iterator on Progress-style error text strings, derived from the
    * error information stored in <code>error</code> (including its possible
    * chain of root causes).
    *
    * @param   error
    *          A throwable which is an instance of
    *          <code>NumberedException</code> or which contains instances
    *          thereof.
    *
    * @return  Iterator on error text strings.
    */
   static Iterator<String> errorTexts(Throwable error)
   {
      return errorTexts(compileErrorEntries(error));
   }
   
   /**
    * Get an iterator on Progress-style error text strings, derived from the
    * error information stored in <code>error</code> (including its possible
    * chain of root causes), and from the given error number and text.
    *
    * @param   num
    *          Error number of the first message to occur in the resulting
    *          iteration.
    * @param   text
    *          Error text of the first message to occur in the resulting
    *          iteration.
    * @param   error
    *          A throwable which is an instance of
    *          <code>NumberedException</code> or which contains instances
    *          thereof.
    *
    * @return  Iterator on error text strings.
    */
   static Iterator<String> errorTexts(int num, String text, Throwable error)
   {
      List<ErrorEntry> list = compileErrorEntries(error);
      ErrorEntry entry = new ErrorEntry(num, text);
      list.add(entry);
      
      return errorTexts(list);
   }
   
   /**
    * Get an iterator on Progress-style error text strings, based upon the
    * <code>ErrorEntry</code> objects stored in a list.
    *
    * @param   entries
    *          List of error entries.
    *
    * @return  Iterator on error text strings.
    */
   private static Iterator<String> errorTexts(final List<ErrorEntry> entries)
   {
      return new Iterator<String>()
      {
         private final Iterator<ErrorEntry> iter = entries.iterator();

         @Override
         public boolean hasNext()
         {
            return iter.hasNext();
         }

         @Override
         public String next()
         {
            ErrorEntry entry = iter.next();
            
            return buildErrorText(entry.num, entry.text);
         }

         @Override
         public void remove()
         {
            throw new UnsupportedOperationException();
         }
      };
   }
   
   /**
    * Extract from <code>error</code> and any chained root cause exceptions
    * of type {@link com.goldencode.p2j.NumberedException NumberedException}
    * error messages and numbers with which to compile a list of error
    * entries.  This list is compiled in reverse order of the root cause
    * chain traversal, such that the most specific errors appear earliest in
    * the list.  If no instances of <code>NumberedException</code> are
    * detected during the traversal, the list will be empty.
    *
    * @param   error
    *          Error from which to extract error messages and numbers.
    *
    * @return  List of <code>ErrorEntry</code> objects.
    */
   private static List<ErrorEntry> compileErrorEntries(Throwable error)
   {
      List<ErrorEntry> list = new LinkedList<ErrorEntry>();
      
      Throwable next = error;
      do
      {
         if (next instanceof NumberedException)
         {
            NumberedException exc = (NumberedException) next;
            int num = exc.getNumber();
            String msg = exc.getMessage();
            ErrorEntry entry = new ErrorEntry(num, msg);
            list.add(0, entry);
         }
         next = next.getCause();
      }
      while (next != null);
      
      return list;
   }
   
   /**
    * Write a formatted log entry with the error text.  In silent error mode
    * generate a warning and in normal mode generate a severe entry. 
    *
    * @param    message
    *           The text to insert into the log.
    * @param    silent
    *           <code>true</code> if silent error mode is active.
    */
   private static void log(String message, boolean silent)
   {
      log(message, null, silent);
   }
   
   /**
    * Write a formatted log entry with the error text.  In silent error mode
    * generate a warning and in normal mode generate a severe entry. 
    *
    * @param    message
    *           The text to insert into the log.
    * @param    trw
    *           The cause of the failure.
    * @param    silent
    *           <code>true</code> if silent error mode is active.
    */
   private static void log(String message, Throwable trw, boolean silent)
   {
      if (isHeadless())
      {
         return;
      }
      
      Level lvl = silent ? Level.FINE : Level.SEVERE;
      
      if (LOG.isLoggable(lvl))
      {
         String err = LogHelper.generate("%s%s", 
                                         (silent ? "(silent mode) " : ""),
                                         message);
         if (trw == null)
         {
            LOG.logp(lvl, "ErrorManager", "", err);
         }
         else
         {
            LOG.logp(lvl, "ErrorManager", "", err, trw);
         }
      }
   }
   
   /**
    * Stores the data related to a single error.
    */
   public static class ErrorEntry
   {
      /** Error number. */
      public final int num;
      
      /** Error text. */
      public final String text;      
      
      /**
       * Construct an instance.
       *
       * @param    num
       *           The error number.
       * @param    text
       *           The text of the error message.
       */
      public ErrorEntry(int num, String text)
      {
         this.num  = num;
         this.text = text;
      }
   }
   
   /**
    * Provides access to the error data when running on the server.
    */
   private static class ServerDataAccess
   implements RemoteErrorData
   {
      /**
       * Queries the state of <code>silent</code> variable.
       *
       * @return   <code>true</code> if silent mode is enabled, otherwise
       *           <code>false</code>.
       */
      @Override
      public boolean isSilent()
      {
         return work.obtain().silent;
      }
   
      /**
       * Sets the <code>silent</code> variable.
       *
       * @param   value
       *          The new state of the variable.
       */
      @Override
      public void setSilent(boolean value)
      {
         work.obtain().silent = value;
      }
   
      /**
       * Queries the state of <code>ignore</code> variable.
       *
       * @return   <code>true</code> if silent mode is enabled, otherwise
       *           <code>false</code>.
       */
      @Override
      public boolean isIgnore()
      {
         return work.obtain().ignore;
      }
   
      /**
       * Sets the <code>ignore</code> variable.
       *
       * @param    value
       *           The new state of the variable.
       */
      @Override
      public void setIgnore(boolean value)
      {
         work.obtain().ignore = value;
      }
   
      /**
       * Queries the state of <code>error</code> variable.
       *
       * @return   <code>true</code> if error flag is turned on, otherwise
       *           <code>false</code>.
       */
      @Override
      public boolean isError()
      {
         return work.obtain().error;
      }
      
      /**
       * Sets the <code>error</code> variable.
       *
       * @param    value
       *           The new state of the variable.
       */
      @Override
      public void setError(boolean value)
      {
         work.obtain().error = value;
      }
   
      /**
       * Queries the state of <code>pendingError</code> variable.
       *
       * @return   <code>true</code> if pending error flag is turned on,
       *           otherwise <code>false</code>.
       */
      @Override
      public boolean isPending()
      {
         return work.obtain().pendingError;
      }
   
      /**
       * Sets the <code>pendingError</code> variable.
       *
       * @param    value
       *           The new state of the variable.
       */
      @Override
      public void setPending(boolean value)
      {
         WorkArea wa = work.obtain();
         
         // check if there was no prior pending error and that there now should
         // be an error pending
         if (!wa.pendingError && value)
         {
            // deferred clear of the error list when the first error occurs
            // after silent error mode is entered
            wa.errList.clear();
         }
            
         wa.pendingError = value;
      }
   
      /**
       * Determines if headless mode should be supported. In headless mode, no
       * client-side output will occur.
       *
       * @return   <code>true</code> if the headless flag is turned on,
       *           otherwise <code>false</code>.
       */
      @Override
      public boolean isHeadless()
      {
         return TransactionManager.isHeadless();
      }
      
      /**
       * Clears the list of error records.
       */
      @Override
      public void clearList()
      {
         work.obtain().errList.clear();
      }
   
      /**
       * Gets the size of the list of error records.
       *
       * @return   The size of the list.
       */
      @Override
      public int getSize()
      {
         return work.obtain().errList.size();
      }
   
      /**
       * Adds a record to the list of error records.
       *
       * @param    record
       *           The object to add to the list of error records.
       */
      @Override
      public void addRecord(Object record)
      {
         work.obtain().errList.add(record);
      }
   
      /**
       * Gets a record from the list of error records.
       *
       * @param    index
       *           The record index.
       *
       * @return   The object associated with the given index in the list.
       */
      @Override
      public Object getRecord(int index)
      {
         return work.obtain().errList.get(index);
      }
   }

   /**
    * Stores global data relating to the state of the current context's
    * transactions.
    */
   private static class WorkArea
   {  
      /** List of errors. */
      private List<Object> errList = new ArrayList<Object>();
   
      /** Stores whether an error condition has occurred previously. */
      private boolean error = false;
      
      /** 
       * No error flag (if <code>true</code> exceptions are silently recorded
       * rather than thrown).
       */ 
      private boolean silent = false;
      
      /** Stores whether an error condition has occurred currently. */
      private boolean pendingError = false;
      
      /**
       * Flags when a query is executed in a CAN-FIND function.
       * In a CAN-FIND, errors are "swallowed" similar to headless mode. 
       */
      private int canFindLevel = 0;
      
      /**
       * <code>true</code> to NO-OP all error recording, display and
       * exception generation.
       */
      private boolean ignore = false;
      
      /** Whether exceptions are processed as non-fatal, warning messages. */
      private boolean warning = false;
      
      /** 
       * Handle containing the value of the ERROR-OBJECT-DETAIL attribute.
       * TODO: this MUST be fixed. 
       *       if initialized with default constructor you'll get in pl/java
       *       ERROR:  java.lang.NoClassDefFoundError: Could not initialize class com.goldencode.p2j.util.handle
       */
      private handle soapFault = null; // new handle();
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Obtains the context-local instance of the contained global work
       * area.
       *
       * @return   The work area associated with this context.
       */
      public WorkArea obtain()
      {
         return this.get();
      }   
      
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       *
       * @return   The newly instantiated work area.
       */
      @Override
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }   
   }
}
