/*
** Module   : LowLevelSocket.java
** Abstract : An interface defining APIs specific to a client SOCKET.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  201309008 Created initial version.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.net.*;

/**
 * Interface defining APIs to connect or use a client socket in P2J. These interface is exported
 * as a service during P2J client startup.
 */
public interface LowLevelSocket
{
   /**
    * Establish a connection to the given host and port.
    * 
    * @param    resourceId
    *           The ID of the associated {@link SocketImpl} resource on server side.
    * @param    host
    *           The IP address or name of the target host.
    * @param    port
    *           The port number.
    * 
    * @return   The ID of the associated socket, if the connection was established.
    *           
    * @throws   IOException
    *           In case of errors while establishing the connection.
    */
   public int connect(int resourceId, String host, int port)
   throws IOException;

   /**
    * Establish a SSL connection to the given host and port.
    * 
    * @param    resourceId
    *           The ID of the associated {@link SocketImpl} resource on server side.
    * @param    host
    *           The IP address or name of the target host.
    * @param    port
    *           The port number.
    * @param    nosessionreuse
    *           Flag indicating that the ssl session ID must not be reused.
    * @param    nohostverify
    *           Flag indicating that the target host name must not be validated against the
    *           certificate's principal name.
    * 
    * @return   The ID of the associated socket, if the connection was established.
    *           
    * @throws   IOException
    *           In case of errors while establishing the SSL connection.
    */
   public int connectSSL(int     resourceId, 
                         String  host,
                         int     port,
                         boolean nosessionreuse,
                         boolean nohostverify)
   throws IOException;

   /**
    * Check if the java socket is still connected.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *           
    * @return   <code>true</code> if the java socket is still connected.
    */
   public boolean isConnected(int id);
   
   /**
    * Disconnect this socket.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *           
    * @return   <code>true</code> if no errors were encountered during disconnect.
    */
   public boolean disconnect(int id);
   
   /**
    * Get the available bytes to be read.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *
    * @return   See above.
    */
   public long getBytesAvailable(int id);
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    * @param    mode
    *           The mode to set for the reading operation.
    *        
    * @return   The number of read bytes. 0 means a timeout during read, -1 meas the stream 
    *           reached EOF (socket closed).
    *           
    * @throws   IOException
    *           In case IO errors were encoubtered during read.
    */
   public long read(int id, memptr buffer, long pos, long len, int mode)
   throws IOException;

   /**
    * Java implementation for the WRITE() SOCKET method, this will write data from the specified 
    * buffer location (position and number of bytes) to the socket. 
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * @param    buffer
    *           The memory buffer from which the data will be written to the socket.
    * @param    pos
    *           The position in the memory buffer from where the data to be taken starts.
    * @param    len
    *           The number of bytes to take from the memory buffer.
    *        
    * @return   The number of written bytes.
    *           
    * @throws   IOException
    *           In case IO errors were encoubtered during write.
    */
   public long write(int id, memptr buffer, long pos, long len)
   throws IOException;
   
   /**
    * In case of a SSL connection, get the digital certificate subject name of the server for
    * the current SSL session; else, return <code>null</code>.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    */
   public String getSslServerName(int id);
   
   /**
    * Get the TCP_NODELAY option value for this socket. 
    * See {@link java.net.Socket#getTcpNoDelay} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public boolean getTcpNoDelay(int id)
   throws SocketException;

   /**
    * Get the SO_LINGER option value for this socket.
    * See {@link java.net.Socket#getSoLinger} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public int getSoLinger(int id)
   throws SocketException;

   /**
    * Get the SO_KEEPALIVE option value for this socket.
    * See {@link java.net.Socket#getKeepAlive} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public boolean getKeepAlive(int id)
   throws SocketException;

   /**
    * Get the SO_REUSEADDR option value for this socket.
    * See {@link java.net.Socket#getReuseAddress} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public boolean getReuseAddress(int id)
   throws SocketException;

   /**
    * Get the SO_RCVBUF option value for this socket.
    * See {@link java.net.Socket#getReceiveBufferSize} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public int getReceiveBufferSize(int id)
   throws SocketException;

   /**
    * Get the SO_SNDBUF option value for this socket.
    * See {@link java.net.Socket#getSendBufferSize} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public int getSendBufferSize(int id)
   throws SocketException;

   /**
    * Get the SO_RCVTIMEO option value for this socket.
    * See {@link java.net.Socket#getSoTimeout} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public int getSoTimeout(int id)
   throws SocketException;

   /**
    * Set the TCP_NODELAY option for this socket.
    * See {@link java.net.Socket#setTcpNoDelay} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    on
    *           The TCP_NODELAY value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setTcpNoDelay(int id, boolean on)
   throws SocketException;

   /**
    * Set the SO_LINGER option for this socket.
    * See {@link java.net.Socket#setSoLinger} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    on
    *           <code>true</code> to turn on the SO_LINGER option.
    * @param    linger
    *           The linger time.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setSoLinger(int id, boolean on, int linger)
   throws SocketException;

   /**
    * Set the SO_KEEPALIVE option for this socket.
    * See {@link java.net.Socket#setKeepAlive} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    on
    *           The SO_KEEPALIVE value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setKeepAlive(int id, boolean on)
   throws SocketException;

   /**
    * Set the SO_REUSEADDR option for this socket.
    * See {@link java.net.Socket#setReuseAddress} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    on
    *           The SO_REUSEADDR value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setReuseAddress(int id, boolean on)
   throws SocketException;

   /**
    * Set the SO_RCVBUF value (the receive buffer size) for this socket.
    * See {@link java.net.Socket#setReceiveBufferSize} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    size
    *           The SO_RCVBUF value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setReceiveBufferSize(int id, int size)
   throws SocketException;

   /**
    * Set the SO_SNDBUF value (the send buffer size) for this socket.
    * See {@link java.net.Socket#setSendBufferSize} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    size
    *           The SO_SNDBUF value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setSendBufferSize(int id, int size)
   throws SocketException;

   /**
    * Set the SO_RCVTIMEO value (the number of seconds the socket waits to receive data before 
    * timing out) for this socket.
    * See {@link java.net.Socket#setSoTimeout} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    timeout
    *           The SO_RCVTIMEO value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setSoTimeout(int id, int timeout)
   throws SocketException;

   /**
    * Get the IP address of this side of this connection.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * 
    * @return   See above.
    */
   public String getLocalHost(int id);

   /**
    * Get the port number of this side of this connection.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * 
    * @return   See above.
    */
   public int getLocalPort(int id);

   /**
    * Get the host IP address of the remote side of this connection.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * 
    * @return   See above.
    */
   public String getRemoteHost(int id);

   /**
    * Get the port number of the remote side of this connection.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * 
    * @return   See above.
    */
   public int getRemotePort(int id);
}
