/*
** Module   : LowLevelSocketImpl.java
** Abstract : Implementation of a client socket.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  201309008 Created initial version.
*/
package com.goldencode.p2j.util;

import java.io.*;
import java.net.*;
import java.net.Socket;
import java.nio.channels.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;

import javax.net.ssl.*;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;

/**
 * This class acts as a delegate for the server-side to the created java socket.
 */
public class LowLevelSocketImpl
implements LowLevelSocket,
           SessionListener
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(LowLevelSocketImpl.class);
   
   /** Token used to authenticate with the dispatcher when registering APIs. */
   private static Object modToken = null;

   /** The context-local data of this class. */
   private static ContextLocal<WorkArea> local = 
   new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      };
   };
   
   /** The single instance available of this class. */
   private static final LowLevelSocketImpl instance = new LowLevelSocketImpl();
   
   /**
    * Private c'tor, to not allow explicit usage of this class. Its intended goal is to be 
    * exported as a network server.
    */
   private LowLevelSocketImpl()
   {
      // no-op
   }
   
   /**
    * Export the singleton instance of this class as a network server.
    * 
    * @param    session
    *           The P2J session. May be <code>null</code> if running in embedded mode.
    * @param    single
    *           <code>true</code> to startup within the server process which must bypass the 
    *           shared infrastructure initialization. Use <code>false</code> for the normal client
    *           JVM startup.
    */
   public static synchronized void initialize(Session session, boolean single)
   {
      final Class<?> ifc = LowLevelSocket.class;
      
      modToken = RemoteObject.registerServer(ifc, instance, modToken, single);

      if (session != null)
      {
         session.addSessionListener(instance);
      }
   }
   
   /**
    * When the server socket accepts a new connection, this will make sure the new socket is
    * registered and a {@link SocketImpl} resource is also created on P2J server side.
    * 
    * @param    client
    *           The client socket.
    * @param    ssl
    *           Flag indicating if this is a SSL connection.
    *           
    * @return   The created {@link SocketImpl} instance with the details of the new socket.
    *
    * @throws   IOException
    *           If the socket's input or output streams couldn't be obtained.
    */
   static SocketData registerClientSocket(Socket client, boolean ssl)
   throws IOException
   {
      String sslServerName = null;
      if (ssl)
      {
         try
         {
            sslServerName = ((SSLSocket) client).getSession().getPeerPrincipal().getName();
         }
         catch (Exception e)
         {
            if (LOG.isLoggable(Level.FINER))
            {
               LOG.log(Level.FINER, "Error determining the principal.", e);
            }
         }
      }

      int id = Utils.uniqueId();
      WorkArea wa = local.get();

      synchronized (wa)
      {
         SocketData sd = new SocketData(id, client, sslServerName);

         wa.addSocket(id, sd);
         
         sd.startThreads();

         return sd;
      }
   }

   /**
    * Identify the socket {@link SocketData socket data} using the given ID.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *           
    * @return   See above.
    */
   private static SocketData locate(int id)
   {
      return local.get().locate(id);
   }

   /**
    * Identify the associated {@link Socket socket} for the given ID.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *           
    * @return   See above.
    */
   private static Socket locateSocket(int id)
   {
      return local.get().locateSocket(id);
   }
   
   /**
    * Wait until all socket-related threads are initialized.
    * 
    * @param    id
    *           The socket ID on P2J client side.
    * @param    ready
    *           The signal to wait on. If <code>null</code>, it returns <code>true</code>.
    *           
    * @return   <code>true</code> if all threads are initialized or <code>ready</code> is null;
    *           <code>false</code> if someone interrupted the {@link CountDownLatch#await}
    */
   private static boolean isReady(int id, CountDownLatch ready)
   {
      try
      {
         ready.await();
         return true;
      }
      catch (InterruptedException e)
      {
         if (LOG.isLoggable(Level.FINER))
         {
            final String msg = "The [%d] socket is not initialized yet!";
            LOG.log(Level.FINER, String.format(msg, id), e);
         }

         return false;
      }
   }

   /**
    * Establish a connection to the given host and port.
    * 
    * @param    resourceId
    *           The ID of the associated {@link SocketImpl} resource on server side.
    * @param    host
    *           The IP address or name of the target host.
    * @param    port
    *           The port number.
    * 
    * @return   The ID of the associated socket, if the connection was established.
    *           
    * @throws   IOException
    *           In case of errors while establishing the connection.
    */
   @Override
   public int connect(int resourceId, String host, int port)
   throws IOException
   {
      int id = -1;

      InetSocketAddress address = new InetSocketAddress(host, port);
      Socket socket = new Socket();
      socket.bind(null);
      socket.connect(address);
      
      id = Utils.uniqueId();
      
      WorkArea wa = local.get();
      synchronized (wa)
      {
         SocketData sd = new SocketData(id, resourceId, socket, null);
         wa.addSocket(id, sd);

         sd.startThreads();
      }

      return id;
   }

   /**
    * Establish a SSL connection to the given host and port.
    * <p>
    * TODO: this is not yet implemented. Following are some findings/implementation notes:
    * <ul>
    *    <li><b>nosessionreuse</b>: "If specified, the connection does not reuse the SSL session 
    *        ID when reconnecting to the same SSL-enabled server socket." - couldn't find anything
    *        in JSSE which would allow control over the SSLSession attached to the SSLSocket.</li>
    *    <li>The certificate sent by the server socket needs to be validated. Need to determine if
    *        <b>nohostverify</b> option will disable validation completely (including expiration
    *        date) or it will just exclude the server host name from the validation.</li>
    *    <li>Documentation states that OpenEdge keeps the public keys (certificates) in a global 
    *        store. We can use the same storage facilities offered by the {@link TransportSecurity}
    *        and {@link SecurityManager}.</li>
    *    <li>Need to determine if the global certificate store is per progress installation or 
    *        per progress client.</li> 
    * </ul>
    * 
    * @param    resourceId
    *           The ID of the associated {@link SocketImpl} resource on server side.
    * @param    host
    *           The IP address or name of the target host.
    * @param    port
    *           The port number.
    * @param    nosessionreuse
    *           Flag indicating that the ssl session ID must not be reused.
    * @param    nohostverify
    *           Flag indicating that the target host name must not be validated against the
    *           certificate's principal name.
    * 
    * @return   The ID of the associated socket, if the connection was established.
    *           
    * @throws   IOException
    *           In case of errors while establishing the SSL connection.
    */
   @Override
   public int connectSSL(int     resourceId, 
                         String  host,
                         int     port,
                         boolean nosessionreuse,
                         boolean nohostverify)
   throws IOException
   {
      int id = -1;

      InetSocketAddress address = new InetSocketAddress(host, port);
      Socket lowSocket = new Socket();
      lowSocket.setReuseAddress(true);
      lowSocket.bind(null);
      lowSocket.connect(address);

      // TODO: how does it determine the certificate to be used for this connection?
      SSLContext        ctx        = null; //SSLContext.getInstance("TLS");
      // SSLSessionContext sslCtx  = ctx.getServerSessionContext();
      SSLSocketFactory  factory = ctx.getSocketFactory();
      Socket            socket  = factory.createSocket(lowSocket, host, port, true);
      SSLSession        sslSess = ((SSLSocket) socket).getSession();
      String      sslServerName = sslSess.getPeerPrincipal().getName();
      
      // for client sockets, host verification is needed: unless -nohostverify is set, 
      // the common name from the server's certificate and the target host name must match
      if (!nohostverify && !sslServerName.equals(host))
      {
         // TODO: return error code
         return -1;
      }

      /*
      if (sslCtx != null)
      {
         sslCtx.setSessionCacheSize(1);
         sslCtx.getsess
      }
      else
      {
         // TODO: log warning
      }
      */
      
      id = Utils.uniqueId();
      
      WorkArea wa = local.get();
      synchronized (wa)
      {
         SocketData sd = new SocketData(id, resourceId, socket, sslServerName);
         wa.addSocket(id, sd);

         sd.startThreads();
      }

      // TODO: return error code
      return id;
   }

   /**
    * Check if the java socket is still connected.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *           
    * @return   <code>true</code> if the java socket is still connected.
    */
   @Override
   public boolean isConnected(int id)
   {
      return !locate(id).listener.isTerminated();
   }

   /**
    * Disconnect this socket.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *           
    * @return   <code>true</code> if no errors were encountered during disconnect.
    */
   @Override
   public boolean disconnect(int id)
   {
      WorkArea wa = local.get();
      synchronized (wa)
      {
         SocketData sd = wa.locate(id);
         try
         {
            sd.socket.close();
         }
         catch (IOException e) 
         {
            if (LOG.isLoggable(Level.FINER))
            {
               final String msg = "Error closing socket [%d].";
               LOG.log(Level.FINER, String.format(msg, id), e);
            }
         }

         synchronized (sd.listener)
         {
            sd.listener.notify();
         }
         
         wa.removeSocket(id);
      }
      
      return true;
   }
   
   /**
    * Get the available bytes to be read.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *
    * @return   See above.
    */
   @Override
   public long getBytesAvailable(int id)
   {
      SocketData sd = locate(id);
      
      return sd.input.getBytesAvailable();
   }
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * <p>
    * During READ processing, the {@link ReadListener read listener} will be blocked from raising
    * new READ-RESPONSE events.
    *   
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    * @param    mode
    *           The mode to set for the reading operation.
    *        
    * @return   The number of read bytes. 0 means a timeout during read, -1 meas the stream 
    *           reached EOF (socket closed).
    *           
    * @throws   IOException
    *           In case IO errors were encoubtered during read.
    */
   @Override
   public long read(int id, memptr buffer, long pos, long len, int mode)
   throws IOException
   {
      final int CHUNK_SIZE = 512;
      SocketData sd = locate(id);

      // do not allow the socket to listen for incoming data and raise READ-RESPONSE events, 
      // as we are processing a READ
      sd.listener.startRead();

      long bytesRead = 0;
      
      try
      {
         if (mode == com.goldencode.p2j.util.Socket.READ_AVAILABLE)
         {
            // in this mode, we need to block until at least one byte is received and then read
            // at most len-1 bytes without blocking (len-1 as a byte was read with blocking) 
            do
            {
               // read a number of bytes, but make sure we don't exceed the remainder length and
               // always use chunks
               int toRead = (int) Math.min(sd.input.getBytesAvailable(), len);
               toRead = Math.min(CHUNK_SIZE, toRead);
               
               byte[] b = new byte[toRead];
               
               // save the actual number of bytes read (may be different)
               int read = sd.input.read(b, 0, toRead);
               
               for (int i = 0; i < read; i++)
               {
                  buffer.setByte(b[i], pos);
                  pos = pos + 1;
               }
               
               // decrement the number of bytes to be read
               len = len - read;
               
               bytesRead = bytesRead + read;
               
               // terminate when no more data is being received or the max length is reached
            }
            while (sd.input.getBytesAvailable() > 0 && len > 0);
         }
         else
         {
            // in this mode, we need to read exactly len bytes
            while (len > 0)
            {
               // while more data is needed to be read, read it in chunks
               int toRead = (int) Math.min(CHUNK_SIZE, len);
               byte[] b = new byte[toRead];
               int read = sd.input.read(b, 0, toRead);
               
               for (int i = 0; i < read; i++)
               {
                  buffer.setByte(b[i], pos);
                  pos = pos + 1;
               }
               
               // decrement the number of bytes to be read
               len = len - read;

               bytesRead = bytesRead + read;
            }
         }
      }
      finally
      {
         sd.listener.endRead();
      }
      
      return bytesRead;
   }
   
   /**
    * Java implementation for the WRITE() SOCKET method, this will write data from the specified 
    * buffer location (position and number of bytes) to the socket. 
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * @param    buffer
    *           The memory buffer from which the data will be written to the socket.
    * @param    pos
    *           The position in the memory buffer from where the data to be taken starts.
    * @param    len
    *           The number of bytes to take from the memory buffer.
    *        
    * @return   The number of written bytes.
    *           
    * @throws   IOException
    *           In case IO errors were encoubtered during write.
    */
   @Override
   public long write(int id, memptr buffer, long pos, long len)
   throws IOException
   {
      SocketData sd = locate(id);

      long writtenLength = 0;
      
      // the data is in memory (native if memptr, java heap if raw). we need to access it from 
      // there; write it in chunks of 32000 bytes (raw max size). this ensures we always bring
      // maximum amount of bytes into java heap (without using native memory again).
      while (len > 0)
      {
         // bring into java heap (in case memptr is used) at most raw.MAX_SIZE bytes from the 
         // buffer
         BinaryData data = buffer.getBytes(pos, Math.min(len, raw.MAX_SIZE));
         byte[] b = data.getByteArray();

         len -= b.length;
         pos += b.length;
         writtenLength += b.length;

         sd.output.write(b);
      }
      
      return writtenLength;
   }
   
   /**
    * In case of a SSL connection, get the digital certificate subject name of the server for
    * the current SSL session; else, return <code>null</code>.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    */
   @Override
   public String getSslServerName(int id)
   {
      return locate(id).sslServerName;
   }
   
   /**
    * Get the TCP_NODELAY option value for this socket. 
    * See {@link java.net.Socket#getTcpNoDelay} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   @Override
   public boolean getTcpNoDelay(int id)
   throws SocketException
   {
      return locateSocket(id).getTcpNoDelay();
   }

   /**
    * Get the SO_LINGER option value for this socket.
    * See {@link java.net.Socket#getSoLinger} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   @Override
   public int getSoLinger(int id)
   throws SocketException
   {
      return locateSocket(id).getSoLinger();
   }

   /**
    * Get the SO_KEEPALIVE option value for this socket.
    * See {@link java.net.Socket#getKeepAlive} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   @Override
   public boolean getKeepAlive(int id)
   throws SocketException
   {
      return locateSocket(id).getKeepAlive();
   }

   /**
    * Get the SO_REUSEADDR option value for this socket.
    * See {@link java.net.Socket#getReuseAddress} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   @Override
   public boolean getReuseAddress(int id)
   throws SocketException
   {
      return locateSocket(id).getReuseAddress();
   }

   /**
    * Get the SO_RCVBUF option value for this socket.
    * See {@link java.net.Socket#getReceiveBufferSize} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   @Override
   public int getReceiveBufferSize(int id)
   throws SocketException
   {
      return locateSocket(id).getReceiveBufferSize();
   }

   /**
    * Get the SO_SNDBUF option value for this socket.
    * See {@link java.net.Socket#getSendBufferSize} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   @Override
   public int getSendBufferSize(int id)
   throws SocketException
   {
      return locateSocket(id).getSendBufferSize();
   }

   /**
    * Get the SO_RCVTIMEO option value for this socket.
    * See {@link java.net.Socket#getSoTimeout} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   @Override
   public int getSoTimeout(int id)
   throws SocketException
   {
      return locateSocket(id).getSoTimeout();
   }

   /**
    * Set the TCP_NODELAY option for this socket.
    * See {@link java.net.Socket#setTcpNoDelay} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    on
    *           The TCP_NODELAY value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   @Override
   public void setTcpNoDelay(int id, boolean on)
   throws SocketException
   {
      locateSocket(id).setTcpNoDelay(on);
   }

   /**
    * Set the SO_LINGER option for this socket.
    * See {@link java.net.Socket#setSoLinger} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    on
    *           <code>true</code> to turn on the SO_LINGER option.
    * @param    linger
    *           The linger time.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   @Override
   public void setSoLinger(int id, boolean on, int linger)
   throws SocketException
   {
      locateSocket(id).setSoLinger(on, linger);
   }

   /**
    * Set the SO_KEEPALIVE option for this socket.
    * See {@link java.net.Socket#setKeepAlive} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    on
    *           The SO_KEEPALIVE value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   @Override
   public void setKeepAlive(int id, boolean on)
   throws SocketException
   {
      locateSocket(id).setKeepAlive(on);
   }

   /**
    * Set the SO_REUSEADDR option for this socket.
    * See {@link java.net.Socket#setReuseAddress} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    on
    *           The SO_REUSEADDR value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   @Override
   public void setReuseAddress(int id, boolean on)
   throws SocketException
   {
      locateSocket(id).setReuseAddress(on);
   }

   /**
    * Set the SO_RCVBUF value (the receive buffer size) for this socket.
    * See {@link java.net.Socket#setReceiveBufferSize} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    size
    *           The SO_RCVBUF value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   @Override
   public void setReceiveBufferSize(int id, int size)
   throws SocketException
   {
      locateSocket(id).setReceiveBufferSize(size);
   }

   /**
    * Set the SO_SNDBUF value (the send buffer size) for this socket.
    * See {@link java.net.Socket#setSendBufferSize} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    size
    *           The SO_SNDBUF value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   @Override
   public void setSendBufferSize(int id, int size)
   throws SocketException
   {
      locateSocket(id).setSendBufferSize(size);
   }

   /**
    * Set the SO_RCVTIMEO value (the number of seconds the socket waits to receive data before 
    * timing out) for this socket.
    * See {@link java.net.Socket#setSoTimeout} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    timeout
    *           The SO_RCVTIMEO value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   @Override
   public void setSoTimeout(int id, int timeout)
   throws SocketException
   {
      locateSocket(id).setSoTimeout(timeout);
   }

   /**
    * Get the IP address of this side of this connection.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * 
    * @return   See above.
    */
   @Override
   public String getLocalHost(int id)
   {
      return locateSocket(id).getLocalAddress().getHostAddress();
   }

   /**
    * Get the port number of this side of this connection.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * 
    * @return   See above.
    */
   @Override
   public int getLocalPort(int id)
   {
      return locateSocket(id).getLocalPort();
   }

   /**
    * Get the host IP address of the remote side of this connection.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * 
    * @return   See above.
    */
   @Override
   public String getRemoteHost(int id)
   {
      return locateSocket(id).getInetAddress().getHostAddress();
   }

   /**
    * Get the port number of the remote side of this connection.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * 
    * @return   See above.
    */
   @Override
   public int getRemotePort(int id)
   {
      return locateSocket(id).getPort();
   }
   
   /**
    * When the P2J client session is ending, disconnect all active sockets.
    * 
    * @param    session
    *           The session that is ending.
    */
   @Override
   public void terminate(Session session)
   {
      WorkArea wa = local.get();
      synchronized (wa)
      {
         Map<Integer, SocketData> copy = new HashMap<Integer, SocketData>(wa.sockets);
         for (SocketData socket : copy.values())
         {
            disconnect(socket.id);
         }
      }
   }

   /**
    * A container holding the various client-socket data.
    */
   static class SocketData
   {
      /** The ID of this socket, on P2J client-side. */
      final int id;
      
      /** The ID of this {@link com.goldencode.p2j.util.Socket} resource, on P2J server-side. */
      final int resourceId;
      
      /** Signals when all the threads have started. */
      private final CountDownLatch ready = new CountDownLatch(2);
      
      /** The associated socket. */
      private final Socket socket;
      
      /** The SSL server name, in case of a SSL connection. */
      private final String sslServerName;
      
      /** Listener dedicated to posting READ-RESPONSE events. */
      private final ReadListener listener;
      
      /**
       * A specialized instance dedicated to reading data from the socket, in a separate thread.
       */
      private final StagedInput input;
      
      /** This socket's output stream. */
      private final OutputStream output;
      
      /**
       * Save the data for this socket and create the specialized instances to 
       * {@link ReadListener raise READ-RESPONSE events} and to {@link StagedInput read data} from
       * the socket.
       * 
       * @param    id
       *           The ID of the associated socket on P2J Client side.
       * @param    socket
       *           The associated java socket.
       * @param    sslServerName
       *           The SSL server name, in case of a SSL connection. <code>null</code> otherwise.
       * 
       * @throws   IOException
       *           If the socket's input or output streams couldn't be obtained.
       */
      public SocketData(int id, Socket socket, String sslServerName)
      throws IOException
      {
         this.id = id;
   
         this.socket = socket;
         this.sslServerName = sslServerName;
         
         this.input = new StagedInput(id, socket, socket.getInputStream(), ready);
         this.output = socket.getOutputStream();
   
         // create the resource on the server if input/output were obtained
         ServerExports server = ThinClient.getInstance().getServer();
         this.resourceId = server.createSocket(id);
   
         this.listener = new ReadListener(this, ready);
      }
   
      /**
       * Save the data for this socket and create the specialized instances to 
       * {@link ReadListener raise READ-RESPONSE events} and to {@link StagedInput read data} from
       * the socket.
       * 
       * @param    id
       *           The ID of the associated socket on P2J Client side.
       * @param    resourceId
       *           The ID of the associated {@link com.goldencode.p2j.util.Socket} on P2J Server
       *           side.
       * @param    socket
       *           The associated java socket.
       * @param    sslServerName
       *           The SSL server name, in case of a SSL connection. <code>null</code> otherwise.
       * 
       * @throws   IOException
       *           If the socket's input or output streams couldn't be obtained.
       */
      public SocketData(int id, int resourceId, Socket socket, String sslServerName)
      throws IOException
      {
         this.id = id;
         this.resourceId = resourceId;
         this.socket = socket;
         this.sslServerName = sslServerName;
         
         this.input = new StagedInput(id, socket, socket.getInputStream(), ready);
         this.output = socket.getOutputStream();
         
         this.listener = new ReadListener(this, ready);
      }
   
      /**
       * Start the {@link ReadListener listener} and {@link StagedInput reader} threads.
       */
      public void startThreads()
      {
         this.input.start();
         this.listener.start();
      }
   }

   /**
    * Container for the context-local data. This will also ensure sync access to the 
    * {@link #sockets socket registry}.
    */
   private static class WorkArea
   {
      /** A registry containing the open sockets. */
      private final Map<Integer, SocketData> sockets = new HashMap<Integer, SocketData>();
      
      /**
       * Add a new socket to the {@link #sockets registry}.
       * 
       * @param    id
       *           The ID of the associated socket on P2J Client side.
       * @param    sd
       *           The data for this socket.
       */
      public synchronized void addSocket(int id, SocketData sd)
      {
         sockets.put(id, sd);
      }
      
      /**
       * Locate the socket with the given ID from the {@link #sockets registry}.
       * 
       * @param    id
       *           The ID of the associated socket on P2J Client side.
       */
      private synchronized Socket locateSocket(int id)
      {
         return sockets.get(id).socket;
      }

      /**
       * Locate the {@link SocketData} for the the given ID.
       * 
       * @param    id
       *           The ID of the associated socket on P2J Client side.
       */
      private synchronized SocketData locate(int id)
      {
         return sockets.get(id);
      }
      
      /**
       * Remove the socket with the given ID from the {@link #sockets registry}.
       * 
       * @param    id
       *           The ID of the associated socket on P2J Client side.
       */
      private synchronized void removeSocket(int id)
      {
         sockets.remove(id);
      }
   }
   
   /**
    * This class is dedicated to listening for incoming socket data. While the socket is still 
    * connected, it will check if the {@link StagedInput reader} can read some data without 
    * blocking. If so, it will post a READ-RESPONSE event on the server (the server-side is 
    * responsible of reaching back to the client and posting the event, as the server-side holds
    * the implementation).
    * <p>
    * Once a byte can be read, it will wait for a {@link LowLevelSocketImpl#read} to be executed,
    * as it must not post more than a READ-RESPONSE event per "incoming data" case.
    */
   private static class ReadListener
   implements Runnable
   {
      /** The associated {@link SocketData} instance. */
      private final SocketData sd;
      
      /** Flag indicating we are executing a READ. */
      private boolean inRead = false;
      
      /** Flag indicating the listener must been terminated. */
      private boolean terminated = false;

      /** The signal to use when this thread is ready. */
      private CountDownLatch ready;
      
      /**
       * Initialize this instance.
       * 
       * @param    sd
       *           The {@link SocketData} instance.
       */
      public ReadListener(SocketData sd, CountDownLatch ready)
      {
         this.sd = sd;
         this.ready = ready;
      }
      
      /**
       * Set the {@link #terminated} flag to the specified value.
       * 
       * @param    terminated
       *           The new value for the {@link #terminated} flag.
       */
      public synchronized void setTerminated(boolean terminated)
      {
         this.terminated = terminated;
      }
      
      /**
       * Check if the listener thread is terminated. This is true if {@link SocketData#socket} is
       * closed or the {@link #terminated} flag is true.
       * 
       * @return    See above.
       */
      public synchronized boolean isTerminated()
      {
         return terminated || sd.socket.isClosed();
      }
      
      /**
       * Execute a loop to post READ-RESPONSE events until the {@link #isTerminated()} returns.
       * <code>true</code>. 
       * <p>
       * Once the loop has been terminated, a final READ-RESPONSE event will be posted.
       */
      @Override
      public void run()
      {
         ready.countDown();

         try
         {
            ServerExports server = ThinClient.getInstance().getServer();
            
            while (!isTerminated())
            {
               // make sure no read is active
               waitForNotify(false);

               // a read can start here, but all server events during read all be discarded
               
               if (sd.input.hasMoreData())
               {
                  // post a ServerEvent. for this, must use a server-side call (as the server
                  // event is implemented there)
                  server.readResponseEvent(sd.resourceId);
                  
                  // do not process more events until a READ is executed
                  waitForNotify(true);
                  waitForNotify(false);
               }
               else
               {
                  setTerminated(true);
               }
            }
            
            // we have terminated, send a new READ-RESPONSE event
            server.readResponseEvent(sd.resourceId);
         }
         finally
         {
            SessionManager.get().deregisterAsyncThread(Thread.currentThread());
            
            // cleanup
            local.get().removeSocket(sd.id);
         }
      }
      
      /**
       * Start a new thread, in which this instance will be executed.  The thread's name will be
       * <code>Socket listener [%d]</code>, where <code>%d</code> is replaced with the 
       * {@link SocketData#id} value.
       * <p>
       * This thread is marked as async, as all outgoing requests must be processed async, in
       * their own {@link Dispatcher dispatcher}.
       */
      public void start()
      {
         Thread thread = new AssociatedThread(this);
         // mark this thread as async
         SessionManager.get().registerAsyncThread(thread);
         String name = String.format("Socket listener [%d]", sd.id);
         thread.setName(name);
         thread.start();
      }

      /**
       * Called when a READ is started. This will set the {@link #inRead} flag to true and notify
       * all listeners.
       */
      public synchronized void startRead()
      {
         if (!isReady(sd.id, ready))
         {
            return;
         }

         inRead = true;
         notify();
      }
      
      /**
       * Called when a REAd is ended. This will set the {@link #inRead} flag to false and notify
       * all listeners.
       */
      public synchronized void endRead()
      {
         if (!isReady(sd.id, ready))
         {
            return;
         }

         inRead = false;
         notify();
      }
      
      /**
       * Wait for notification that a read has either started or ended.
       * 
       * @param    inRead
       *           <code>true</code> to wait for a READ start, <code>false</code> to wait for a 
       *           READ end.
       */
      private synchronized void waitForNotify(boolean inRead)
      {
         while (this.inRead != inRead && !sd.socket.isClosed())
         {
            try
            {
               // wait for the other thread to do its job
               wait();
            }
            catch (InterruptedException exc)
            {
               // ignore
            }
         }
      }
   }
   
   /**
    * Due to the nature of java sockets and how the 4GL sockets behave, all reading must be done
    * in a dedicated thread. Also, reading data via a java socket input stream is not thread-safe; 
    * thus, reading data must be staged, following these rules, after a byte is read:
    * <ol>
    *    <li>Save it in {@link #lastByte} field.</li>
    *    <li>Notify the {@link ReadListener} (which waits for data in {@link #hasMoreData}, 
    *        via the {@link #lLock}.
    *    <li>If EOF is reached, then end the loop.</li> 
    *    <li>Wait for someone to read the data (on the {@link #rLock}). The {@link #rLock} will be
    *        notified only after a {@link #read} call is finished.</li>
    * </ol>
    * The {@link ReadListener} will always call {@link #hasMoreData()}, which will 
    * block on the {@link #lLock} until a byte was read.
    * <p>
    * A thread wanting to read data from the socket will always call {@link #read}, which will
    * block for a byte to be read using {@link #hasMoreData()}.
    * <p>
    * A thread wanting to interrogate the available bytes on the input stream will always call
    * {@link #getBytesAvailable}, which will use the {@link InputStream#available} to determine
    * the available bytes.
    * <p>
    * This approach was chosen after chasing a few other solution:
    * <ol>
    *    <li>Using {@link InputStream#available()} on the socket's input stream to check if data
    *        is incoming is not enough because this will not block. A solution which calls 
    *        <code>available()</code> in a loop (eventually after waiting for some millis) will 
    *        pose unnecessary overhead on the CPU or it will cause delays from posting the 
    *        READ-RESPONSE event.</li>
    *    <li>If channels are used, a {@link Selector} can be used to block until data is available.
    *        But this will force the input and output stream to be non-blocking. Also, when using
    *        channels in blocking mode, read and write can't be used in multiple threads; if a 
    *        thread is blocking on a read, then a write will be possible from another thread only
    *        if the read has released the channel's {@link SocketChannel#blockingLock() lock}.
    *     </li>
    * </ol>
    */
   private static class StagedInput
   implements Runnable
   {
      /** Value indicating that no byte was read. */
      private static final int NO_BYTE = -2;
      
      /** Lock used by the {@link ReadListener}. */
      private final Object lLock = new Object();

      /** Lock used when reading. */
      private final Object rLock = new Object();

      /** The socket's input stream. */
      private final InputStream input;
      
      /** The associated java socket. */
      private final Socket socket;
      
      /** The ID of this socket on P2J client side. */
      private final int id;

      /** The signal to use when this thread is ready. */
      private CountDownLatch ready;
      
      /** The last byte read. */
      private int lastByte = NO_BYTE;
      
      /**
       * Initialize this staged input stream.
       * 
       * @param    id
       *           The ID of the associated socket on P2J Client side.
       * @param    socket
       *           The associated java socket.
       * @param    input
       *           The socket's input stream.
       * @param    ready
       *           The signal to use when this thread is ready.
       */
      public StagedInput(int id, Socket socket, InputStream input, CountDownLatch ready)
      {
         this.id = id;
         this.socket = socket;
         this.input = input;
         this.ready = ready;
      }
      
      /**
       * Wait for data to be received by the socket, in a dedicated thread.
       */
      @Override
      public void run()
      {
         ready.countDown();
         
         try
         {
            while (!socket.isClosed())
            {
               try
               {
                  int b = input.read();
   
                  // I've read data, notify the listener
                  synchronized (lLock)
                  {
                     lastByte = b;
                     lLock.notify();
                  }
                  
                  if (b == -1)
                  {
                     // EOF reached
                     break;
                  }
   
                  synchronized (rLock)
                  {
                     // wait for the data to be read
                     try
                     {
                        rLock.wait();
                     }
                     catch (InterruptedException e)
                     {
                        // ignore this, but log it
                        if (LOG.isLoggable(Level.FINER))
                        {
                           LOG.log(Level.FINER,
                                   "The lock waiting for someone to read the data was interrupted.",
                                   e); 
                        }
                     }
                  }
               }
               catch (IOException e)
               {
                  if (LOG.isLoggable(Level.FINER))
                  {
                     LOG.log(Level.FINER, "Problems reading the initial byte!", e);
                  }
   
                  // could not read, exit
                  break;
               }
            }
         }
         finally
         {
            SessionManager.get().deregisterAsyncThread(Thread.currentThread());
         }
      }
      
      /**
       * Check if the socket has more data to be read. This will block until a byte was read or
       * EOF is reached.
       * 
       * @return   <code>true</code> if data is incoming.
       */
      public boolean hasMoreData()
      {
         if (!isReady(id, ready))
         {
            return false;
         }

         try
         {
            return hasMoreDataWorker(0);
         }
         catch (SocketTimeoutException exc)
         {
            return false;
         }
      }
      
      /**
       * Determine the number of available bytes on the socket's input stream.  This will include
       * {@link #lastByte} too, if it was read previously.
       * 
       * @return   See above.
       */
      public int getBytesAvailable()
      {
         if (!isReady(id, ready))
         {
            return 0;
         }

         synchronized (lLock)
         {
            try
            {
               return input.available() + (lastByte >= 0 ? 1 : 0);
            }
            catch (IOException e)
            {
               if (LOG.isLoggable(Level.FINER))
               {
                  final String msg = "Error in getBytesAvailable for %d";
                  LOG.log(Level.FINER, String.format(msg, id), e);
               }

               return 0;
            }
         }
      }
      
      /**
       * Read <code>len</code> number of bytes and place them in the specified <code>buffer</code>,
       * starting with the specified position.
       * 
       * @param    buffer
       *           The buffer where to place the read data.
       * @param    pos
       *           The position in the buffer where to place the data.
       * @param    len
       *           The number of bytes to read.
       *
       * @return   The number of read bytes.
       * 
       * @throws   IOException
       *           In case of read errors.
       */
      public int read(byte[] buffer, int pos, int len)
      throws IOException
      {
         if (!isReady(id, ready))
         {
            return 0;
         }

         if (len == 0)
         {
            // nothing to read
            return 0;
         }

         int timeout = 0;
         try
         {
            timeout = socket.getSoTimeout();
         }
         catch (IOException e)
         {
            if (LOG.isLoggable(Level.FINER))
            {
               LOG.log(Level.FINER, "Error determining SO_RCVTIMEO, assuming 0", e);
            }
         }
         
         // wait for data to be available, using the current SO-TIMEOUT value
         if (!hasMoreDataWorker(timeout))
         {
            return -1;
         }
         
         // consume the first byte
         byte b = -1;
         synchronized (lLock)
         {
            b = (byte) lastByte;
            lastByte = NO_BYTE;
         }

         synchronized (rLock)
         {
            buffer[0] = b;
            pos = pos + 1;
            len = len - 1;
            int read = 1;

            if (len > 0)
            {
               try
               {
                  read = read + input.read(buffer, pos, len);
               }
               finally
               {
                  rLock.notify();
               }
            }
            
            return read;
         }
      }

      /**
       * Start a new thread, in which this instance will be executed.  The thread's name will be
       * <code>Socket reader [%d]</code>, where <code>%d</code> is replaced with the {@link #id}
       * value.
       * <p>
       * This thread is marked as async, as all outgoing requests must be processed async, in
       * their own {@link Dispatcher dispatcher}.
       */
      public void start()
      {
         Thread thread = new AssociatedThread(this);
         // mark this thread as async
         SessionManager.get().registerAsyncThread(thread);
         String name = String.format("Socket reader [%d]", id);
         thread.setName(name);
         thread.start();
      }

      /**
       * Check if the socket has more data to be read. This will block until a byte was read or
       * EOF is reached.
       * 
       * @param    timeout
       *           the value of millis to wait; <code>0</code> to wait indefinitely.
       * 
       * @return   <code>true</code> if data is incoming.
       * 
       * @throws   SocketTimeoutException
       *           If no byte was read.
       */
      private boolean hasMoreDataWorker(int timeout)
      throws SocketTimeoutException
      {
         int b = NO_BYTE;
      
         synchronized (lLock)
         {
            while (lastByte == NO_BYTE)
            {
               try
               {
                  lLock.wait(timeout);
               }
               catch (InterruptedException e)
               {
                  // someone interrupted us, terminate the loop
                  break;
               }
            }
            
            if (lastByte != NO_BYTE)
            {
               b = lastByte;
            }
         }
         
         if (b == NO_BYTE)
         {
            final String msg = "Timeout during read for %d";
            throw new SocketTimeoutException(String.format(msg, id));
         }
         
         return b >= 0;
      }
   }
}
