/*
** Module   : LowLevelSocketListener.java
** Abstract : An interface defining server-socket related APIs which execute on P2J client-side.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130902 Created initial version.
*/

package com.goldencode.p2j.util;

import java.io.IOException;

/**
 * Interface for Server-Socket APIs which need to be implemented on the P2J client-side.
 */
public interface LowLevelSocketListener
{
   /**
    * Create a server socket and let it listen on the given port.
    * 
    * @param    resourceId
    *           The ID of the associated {@link SocketImpl} resource on server side.
    * @param    port
    *           The port number.
    * @param    qsize
    *           The backlog size for this socket.
    * 
    * @return   The ID of the associated {@link LowLevelSocketListenerImpl} instance when 
    *           the connection was established.
    *           
    * @throws   IOException
    *           In case of errors while starting the server.
    */
   public int enableConnections(int resourceId, int port, int qsize)
   throws IOException;

   /**
    * Create a SSL server socket and let it listen on the given port.
    * 
    * @param    resourceId
    *           The ID of the associated {@link SocketImpl} resource on server side.
    * @param    port
    *           The port number.
    * @param    qsize
    *           The backlog size for this socket.
    * @param    keyalias
    *           The alias for the digital certificate.
    * @param    keyaliaspasswd
    *           The password for the key of the digital certificate.
    * @param    nosessioncache
    *           Flag indicating the SSL client session cache is disabled.
    * @param    sessiontimeout
    *           The number of seconds before the SSL client request is timeout.
    * 
    * @return   The ID of the associated {@link LowLevelSocketListenerImpl} instance when 
    *           the connection was established.
    *           
    * @throws   IOException
    *           In case of errors while starting the server socket.
    */
   public int enableSSLConnections(int     resourceId,
                                   int     port,
                                   int     qsize,
                                   String  keyalias,
                                   String  keyaliaspasswd,
                                   boolean nosessioncache,
                                   int     sessiontimeout)
   throws IOException;

   /**
    * Terminates the listening loop for this server socket. The existing established connections
    * are not terminated.
    * 
    * @param    id
    *           The ID of {@link LowLevelSocketListenerImpl} instance on P2J Client side.
    *           
    * @return   <code>true</code> if no errors were encountered during disconnect.
    */
   public boolean disableConnections(int id);
}
