/*
** Module   : LowLevelSocketListenerImpl.java
** Abstract : Implementation of a server socket.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130908 Created initial version.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;
import java.util.logging.*;
import java.net.*;
import java.net.Socket;

import javax.net.ssl.*;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.util.LowLevelSocketImpl.SocketData;

/**
 * This class acts as a delegate for the server-side to the create java socket.
 */
public class LowLevelSocketListenerImpl
implements LowLevelSocketListener,
           SessionListener
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(LowLevelSocketListenerImpl.class);

   /** Token used to authenticate with the dispatcher when registering APIs. */
   private static Object modToken = null;

   /** The context-local data of this class: a registry of server sockets by their ID. */
   private static ContextLocal<Map<Integer, ServerSocketData>> sockets = 
   new ContextLocal<Map<Integer, ServerSocketData>>()
   {
      protected Map<Integer,ServerSocketData> initialValue()
      {
         return new HashMap<Integer, LowLevelSocketListenerImpl.ServerSocketData>();
      };
   };
   
   /** The single instance available of this class. */
   private static final LowLevelSocketListenerImpl instance = new LowLevelSocketListenerImpl();
   
   /**
    * Private c'tor, to not allow explicit usage of this class. Its intended goal is to be 
    * exported as a network server.
    */
   private LowLevelSocketListenerImpl()
   {
      // no-op
   }
   
   /**
    * Export the singleton instance of this class as a network server.
    * 
    * @param    session
    *           The P2J session. May be <code>null</code> if running in embedded mode.
    * @param    single
    *           <code>true</code> to startup within the server process which must bypass the 
    *           shared infrastructure initialization. Use <code>false</code> for the normal client
    *           JVM startup.
    */
   public static synchronized void initialize(Session session, boolean single)
   {
      final Class<?> ifc = LowLevelSocketListener.class;
      
      modToken = RemoteObject.registerServer(ifc, instance, modToken, single);
      
      if (session != null)
      {
         session.addSessionListener(instance);
      }
   }

   /**
    * Create a server socket and let it listen on the given port.
    * 
    * @param    resourceId
    *           The ID of the associated {@link SocketImpl} resource on server side.
    * @param    port
    *           The port number.
    * @param    qsize
    *           The backlog size for this socket.
    * 
    * @return   The ID of the associated {@link LowLevelSocketListenerImpl} instance when 
    *           the connection was established.
    *           
    * @throws   IOException
    *           In case of errors while starting the server.
    */
   @Override
   public int enableConnections(int resourceId, int port, int qsize)
   throws IOException
   {
      ServerSocket socket = new ServerSocket(port, qsize);

      int id = Utils.uniqueId();

      Map<Integer, ServerSocketData> map = sockets.get();
      synchronized (map)
      {
         ServerSocketData data = new ServerSocketData(id, resourceId, socket, null);
         map.put(id, data);
         
         data.startThreads();
      }

      return id;
   }

   /**
    * Create a SSL server socket and let it listen on the given port.
    * <p>
    * TODO: this is not yet implemented. Following are some findings/implementation notes:
    * <ul>
    *    <li><b>nosessioncache</b> - this disables the SSL session cache. The {@link SSLContext} 
    *        or {@link SSLSessionContext} classes don't seem to offer APIs to disable the SSL 
    *        session cache.</li>
    *    </li>
    *    <li><b>sessiontimeout</b> represents "The maximum amount of time, in seconds, that the 
    *        server waits before it rejects a SSL client’s request to resume a session. The 
    *        default value is 180 seconds."  This parameter looks like is something which should 
    *        be set by the client and not by the server, as (CA thinkgs) it means "if client is 
    *        idle more than x seconds, than terminate connection". 
    *        {@link SSLSocket}, {@link SSLContext} or {@link SSLSessionContext} don't have APIs
    *        which would suggest that would set a "disconnect idle client after" value.</li>
    *    <li>Documentation states that OpenEdge keeps the private keys in a global store. We can
    *        use the same storage facilities offered by the {@link TransportSecurity} and
    *        {@link SecurityManager}, but we will need to provide a facility to encrypt the key's
    *        password.</li>
    *    <li>Need to determine if the global store is per progress installation or per progress
    *        client.</li> 
    *    <li>there is a <code>default_server</code> key with the default password set to 
    *        <code>password</code>. This is used when connecting in SSL mode, but no key alias or 
    *        password is specified. Confirmation is needed if these defaults can be configurable 
    *        in OpenEdge.</li>
    *    <li>the password specified via keyaliaspasswd option in ENABLE-CONNECTIONS is encrypted. 
    *        4GL has a genpassword utility to encrypt a password; the assumption is that it uses 
    *        a master password (as salt) and the given password to generate some hash value. 
    *        In our implementation, we will need to encrypt the password too, and I think we can 
    *        use a similar approach: we can keep the (keyalias, keypassword) in some separate 
    *        directory node/configuration file and just check the hash sent by the client against  
    *        the hash obtained from the master password and the associated alias key; if these 
    *        match, use the real keypassword to decript the private key.</li>
    * </ul>
    *
    * @param    resourceId
    *           The ID of the associated {@link SocketImpl} resource on server side.
    * @param    port
    *           The port number.
    * @param    qsize
    *           The backlog size for this socket.
    * @param    keyalias
    *           The alias for the digital certificate.
    * @param    keyaliaspasswd
    *           The password for the key of the digital certificate.
    * @param    nosessioncache
    *           Flag indicating the SSL client session cache is disabled.
    * @param    sessiontimeout
    *           The number of seconds before the SSL client request is timeout.
    * 
    * @return   The ID of the associated {@link LowLevelSocketListenerImpl} instance when 
    *           the connection was established.
    *           
    * @throws   IOException
    *           In case of errors while starting the server socket.
    */
   @Override
   public int enableSSLConnections(int     resourceId,
                                   int     port,
                                   int     qsize,
                                   String  keyalias,
                                   String  keyaliaspasswd,
                                   boolean nosessioncache,
                                   int     sessiontimeout)
   throws IOException
   {
      int id = -1;

      // TODO: there should be a central store where these certificates are kept.
      SSLContext             ctx     = null;
      // SSLSessionContext      sslCtx  = ctx.getClientSessionContext();
      SSLServerSocketFactory factory = ctx.getServerSocketFactory();
      SSLServerSocket        secure  = (SSLServerSocket) factory.createServerSocket(port);
      
      /*
      if (sslCtx != null)
      {
         // TODO: how to disable the cache?
         // sslCtx.setSessionCacheSize(nosessioncache ? 1 : 0);
         sslCtx.setSessionTimeout(sessiontimeout);
      }
      else
      {
         // TODO: log warning that nosessioncache and sessiontimeout could not be set!
      }
      */

      // TODO: check this
      secure.setWantClientAuth(true);
      
      id = Utils.uniqueId();
      
      Map<Integer, ServerSocketData> map = sockets.get();
      synchronized (map)
      {
         ServerSocketData data = new ServerSocketData(id, resourceId, null, secure);
         map.put(id, data);
         
         data.listener.start();
      }

      return id;
   }

   /**
    * Terminates the listening loop for this server socket. The existing established connections
    * are not terminated.
    * 
    * @param    id
    *           The ID of {@link LowLevelSocketImpl} instance on P2J Client side.
    *           
    * @return   <code>true</code> if no errors were encountered during disconnect.
    */
   @Override
   public boolean disableConnections(int id)
   {
      Map<Integer, ServerSocketData> map = sockets.get();
      synchronized (map)
      {
         ServerSocketData server = map.remove(id);
         server.terminate();
      }
      
      return true;
   }

   /**
    * When the P2J client session is ending, terminate all active server sockets.
    * 
    * @param    session
    *           The session that is ending.
    */
   @Override
   public void terminate(Session session)
   {
      Map<Integer, ServerSocketData> data = sockets.get();
      
      synchronized (data)
      {
         Map<Integer, ServerSocketData> copy = new HashMap<Integer, ServerSocketData>(data);
         for (ServerSocketData server : copy.values())
         {
            disableConnections(server.id);
         }
      }
   }

   /**
    * Container of various server-socket related data.
    */
   private static class ServerSocketData
   {
      /** The local ID of this server socket. */
      private final int id;
      
      /** The non-ssl server socket. <code>null</code> if ssl mode. */
      private final ServerSocket socket;
      
      /** The ssl server socket. <code>null</code> if no-ssl mode. */
      private final SSLServerSocket sslSocket;
      
      /** Instance of this class will start listening for connections in a different thread. */
      private final ConnectListener listener;
      
      /** Flag indicating that the server socket is terminated. */
      private boolean terminated = false;
      
      /**
       * Save the given data in a new {@link ServerSocketData} instance.
       * 
       * @param    id
       *           The local ID for this server socket.
       * @param    resourceId
       *           The resource ID of the server socket on P2J serer side.
       * @param    socket
       *           The server-socket for non-ssl mode; <code>null</code> if ssl mode.
       * @param    sslSocket
       *           The server-socket for ssl mode; <code>null</code> if no-ssl mode.
       */
      public ServerSocketData(int             id,
                              int             resourceId,
                              ServerSocket    socket,
                              SSLServerSocket sslSocket)
      {
         this.id = id;
         this.socket = socket;
         this.sslSocket = sslSocket;
         
         this.listener = new ConnectListener(resourceId, this);
      }
      
      /**
       * Start the {@link #listener connect listener} thread.
       */
      public void startThreads()
      {
         this.listener.start();
      }

      /**
       * Terminate this server socket.
       */
      public synchronized void terminate()
      {
         this.terminated = true;
         
         try
         {
            if (socket != null)
            {
               socket.close();
            }
            else 
            {
               sslSocket.close();
            }
         }
         catch (IOException e)
         {
            if (LOG.isLoggable(Level.FINER))
            {
               final String msg = "Error terminating server socket [%d]";
               LOG.log(Level.FINER, String.format(msg, id), e);
            }

            // ignore exceptions triggered by socket close
         }
      }
      
      /**
       * Check if this server socket is terminated.
       * 
       * @return   The {@link #terminated} flag.
       */
      public synchronized boolean isTerminated()
      {
         return terminated;
      }
   }
   
   /**
    * Start a loop in a different thread and listen for incoming connections.
    */
   private static class ConnectListener
   implements Runnable
   {
      /** The remote resource ID for this server-socket. */
      private final int resourceId;
      
      /** Back reference to the data for this server socket. */
      private final ServerSocketData data;

      /**
       * Initialize this instance.
       * 
       * @param    resourceId
       *           The ID of the {@link SocketListener resource} on P2J server side.
       * @param    data
       *           The associated {@link ServerSocketData}.
       */
      public ConnectListener(int resourceId, ServerSocketData data)
      {
         this.resourceId = resourceId;
         this.data = data;
      }

      /**
       * Listen for incoming connections, in a different thread.
       */
      public void run()
      {
         try
         {
            ServerExports server = ThinClient.getInstance().getServer();
            
            while (!data.isTerminated())
            {
               SocketData clientData = accept();
               
               if (clientData != null)
               {
                  server.connectEvent(resourceId, clientData.resourceId);
               }
            }
         }
         finally
         {
            SessionManager.get().deregisterAsyncThread(Thread.currentThread());
         }
      }
      
      /**
       * Start a new thread, in which this instance will be executed.  The thread's name will be
       * <code>Server socket listener [%d]</code>, where <code>%d</code> is replaced with the 
       * {@link SocketData#id} value.
       * <p>
       * This thread is marked as async, as all outgoing requests must be processed async, in
       * their own {@link Dispatcher dispatcher}.
       */
      public void start()
      {
         Thread t = new AssociatedThread(this);
         // mark this thread as async
         SessionManager.get().registerAsyncThread(t);
         String name = String.format("Server socket listener [%d]", data.id);
         t.setName(name);
         t.start();
      }

      /**
       * Wait for a new incoming connection.
       * 
       * @return   The {@link SocketData} for this client socket.
       */
      private SocketData accept()
      {
         boolean ssl = (data.sslSocket != null);
         SocketData cldata = null;
         
         SessionManager.get().getSession();
         
         try
         {
            Socket client = ssl ? data.sslSocket.accept() : data.socket.accept();

            cldata = LowLevelSocketImpl.registerClientSocket(client, ssl);
         }
         catch (IOException e)
         {
            if (LOG.isLoggable(Level.FINER))
            {
               final String msg = "Error while accepting a new connection on [%d]";
               LOG.log(Level.FINER, String.format(msg, data.id), e);
            }

            // ignore errors
         }
         
         return cldata;
      }
   }
}
