/*
** Module   : Socket.java
** Abstract : An interface holding the SOCKET specific methods and 
**            attributes api.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CS  20130204 Created initial version.
** 002 CS  20130204 Moved the SSL-SERVER attribute to it's own interface.
** 003 CA  20130908 Javadoc changes.
*/

package com.goldencode.p2j.util;

/**
 * Interface for SOCKET methods and attributes
 * <p>
 * Exposes the following Attributes and methods:
 * <table>
 * <tr>
 *    <th>attribute or method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>LOCAL-HOST</td>
 *    <td>{@link #getLocalHost()}</td>
 * </tr>
 * <tr>
 *    <td>LOCAL-PORT</td>
 *    <td>{@link #getLocalPort()}</td>
 * </tr>
 * 
 * <tr>
 *    <td>REMOTE-HOST</td>
 *    <td>{@link #getRemoteHost()}</td>
 * </tr>
 * <tr>
 *    <td>REMOTE-PORT</td>
 *    <td>{@link #getRemotePort()}</td>
 * </tr>
 * <tr>
 *    <td>BYTES-READ</td>
 *    <td>{@link #getBytesRead()}</td>
 * </tr>
 * <tr>
 *    <td>BYTES-WRITTEN</td>
 *    <td>{@link #getBytesWritten()}</td>
 * </tr>
 * <tr>
 *    <td>GET-SOCKET-OPTION()</td>
 *    <td>{@link #getSocketOption(character)}, 
 *        {@link #getSocketOption(String)}</td>
 * </tr>
 * 
 * <tr>
 *    <td>GET-BYTES-AVAILABLE()</td>
 *    <td>{@link #getBytesAvailable()}</td>
 * </tr>
 * 
 * <tr>
 *    <td>READ()</td>
 *    <td>{@link #read(memptr, double, double)}, 
 *        {@link #read(memptr, NumberType, double)}, 
 *        {@link #read(memptr, double, NumberType)}, 
 *        {@link #read(memptr, NumberType, NumberType)},
 *        {@link #read(memptr, double, double, double)}, 
 *        {@link #read(memptr, NumberType, double, double)}, 
 *        {@link #read(memptr, double, NumberType, double)}, 
 *        {@link #read(memptr, NumberType, NumberType, double)}
 *        {@link #read(memptr, double, double, NumberType)}, 
 *        {@link #read(memptr, NumberType, double, NumberType)}, 
 *        {@link #read(memptr, double, NumberType, NumberType)}, 
 *        {@link #read(memptr, NumberType, NumberType, NumberType)}</td>
 * </tr>
 * 
 * <tr>
 *    <td>WRITE()</td>
 *    <td>{@link #write(memptr, double, double)},
 *        {@link #write(memptr, NumberType, double)},
 *        {@link #write(memptr, double, NumberType)},
 *        {@link #write(memptr, NumberType, NumberType)},</td>
 * </tr>
 * <tr>
 *    <td>SET-READ-RESPONSE-PROCEDURE()</td>
 *    <td>{@link #setReadResponseProcedure(String)}, 
 *        {@link #setReadResponseProcedure(character)},
 *        {@link #setReadResponseProcedure(String, handle)}, 
 *        {@link #setReadResponseProcedure(character, handle)} </td>
 * </tr>
 * <tr>
 *    <td>SET-SOCKET-OPTION()</td>
 *    <td>{@link #setSocketOption(String, String)},
 *        {@link #setSocketOption(character, String)},
 *        {@link #setSocketOption(String, character)},
 *        {@link #setSocketOption(character, character)}, </td>
 * </tr>
 * </table>
 * This interface is used as a returned type by the 
 * {@link handle#unwrapSocket(handle)} API, which, will return a
 * {@link SocketImpl} for a Socket type handle.
 */
public interface Socket
extends CommonHandleChain
{
   /** Constant defining socket read mode: READ-AVAILABLE */
   public static final int READ_AVAILABLE = 1;
   
   /** Constant for socket read mode: READ-EXACT-NUM */
   public static final int READ_EXACT_NUM = 2;

   /**
    * Java implementation for getting the LOCAL-HOST SOCKET attribute which contains the address 
    * in IP form of the machine to which the Socket is connected.
    * 
    * @return   The address in IP form of the machine to which the Socket is connected.
    */
   public character getLocalHost();
   
   /**
    * Java implementation for getting the LOCAL-PORT SOCKET attribute which contains the local 
    * port of the machine to which the Socket is connected.
    * 
    * @return   The local port of the machine to which the Socket is connected.
    */
   public integer getLocalPort();
   
   /**
    * Java implementation for getting the REMOTE-HOST SOCKET attribute which contains the address
    * in IP form of the machine to which the Socket is connected.
    * 
    * @return    The remote address in IP form of the machine to which the Socket is connected.
    */
   public character getRemoteHost();
   
   /**
    * Java implementation for getting the REMOTE-PORT SOCKET attribute which contains the local 
    * port of the machine to which the Socket is connected.
    * 
    * @return    The remote port of the machine to which the Socket is connected.
    */
   public integer getRemotePort();
   
   /**
    * Java implementation for getting the  BYTES-READ SOCKET attribute which contains the number 
    * of bytes the last read() method call returned from the socket.
    * 
    * @return   The number of bytes read the read() method has returned from the socket.
    */
   public integer getBytesRead();
   
   /**
    * Java implementation for getting the  BYTES-WRITTEN SOCKET attribute which contains the 
    * number of bytes the last write() method has put in the socket.
    * 
    * @return   The number of bytes the last write() method has put in the socket.
    */
   public integer getBytesWritten();
   
   /**
    * Java implementation for the GET-SOCKET-OPTION() SOCKET method.  This will return the 
    * specified socket option value.
    * <p>
    * The following options names can be used:
    * <ul>
    *    <li>TCP-NODELAY</li>
    *    <li>SO-LINGER</li>
    *    <li>SO-KEEPALIVE</li>
    *    <li>SO-REUSEADDR</li>
    *    <li>SO-RCVBUF</li>
    *    <li>SO-SNDBUF</li>
    *    <li>SO-RCVTIMEO</li>
    * <ul>
    * 
    * @param    optionName
    *           The name of the socket option we want to retrieve.
    *         
    * @return   The specified socket option value.
    */
   public character getSocketOption(character optionName);
   
   /**
    * Java implementation for the GET-SOCKET-OPTION() SOCKET method.  This will return the 
    * specified socket option value.
    * <p>
    * The following options names can be used:
    * <ul>
    *    <li>TCP-NODELAY</li>
    *    <li>SO-LINGER</li>
    *    <li>SO-KEEPALIVE</li>
    *    <li>SO-REUSEADDR</li>
    *    <li>SO-RCVBUF</li>
    *    <li>SO-SNDBUF</li>
    *    <li>SO-RCVTIMEO</li>
    * <ul>
    * 
    * @param    optionName
    *           The name of the socket option we want to retrieve.
    *         
    * @return   The specified socket option value.
    */
   public character getSocketOption(String optionName);
   
   /**
    * Java implementation for the GET-BYTES-AVAILABLE() SOCKET method, this will return the number
    * of bytes that can be read at the current moment from the socket.
    * 
    * @return   The number of bytes that can be read at the current moment from the socket.
    */
   public integer getBytesAvailable();
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    *        
    * @return   <code>true</code> if the read was successful or <code>false</code> otherwise.
    */
   public logical read(memptr buffer, double pos, double len); 
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    *        
    * @return   <code>true</code> if the read was successful or <code>false</code> otherwise.
    */
   public logical read(memptr buffer, NumberType pos, double len); 
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    *        
    * @return   <code>true</code> if the read was successful or <code>false</code> otherwise.
    */
   public logical read(memptr buffer, double pos, NumberType len); 
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    *        
    * @return   <code>true</code> if the read was successful or <code>false</code> otherwise.
    */
   public logical read(memptr buffer, NumberType pos, NumberType len);
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    * @param    mode
    *           The mode to set for the reading operation.
    *        
    * @return   <code>true</code> if the read was successful or <code>false</code> otherwise.
    */
   public logical read(memptr buffer, double pos, double len, double mode);
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    * @param    mode
    *           The mode to set for the reading operation.
    *        
    * @return   <code>true</code> if the read was successful or <code>false</code> otherwise.
    */
   public logical read(memptr buffer, NumberType pos, double len, double mode); 
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    * @param    mode
    *           The mode to set for the reading operation.
    *        
    * @return   <code>true</code> if the read was successful or <code>false</code> otherwise.
    */
   public logical read(memptr buffer, double pos, NumberType len, double mode);
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    * @param    mode
    *           The mode to set for the reading operation.
    *        
    * @return   <code>true</code> if the read was successful or <code>false</code> otherwise.
    */
   public logical read(memptr buffer, NumberType pos, NumberType len, double mode);
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    * @param    mode
    *           The mode to set for the reading operation.
    *        
    * @return   <code>true</code> if the read was successful or <code>false</code> otherwise.
    */
   public logical read(memptr buffer, double pos, double len, NumberType mode);
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    * @param    mode
    *           The mode to set for the reading operation.
    *        
    * @return   <code>true</code> if the read was successful or <code>false</code> otherwise.
    */
   public logical read(memptr buffer, NumberType pos, double len, NumberType mode); 
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    * @param    mode
    *           The mode to set for the reading operation.
    *        
    * @return   <code>true</code> if the read was successful or <code>false</code> otherwise.
    */
   public logical read(memptr buffer, double pos, NumberType len, NumberType mode); 
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    * @param    mode
    *           The mode to set for the reading operation.
    *        
    * @return   <code>true</code> if the read was successful or <code>false</code> otherwise.
    */
   public logical read(memptr buffer, NumberType pos, NumberType len, NumberType mode);
   
   /**
    * Java implementation for the WRITE() SOCKET method, this will write data from the specified 
    * buffer location (position and number of bytes) to the socket. 
    * 
    * @param    buffer
    *           The memory buffer from which the data will be written to the socket.
    * @param    pos
    *           The position in the memory buffer from where the data to be taken starts.
    * @param    len
    *           The number of bytes to take from the memory buffer.
    *         
    * @return   <code>true</code> if the write was successful or <code>false</code> otherwise.
    */
   public logical write(memptr buffer, double pos, double len);
   
   /**
    * Java implementation for the WRITE() SOCKET method, this will write data from the specified 
    * buffer location (position and number of bytes) to the socket. 
    * 
    * @param    buffer
    *           The memory buffer from which the data will be written to the socket.
    * @param    pos
    *           The position in the memory buffer from where the data to be taken starts.
    * @param    len
    *           The number of bytes to take from the memory buffer.
    *         
    * @return   <code>true</code> if the write was successful or <code>false</code> otherwise.
    */
   public logical write(memptr buffer, NumberType pos, double len);
   
   /**
    * Java implementation for the WRITE() SOCKET method, this will write data from the specified 
    * buffer location (position and number of bytes) to the socket. 
    * 
    * @param    buffer
    *           The memory buffer from which the data will be written to the socket.
    * @param    pos
    *           The position in the memory buffer from where the data to be taken starts.
    * @param    len
    *           The number of bytes to take from the memory buffer.
    *         
    * @return   <code>true</code> if the write was successful or <code>false</code> otherwise.
    */
   public logical write(memptr buffer, double pos, NumberType len);
   
   /**
    * Java implementation for the WRITE() SOCKET method, this will write data from the specified 
    * buffer location (position and number of bytes) to the socket. 
    * 
    * @param    buffer
    *           The memory buffer from which the data will be written to the socket.
    * @param    pos
    *           The position in the memory buffer from where the data to be taken starts.
    * @param    len
    *           The number of bytes to take from the memory buffer.
    *         
    * @return   <code>true</code> if the write was successful or <code>false</code> otherwise.
    */
   public logical write(memptr buffer, NumberType pos, NumberType len);
   
   /**
    * Java implementation for the SET-READ-RESPONSE-PROCEDURE() SOCKET method.  This will set the
    * procedure to execute when a READ-RESPONSE event is received.
    * 
    * @param    procedureName
    *           The name of the procedure to be executed on READ-RESPONSE event.
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setReadResponseProcedure(String procedureName);
   
   /**
    * Java implementation for the SET-READ-RESPONSE-PROCEDURE() SOCKET method.  This will set the
    * procedure to execute when a READ-RESPONSE event is received.
    * 
    * @param    procedureName
    *           The name of the procedure to be executed on READ-RESPONSE event.
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setReadResponseProcedure(character procedureName);
   
   /**
    * Java implementation for the SET-READ-RESPONSE-PROCEDURE() SOCKET method.  This will set the
    * procedure to execute when a READ-RESPONSE event is received.
    * 
    * @param    procedureName
    *           The name of the procedure to be executed on READ-RESPONSE event.
    * @param    procHandle
    *           The context where the specified procedure belongs. Validation will be performed
    *           when the READ-RESPONSE event is executed.
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setReadResponseProcedure(String procedureName, handle procHandle); 

   /**
    * Java implementation for the SET-READ-RESPONSE-PROCEDURE() SOCKET method.  This will set the
    * procedure to execute when a READ-RESPONSE event is received.
    * 
    * @param    procedureName
    *           The name of the procedure to be executed on READ-RESPONSE event.
    * @param    procHandle
    *           The context where the specified procedure belongs. Validation will be performed
    *           when the READ-RESPONSE event is executed.
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setReadResponseProcedure(character procedureName, handle procHandle);

   /**
    * Java implementation for the SET-SOCKET-OPTION method.  This will set the value for the 
    * specified option.
    * <p>
    * The following options are available:
    * <ul>
    *    <li>TCP-NODELAY
    *    <li>SO-LINGER
    *    <li>SO-KEEPALIVE
    *    <li>SO-REUSEADDR
    *    <li>SO-RCVBUF
    *    <li>SO-SNDBUF
    *    <li>SO-RCVTIMEO
    * <ul>
    * 
    * @param    optionName
    *           The name of the socket option we want to set.
    * @param    optionVal
    *           The value of the option we want to set
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setSocketOption(String optionName, String optionVal);
   
   /**
    * Java implementation for the SET-SOCKET-OPTION method.  This will set the value for the 
    * specified option.
    * <p>
    * The following options are available:
    * <ul>
    *    <li>TCP-NODELAY
    *    <li>SO-LINGER
    *    <li>SO-KEEPALIVE
    *    <li>SO-REUSEADDR
    *    <li>SO-RCVBUF
    *    <li>SO-SNDBUF
    *    <li>SO-RCVTIMEO
    * <ul>
    * 
    * @param    optionName
    *           The name of the socket option we want to set.
    * @param    optionVal
    *           The value of the option we want to set
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setSocketOption(character optionName, String optionVal);
   
   /**
    * Java implementation for the SET-SOCKET-OPTION method.  This will set the value for the 
    * specified option.
    * <p>
    * The following options are available:
    * <ul>
    *    <li>TCP-NODELAY
    *    <li>SO-LINGER
    *    <li>SO-KEEPALIVE
    *    <li>SO-REUSEADDR
    *    <li>SO-RCVBUF
    *    <li>SO-SNDBUF
    *    <li>SO-RCVTIMEO
    * <ul>
    * 
    * @param    optionName
    *           The name of the socket option we want to set.
    * @param    optionVal
    *           The value of the option we want to set
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setSocketOption(String optionName, character optionVal);
   
   /**
    * Java implementation for the SET-SOCKET-OPTION method.  This will set the value for the 
    * specified option.
    * <p>
    * The following options are available:
    * <ul>
    *    <li>TCP-NODELAY
    *    <li>SO-LINGER
    *    <li>SO-KEEPALIVE
    *    <li>SO-REUSEADDR
    *    <li>SO-RCVBUF
    *    <li>SO-SNDBUF
    *    <li>SO-RCVTIMEO
    * <ul>
    * 
    * @param    optionName
    *           The name of the socket option we want to set.
    * @param    optionVal
    *           The value of the option we want to set
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setSocketOption(character optionName, character optionVal);
}
