/*
** Module   : SocketListener.java
** Abstract : An interface holding the SERVER-SOCKET specific methods and 
**            attributes api.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 CS  20130204 Created initial version.
** 002 CA  20130908 Javadoc changes.
*/

package com.goldencode.p2j.util;

/**
 * Interface for SERVER-SOCKET methods and attributes
 * <p>
 * Exposes the following Attributes and methods:
 * <table>
 * <tr>
 *    <th>attribute or method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>ENABLE-CONNECTIONS</td>
 *    <td>{@link #enableConnections(String)}, 
 *        {@link #enableConnections(character)}
 *    </td>
 * </tr>
 * <tr>
 *    <td>DISABLE-CONNECTIONS</td>
 *    <td>{@link #disableConnections()}</td>
 * </tr>
 * <tr>
 *    <td>SET-CONNECT-PROCEDURE</td>
 *    <td>{@link #setConnectProcedure(String)},
 *        {@link #setConnectProcedure(character)},
 *        {@link #setConnectProcedure(String, handle) },
 *        {@link #setConnectProcedure(character, handle)},
 *        
 *    </td>
 * </tr>
 * </table>
 * This interface is used as a returned type by the 
 * {@link handle#unwrapSocketListener(handle)} API, which, will return a
 * {@link SocketListenerImpl} for a Server-Socket type handle.
 */
public interface SocketListener 
extends CommonHandleChain
{
   /**
    * Java implementation for the ENABLE-CONNECTION() method.  This will attempt to start 
    * listening on the specified port and will return <code>true</code> if operation succeeded, 
    * or <code>false</code> otherwise.
    * <p>
    * The following options can be specified as valid ENABLE-CONNECTIONS() options:
    * <ul>
    *    <li><b>-S port</b> The port to which to connect.</li>
    *    <li><b>-ssl</b> If specified the connection will in ssl mode.</li>
    *    <li><b>-pf config-file-name</b> Specifies the file from which options will be read.</li>
    *    <li><b>-qsize maxconn</b> The maximum number of connections to handle  before refusing 
    *        extra incoming connections.</li>
    *    <li><b>-keyalias alias</b> The alias for the digital certificate, default is:
    *        <b>default_server.</b></li>
    *    <li><b>-keyaliaspasswd pwd</b> The password for the key of the digital certificate. 
    *        To use always with -keyalias option.</li>
    *    <li><b>-nossesioncache</b> Disables SSL client session cache.</li>
    *    <li><b>-sessiontimeout time</b> The number of seconds before the SSL client request is 
    *        timeout, default value is 180.</li>
    * </ul>
    * 
    * @param    option
    *           The set of options that define the enable-connections method.
    * 
    * @return   <code>true</code> if operation succeeded, or <code>false</code> otherwise.
    */
   public logical enableConnections(String option);
   
   /**
    * Java implementation for the ENABLE-CONNECTION() method.  This will attempt to start 
    * listening on the specified port and will return <code>true</code> if operation succeeded, 
    * or <code>false</code> otherwise.
    * <p>
    * The following options can be specified as valid ENABLE-CONNECTIONS() options:
    * <ul>
    *    <li><b>-S port</b> The port to which to connect.</li>
    *    <li><b>-ssl</b> If specified the connection will in ssl mode.</li>
    *    <li><b>-pf config-file-name</b> Specifies the file from which options will be read.</li>
    *    <li><b>-qsize maxconn</b> The maximum number of connections to handle  before refusing 
    *        extra incoming connections.</li>
    *    <li><b>-keyalias alias</b> The alias for the digital certificate, default is:
    *        <b>default_server.</b></li>
    *    <li><b>-keyaliaspasswd pwd</b> The password for the key of the digital certificate. 
    *        To use always with -keyalias option.</li>
    *    <li><b>-nossesioncache</b> Disables SSL client session cache.</li>
    *    <li><b>-sessiontimeout time</b> The number of seconds before the SSL client request is 
    *        timeout, default value is 180.</li>
    * </ul>
    * 
    * @param    option
    *           The set of options that define the enable-connections method.
    * 
    * @return   <code>true</code> if operation succeeded, or <code>false</code> otherwise.
    */
   public logical enableConnections(character option);
   
   /**
    * Java implementation for the DISABLE-CONNECTIONS() SERVER-SOCKET method.  This will stop 
    * listening and accepting any new connections, but will interrupt the existing ones.
    * 
    * @return    <code>true</code> if the operation succeeded or <code>false</code> otherwise.
    */
   public logical disableConnections();
   
   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method.  This will set 
    * the procedure to execute when a CONNECT event is received.
    * 
    * @param    procedureName
    *           The name of the procedure to be executed on CONNECT event.
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setConnectProcedure(String procedureName);
   
   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method.  This will set 
    * the procedure to execute when a CONNECT event is received.
    * 
    * @param    procedureName
    *           The name of the procedure to be executed on CONNECT event.
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setConnectProcedure(character procedureName);
   
   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method.  This will set 
    * the procedure to execute when a CONNECT event is received.
    * 
    * @param    procedureName
    *           The name of the procedure to be executed on CONNECT event.
    * @param    procHandle
    *           The context where the specified procedure belongs. Validation will be performed
    *           when the CONNECT event is executed.
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setConnectProcedure(String procedureName, handle procHandle);
   
   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method.  This will set 
    * the procedure to execute when a CONNECT event is received.
    * 
    * @param    procedureName
    *           The name of the procedure to be executed on CONNECT event.
    * @param    procHandle
    *           The context where the specified procedure belongs. Validation will be performed
    *           when the CONNECT event is executed.
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setConnectProcedure(character procedureName, handle procHandle);
}
