/*
** Module   : SocketListenerImpl.java
** Abstract : A class for implementing the SERVER-SOCKET method and attributes
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------------------Description---------------------------------
** 001 CS  20130204 Created initial version.
** 002 CA  20130221 Added Deletable support. setType was renamed to setResourceType.
** 003 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 004 CA  20130603 resourceDelete must return a boolean value.
** 005 CA  20130902 Added runtime support for server sockets.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;

/**
 * A class for implementing the SERVER-SOCKET method and attributes.
 */ 
public class SocketListenerImpl 
extends SensitiveResource
implements SocketListener,
           Sensitive
{   
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(SocketListenerImpl.class);

   /** Client side network instance for handling sockets. */
   private static final ContextLocal<LowLevelSocketListener> remoteSocket =
   new ContextLocal<LowLevelSocketListener>()
   {
      protected LowLevelSocketListener initialValue()
      {
         Class<?> cls = LowLevelSocketListener.class;
         Session session = SessionManager.get().getSession();
         
         if (session != null)
         {
            return (LowLevelSocketListener) RemoteObject.obtainNetworkInstance(cls);
         }
         else
         {
            return (LowLevelSocketListener) RemoteObject.obtainLocalInstance(cls, true);
         }
      };
   };
   
   /** The name of the procedure to be executed when new clients connect, if any. */
   private final character connectProcedure = new character();
   
   /** The context of the procedure to be executed when new clients connect, if any. */
   private final handle connectProcedureContext = new handle();
   
   /** Boolean flag to track connection state. */
   private boolean connected = false;
   
   /** The associated server socket id on P2J client side. If not connected, defaults to -1. */
   private int serverSocketId = -1;
   
   /** Flag to track the valid state of this resource. */
   private boolean valid = true;
   
   /**
    * Default constructor which initializes SERVER-SOCKET specific settings.
    */
   public SocketListenerImpl()
   {
      setResourceType("SERVER-SOCKET");
      setUnknown(false);
   }

   /**
    * Reports if this object is valid for use.  
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return valid;
   }
   
   /**
    * Java implementation for the ENABLE-CONNECTION() method.  This will attempt to start 
    * listening on the specified port and will return <code>true</code> if operation succeeded, 
    * or <code>false</code> otherwise.
    * <p>
    * The following options can be specified as valid ENABLE-CONNECTIONS() options:
    * <ul>
    *    <li><b>-S port</b> The port to which to connect.</li>
    *    <li><b>-ssl</b> If specified the connection will in ssl mode.</li>
    *    <li><b>-pf config-file-name</b> Specifies the file from which options will be read.</li>
    *    <li><b>-qsize maxconn</b> The maximum number of connections to handle  before refusing 
    *        extra incoming connections.</li>
    *    <li><b>-keyalias alias</b> The alias for the digital certificate, default is:
    *        <b>default_server.</b></li>
    *    <li><b>-keyaliaspasswd pwd</b> The password for the key of the digital certificate. 
    *        To use always with -keyalias option.</li>
    *    <li><b>-nossesioncache</b> Disables SSL client session cache.</li>
    *    <li><b>-sessiontimeout time</b> The number of seconds before the SSL client request is 
    *        timeout, default value is 180.</li>
    * </ul>
    * 
    * @param    option
    *           The set of options that define the enable-connections method.
    * 
    * @return   <code>true</code> if operation succeeded, or <code>false</code> otherwise.
    */
   public logical enableConnections(String option)
   {
      return enableConnections(new character(option));
   }
   
   /**
    * Java implementation for the ENABLE-CONNECTION() method.  This will attempt to start 
    * listening on the specified port and will return <code>true</code> if operation succeeded, 
    * or <code>false</code> otherwise.
    * <p>
    * The following options can be specified as valid ENABLE-CONNECTIONS() options:
    * <ul>
    *    <li><b>-S port</b> The port to which to connect.</li>
    *    <li><b>-ssl</b> If specified the connection will in ssl mode.</li>
    *    <li><b>-pf config-file-name</b> Specifies the file from which options will be read.</li>
    *    <li><b>-qsize maxconn</b> The maximum number of connections to handle  before refusing 
    *        extra incoming connections.</li>
    *    <li><b>-keyalias alias</b> The alias for the digital certificate, default is:
    *        <b>default_server.</b></li>
    *    <li><b>-keyaliaspasswd pwd</b> The password for the key of the digital certificate. 
    *        To use always with -keyalias option.</li>
    *    <li><b>-nossesioncache</b> Disables SSL client session cache.</li>
    *    <li><b>-sessiontimeout time</b> The number of seconds before the SSL client request is 
    *        timeout, default value is 180.</li>
    * </ul>
    * 
    * @param    option
    *           The set of options that define the enable-connections method.
    * 
    * @return   <code>true</code> if operation succeeded, or <code>false</code> otherwise.
    */
   public logical enableConnections(character option)
   {
      if (connected)
      {
         final String msg = "Only one ENABLE-CONNECTIONS can be active at any one time";
         ErrorManager.recordOrShowError(9185, msg, false, false);
         
         return new logical(false);
      }
      
      Map<String, String> parms = parseOptions(option);
      
      if (parms != null && !parms.containsKey("-S"))
      {
         ErrorManager.recordOrShowError(5488, "The -S parameter must be specified to connect",
                                        false, false);
         return new logical(false);
      }
      
      if (parms == null)
      {
         ErrorManager.recordOrShowError(5510, "Invalid parameter string", false, false);
         return new logical(false);
      }

      int qsize = ConnectHelper.getInt(parms, "-qsize", 0);
      String serviceName = ConnectHelper.getString(parms, "-S", null);
      String saveServiceName = serviceName;

      int port = -1;
      try
      {
         port = Integer.parseInt(serviceName);
         serviceName = "";
      }
      catch (NumberFormatException e)
      {
         // in case of server sockets, the target service name must be registered in the 
         // directory's "port_services" mapping
         port = Utils.getPortForService(null, serviceName);
      }
      
      if (port == -1)
      {
         final String msg = "Service name %s could not be resolved to a port!";
         throw new IllegalArgumentException(String.format(msg, serviceName));
      }
         
      boolean ssl = parms.containsKey("-ssl");
      int resId = handle.resourceId(this);
      
      try
      {
         serverSocketId = -1;
         if (!ssl)
         {
            serverSocketId = remoteSocket.get().enableConnections(resId, port, qsize);
         }
         else
         {
            String keyalias = ConnectHelper.getString(parms, "-keyalias", "default_server");
            // TODO: the password must be encrypted?
            String keyaliaspasswd = ConnectHelper.getString(parms, "-keyaliaspasswd", "password");
            boolean nosessioncache = parms.containsKey("-nosessioncache");
            int sessiontimeout = ConnectHelper.getInt(parms, "-sessiontimeout", 180);
            
            serverSocketId = remoteSocket.get().enableSSLConnections(resId, 
                                                                     port,
                                                                     qsize,
                                                                     keyalias,
                                                                     keyaliaspasswd,
                                                                     nosessioncache,
                                                                     sessiontimeout);
   
         }
         
         connected = (serverSocketId != -1);
      }
      catch (BindException e)
      {
         if (LOG.isLoggable(Level.FINER))
         {
            LOG.log(Level.FINER, "Error while establishing connection!", e);
         }

         final String msg = "Service %s transport TCP is busy";
         ErrorManager.recordOrShowError(5485, String.format(msg, serviceName), false);
      }
      catch (IOException e)
      {
         if (LOG.isLoggable(Level.FINER))
         {
            LOG.log(Level.FINER, "Error while establishing connection!", e);
         }

         final String msg = "Unexpected error %s when connecting port %s";
         String errMsg = String.format(msg, e.getMessage(), saveServiceName);
         ErrorManager.recordOrThrowError(5485, errMsg);
      }
      
      return new logical(connected);
   }

   /**
    * Java implementation for the DISABLE-CONNECTIONS() SERVER-SOCKET method.  This will stop 
    * listening and accepting any new connections, but will interrupt the existing ones.
    * 
    * @return    <code>true</code> if the operation succeeded or <code>false</code> otherwise.
    */
   public synchronized logical disableConnections()
   {
      connected = remoteSocket.get().disableConnections(serverSocketId);
      
      if (!connected)
      {
         this.serverSocketId = -1;
      }
      
      return new logical(connected);
   }
   
   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method.  This will set 
    * the procedure to execute when a CONNECT event is received.
    * 
    * @param    procedureName
    *           The name of the procedure to be executed on CONNECT event.
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setConnectProcedure(String procedureName)
   {
      return setConnectProcedure(new character(procedureName), ProcedureManager.thisProcedure());
   }

   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method.  This will set 
    * the procedure to execute when a CONNECT event is received.
    * 
    * @param    procedureName
    *           The name of the procedure to be executed on CONNECT event.
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setConnectProcedure(character procedureName)
   {
      return setConnectProcedure(procedureName, ProcedureManager.thisProcedure());
   }

   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method.  This will set 
    * the procedure to execute when a CONNECT event is received.
    * 
    * @param    procedureName
    *           The name of the procedure to be executed on CONNECT event.
    * @param    procHandle
    *           The context where the specified procedure belongs. Validation will be performed
    *           when the CONNECT event is executed.
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setConnectProcedure(String procedureName, handle procHandle)
   {
      return setConnectProcedure(new character(procedureName), procHandle);
   }

   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method.  This will set 
    * the procedure to execute when a CONNECT event is received.
    * 
    * @param    procedureName
    *           The name of the procedure to be executed on CONNECT event.
    * @param    procHandle
    *           The context where the specified procedure belongs. Validation will be performed
    *           when the CONNECT event is executed.
    *         
    * @return   <code>true</code> if the operation was successful or <code>false</code> otherwise.
    */
   public logical setConnectProcedure(character procedureName, handle procHandle)
   {
      this.connectProcedure.assign(procedureName);
      this.connectProcedureContext.assign(procHandle);

      return new logical(true);
   }

   /**
    * Called by the P2J client side, it will inform the server that a server event was generated.
    * Although the event originates from the P2J Client Side, socket events are server-side 
    * events.
    * 
    * @param    clientSocketId
    *           The resource ID of the socket created for the new connected accepted by the
    *           server socket.
    *           
    * @return   The ID of the generated server event or <code>-1</code> if a event could not be
    *           posted (i.e. the resource is not SENSITIVE).
    */
   public int connectEvent(final int clientSocketId)
   {
      if (!isSensitive().booleanValue())
      {
         // if the resource is not sensitive, will ignore the event.
         return -1;
      }

      int resId = handle.resourceId(this);
      
      Runnable task = new Runnable()
      {
         @Override
         public void run()
         {
            handle h = handle.fromString(new integer(clientSocketId).toStringMessage());

            invokeConnectProcedure(h);

            // the client socket is not deleted/disconnected implicitly
         }
      };

      // raise a CONNECT event
      ServerEvent event = new ServerEvent(resId, Keyboard.CONNECT, task);
      LogicalTerminal.postServerEvent(event);
      
      return event.id();
   }

   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      if (connected)
      {
         final String msg = "Server-Socket is still connected. Cannot DELETE";
         ErrorManager.recordOrThrowError(10096, msg);
         return false;
      }

      this.valid = false;
      return true;
   }
   
   /**
    * Invoke the {@link #connectProcedure connect procedure}, as a new client has connected to
    * this socket.  If {@link #connectProcedure} is unknown, this is a no-op.
    * 
    * @param    socket
    *           A handle referring the connected {@link SocketImpl client socket}.
    */
   private void invokeConnectProcedure(handle socket)
   {
      if (connectProcedure.isUnknown())
      {
         return;
      }
      
      // invoke the event procedure
      try
      {
         SelfManager.pushSelf(new handle(this));
         ControlFlowOps.invokeInWithMode(connectProcedure, connectProcedureContext, "I", socket);
      }
      finally
      {
         SelfManager.popSelf();
      }
   }

   /**
    * Utility method for parsing the {@link #enableConnections} parameters.
    * 
    * @param    options
    *           The string to parse.
    * 
    * @return   The parameter-to-value map or <code>null</code> if something could not be parsed.
    */
   private Map<String, String> parseOptions(character options)
   {
      final Set<String> knownOptions = new HashSet<String>();
      knownOptions.add("-S");
      knownOptions.add("-pf");
      knownOptions.add("-qsize");
      knownOptions.add("-ssl"); // no value
      knownOptions.add("-keyalias");
      knownOptions.add("-keyaliaspasswd");
      knownOptions.add("-nosessioncache"); // no value
      knownOptions.add("-sessiontimeout");
      
      final Set<String> noValueOptions = new HashSet<String>();
      noValueOptions.add("-ssl"); // no value
      
      final Map<String, Integer> intParams = new HashMap<String, Integer>();
      intParams.put("-qsize", 0);
      intParams.put("-sessiontimeout", 0);
      
      return ConnectHelper.parseOptions(options, knownOptions, noValueOptions, intParams);
   }
}
