/*
** Module   : handle.java
** Abstract : Progress 4GL compatible handle object
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -----------------------Description-----------------------------------
** 001 GES 20051125   @23505 Created initial version which represents
**                           the Progress handle data type.
** 002 GES 20060206   @24300 Made compareTo() safe to process any object
**                           type (avoids ClassCastException).
** 003 GES 20060208   @24390 Added a default format string getter.
** 004 GES 20060724   @28161 Added a forced assign() abstract method.
** 005 GES 20060807   @28479 Moved to externalizable interface to optimize
**                           network performance. Throw an exception if
**                           an attempt is made to serialize this class
**                           since it makes no sense to transfer this
**                           between systems.
** 006 GES 20070219   @32175 Added support for the widget-handle built-in
**                           function.
** 007 GES 20070221   @32195 Added support for the TransparentWrapper as
**                           a contained object that exposes its
**                           referenced object instead.
** 008 GES 20080402   @37825 Added a new assign(Object) signature and a
**                           new constructor handle(WrappedResource).
** 009 CA  20090227   @41383 Undefined SELF or FOCUS handles refers a
**                           NullWidget instance - modified the isUnknown
**                           method to return true when the referenced
**                           value is not null, but unknown.
** 010 ECF 20101003          Removed ContextContainer inner class; replaced with
**                           parameterized use of ContextLocal. Fixed javadoc.
** 011 CA  20130112          Major rework to implement generic handles. 
**                           Added c'tor which builds a new instance from a
**                           BaseDataType value, if possible. This should be
**                           used only when constructing an instance using the
**                           value returned by a DYNAMIC-FUNCTION call; 4GL
**                           does some special casting in this case, which is
**                           not implemented yet.
** 012 EVL 20130130          Adding wrap methods to invoke DOM XML object
**                           method calls. The new supported class: XEntity,
**                           XNodeRef and XDocument.
** 013 CS 20130204           Added conversion support for socket and 
**                           server-socket.
** 014 VMN 20130207          Added unwrapBuffer(), unwrapQuery(), unwrapFieldInterface()
** 015 OM  20130129          Added TempTable handle unwrapper.
** 016 EVL 20130208          Adding wrap methods to invoke SAX object method calls. The new
**                           supported class: XCommon, SaxEntity, SaxAttributes, SaxReader and
**                           SaxWriter.
** 017 CA  20130216          Added unwrappers for SOAP-related handles.
** 018 CS  20130218          Added conversion support for app server.
** 019 VMN 20130219          Added support for attributes
**                           ERROR (KW_ERROR)
**                           NAME (KW_NAME)
**                           PRIVATE-DATA (KW_PRIV_DATA)
**                           DATA-TYPE (KW_DATATYPE)
**                           LABEL (KW_LABEL)
**                           DBNAME (KW_DBNAME)
**                           TABLE (KW_TABLE)
**                           WIDTH-CHARS (KW_WIDTH_C)
**                           HEIGHT-CHARS (KW_HEIGHT_C)
**                           WIDTH-PIXELS (KW_WIDTH_P)
**                           HEIGHT-PIXELS (KW_HEIGHT_P)
** 020 CA  20130221          Reordered unwrap APIs. getType was renamed to getResourceType. Added
**                           unwrap APIs for NAMESPACE-URI, UNIQUE-ID, ADM-DATA, CommonField and 
**                           Sizeable.
** 021 OM  20130220          Added dereference static methods for :: punctuation conversion.
** 022 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 023 OM  20130320          Added ResultListHandler to the list of unwrapped interfaces.
** 024 CS  20130321          Added ConnectableServer to the list of unwrapped interfaces.
** 025 GES 20130322          Fixed the BDT c'tor and modified the assign(BDT,boolean) to be the
**                           common location for the _POLY morphing. This is safe because all
**                           current usage would have generated a ClassCastException if that
**                           method was ever used with an invalid type. Now, instead of the CCE,
**                           the code will morph the value as is possible. The morphing code
**                           still needs to be implemented.
** 026 CA  20130529          Remove procedure-related dependencies, exposed the wrapped resource
**                           via a getter.
** 027 OM  20130618          Use default constructor for unknown values of date. Cleanup.
** 028 CA  20130813          Added support for the async request handle.
** 029 CA  20130908          When obtaining the resource ID via resourceId(), make sure the 
**                           resource can be "restored" using the ID.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.classloader.MultiClassLoader;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.xml.*;

/**
 * A class that represents a Progress 4GL compatible handle data type. This
 * is a container for an object of type {@link WrappedResource}.  Such
 * objects can be passed around as a regular {@link BaseDataType} wrapper
 * but the backing object can be indirectly accessed via the {@link #get} 
 * method.
 * <p>
 * The following is the mapping of Progress language features to the
 * corresponding feature in this class:
 * <p>
 * <pre>
 * string        function (default format)      {@link #toString()}
 *                        (user-defined format) {@link #toString(String)} 
 *                        (export format)       {@link #toStringExport()}
 * valid-handle  function                       {@link #isValid}
 * widget-handle function                       {@link #fromString}
 * </pre>
 * <p>
 * There are 2 deviations between this class and its behavior in the Progress
 * 4GL:
 * <p>
 * <ol>
 *    <li> The default format string is larger in Java than in Progress since
 *         the Java implementation uses all possible 32-bit values as its
 *         handle identifiers. This could be made compatible by limiting the
 *         possible values used to those that would fit into the Progress
 *         equivalent. This would probably require that every handle be
 *         registered or assigned a unique value at construction. 
 *    <li> In Progress, it is possible that one could create a loop to 
 *         instantiate all possible handles using the WIDGET-HANDLE function.
 *         This means that one might not have used the STRING function to get
 *         the text representation since that loop could create strings for
 *         all integers in the possible range of handle identifiers. This
 *         seems like a bad practice, but if it is ever used in reality, it
 *         would invalidate the implementation of {@link #fromString} which
 *         depends upon a map maintained during {@link #toString}.  The
 *         solution is the same as for the oversized text representation
 *         problem above (a registry created at construction of every 
 *         handle).
 * </ol>
 */
public class handle
extends BaseDataType
{
   /** The default format string (WARNING: not compatible with Progress). */
   private static final String defaultFormat = ">>>>>>>>>9";   
   
   /** Stores context-local state variables. */
   private static final ContextLocal<WorkArea> work =
   new ContextLocal<WorkArea>()
   {
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }
   };
   
   /**
    * Provides a map of converted API names to legacy 4GL method/attribute 
    * name.  Needed to provide support for invalid attribute/method access
    * (i.e. handle is of a certain type, but the attribute/method belongs to
    * another type). Gets populated on first class access 
    * (see {@link #populateApis()}. Not set final to be able to make it 
    * unmodifiable.
    */
   private static Map<String, String> apisToAttrOrMthd = new HashMap<String, String>();

   /** 
    * Dynamic proxy returned by unwrap APIs when the handle is invalid or not
    * of the expected type.  Gets populated on class load.
    */
   private static HandleCommon invalidAttrAccessProxy = null;

   /** Stores the referenced data of this handle. */
   private WrappedResource value = null;
   
   static
   {
      populateApis();

      ClassLoader cl = MultiClassLoader.class.getClassLoader();
      Class<?>[] ifaces = new Class<?>[] { HandleCommon.class };
      InvalidAttributeAccess handler = new InvalidAttributeAccess();
      invalidAttrAccessProxy = (HandleCommon) Proxy.newProxyInstance(cl, ifaces, handler);
   }
   
   /**
    * Default constructor, creates an instance that represents the unknown
    * value.
    */
   public handle()
   {
   }

   /**
    * This is a special c'tor which should be used only when converting the
    * value returned by a function or method with polymorphic return type into the
    * expected type (i.e. DYNAMIC-FUNCTION()).  In such cases, the 4GL does some
    * automatic type conversion (see {@link #assign(BaseDataType,boolean)}).
    * 
    * @param    value
    *           The value to be used for this instance.
    */
   public handle(BaseDataType value)
   {
      if (value == null || value.isUnknown())
      {
         setUnknown();
      }
      else
      {
         assign(value, false);
      }
   }

   /**
    * Constructs an instance that has the contained object reference and
    * <code>unknown value</code> that exactly matches that of the passed-in 
    * instance.
    *
    * @param    value
    *           The instance to pattern from.
    */
   public handle(handle value)
   {
      assign(value, false);
   }
   
   /**
    * Constructs an instance that has the contained object reference.
    *
    * @param    value
    *           The instance to pattern from.
    */
   public handle(WrappedResource value)
   {
      assign(value, false);
   }

   /**
    * Emitted and to be used only in cases the attribute is read-only. It
    * displays appropriate error message.
    * 
    * @param    h
    *           The handle used to access the attribute.
    * @param    attribute
    *           The read-only attribute.
    */
   public static void readOnlyError(handle h, String attribute)
   {
      if (attribute == null)
      {
         throw new IllegalArgumentException("The attribute is null !");
      }

      final String msg = "**%s is not a settable attribute for %s widget";
      String err = String.format(msg,
                                 attribute.toUpperCase(),
                                 unwrap(h).getResourceType().toStringMessage());
      ErrorManager.displayError(4052, err, false);
   }

   /**
    * Convert the string representation of a handle into the original handle
    * instance. This facility can only work for handles that have been
    * converted at least once into a string via {@link #toString}. Such
    * handles are stored in a weak hash map for conversion purposes.
    *
    * @param    txt
    *           A valid text representation of a handle.
    */
   public static handle fromString(character txt)
   {
      return fromString(txt.toStringMessage());
   }
   
   /**
    * Convert the string representation of a handle into the original handle
    * instance. This facility can only work for handles that have been
    * converted at least once into a string via {@link #toString}. Such
    * handles are stored in a weak hash map for conversion purposes.
    *
    * @param    txt
    *           A valid text representation of a handle.
    */
   public static handle fromString(String txt)
   {
      handle result = null;
      
      try
      {
         Integer key = new Integer(txt);
         
         WorkArea wa = work.get();
         result = (handle) wa.map.get(key);
      }
      
      catch (NumberFormatException nfe)
      {
         // return an invalid handle
         return new handle();
      }
      
      return result;
   }
   
   /**
    * This API should be called only in case of a attribute/method access using a resource which
    * does not support it.  It shows appropriate error messages, depending on handle state and
    * attribute/method.
    * <p>
    * This must not be called in cases of a not-supported attribute invoked as a lvalue (i.e. a
    * setter for a not-supported attribute is invoked). In this case, call {@link #readOnlyError}.
    * 
    * @param    attr
    *           The attribute/method name which is not supported by the handle.
    * @param    h
    *           The handle used to access the attribute/method.
    */
   public static void invalidAttribute(String attr, handle h)
   {
      if (!h._isValid())
      {
         final String errMsg1 = 
            "Invalid handle.  Not initialized or points to a deleted " +
            "object";
         final String errMsg2 = 
           "Cannot access the %s attribute because the widget does not " +
           "exist";
         int[] errIds = { 3135, 3140 };
         String[] msgs = 
         {
            errMsg1,
            String.format(errMsg2, attr)
         };
         
         ErrorManager.recordOrThrowError(errIds, msgs, false);
      }
      else
      {
         // valid handle, wrong API, show error
         final String errMsg = 
            "**%s is not a queryable attribute for %s widget";

         String type = unwrapType(h).getResourceType().toStringMessage();
         String msg = String.format(errMsg, attr, type);
         ErrorManager.displayError(4052, msg, false);
      }
   }
   
   /**
    * Unwrap the given handle to a {@link CommonHandleChain} instance. If the
    * handle is not a procedure, widget, window or frame handle or is invalid, 
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static CommonHandleChain unwrap(handle h)
   {
      return unwrapImpl(h, CommonHandleChain.class);
   }

   /**
    * Unwrap the given handle to a {@link ADMData} instance. If the handle is 
    * not a ADMData handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static ADMData unwrapADMData(handle h)
   {
      return unwrapImpl(h, ADMData.class);
   }

   /**
    * Unwrap the given handle to a {@link AsyncRequest} instance. If the handle is 
    * not a <code>AsynchronousRequest</code> handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static AsyncRequest unwrapAsync(handle h)
   {
      return unwrapImpl(h, AsyncRequest.class);
   }

   /**
    * Unwrap the given handle to a {@link AsyncRequestCountAttribute} instance. If the handle is 
    * not a AsyncRequestCountAttribute handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static AsyncRequestCountAttribute unwrapAsyncRequestCount(handle h)
   {
      return unwrapImpl(h, AsyncRequestCountAttribute.class);
   }

   /**
    * Unwrap the given handle to a {@link Buffer} instance. If the
    * handle is not the Buffer handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @param    h
    *           The handle to be unwrapped.
    *
    * @return   See above.
    */
   public static Buffer unwrapBuffer(handle h)
   {
      return unwrapImpl(h, Buffer.class);
   }

   /**
    * Unwrap the given handle to a {@link FieldInterface} instance. If the
    * handle is not the FieldInterface handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @param    h
    *           The handle to be unwrapped.
    *
    * @return   See above.
    */
   public static FieldInterface unwrapBufferField(handle h)
   {
      return unwrapImpl(h, FieldInterface.class);
   }

   /**
    * Unwrap the given handle to a {@link CommonProcedure} instance. If the
    * handle is not a procedure or the SESSION system handle or is invalid, 
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static CommonProcedure unwrapCommonProcedure(handle h)
   {
      return unwrapImpl(h, CommonProcedure.class);
   }

   /**
    * Unwrap the given handle to a {@link CommonField} instance. If the
    * handle is not CommonField resource or is invalid, 
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static CommonField unwrapCommonField(handle h)
   {
      return unwrapImpl(h, CommonField.class);
   }

   /**
    * Unwrap the given handle to a {@link Connectable} instance. If the
    * handle is not a Connectable handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static Connectable unwrapConnectable(handle h)
   {
      return unwrapImpl(h, Connectable.class);
   }
   
   /**
    * Unwrap the given handle to a {@link ConnectableServer} instance. If the
    * handle is not a ConnectableServer handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static ConnectableServer unwrapConnectableServer(handle h)
   {
      return unwrapImpl(h, ConnectableServer.class);
   }

   /**
    * Unwrap the given handle to a {@link DatabaseInfo} instance. If the
    * handle is not a procedure, widget, window or frame handle or is invalid,
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @param    h
    *           The handle to be unwrapped.
    *
    * @return   See above.
    */
   public static DatabaseInfo unwrapDatabaseInfo(handle h)
   {
      return unwrapImpl(h, DatabaseInfo.class);
   }

   /**
    * Unwrap the given handle to a {@link Dereferenceable} instance. If the
    * handle is not the Dereferenceable handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation to dereference().  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static Dereferenceable unwrapDereferenceable(handle h)
   {
      return unwrapImpl(h, Dereferenceable.class);
   }

   /**
    * Unwrap the given handle to a {@link Errorable} instance. If the
    * handle is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @param    h
    *           The handle to be unwrapped.
    *
    * @return   See above.
    */
   public static Errorable unwrapError(handle h)
   {
      return unwrapImpl(h, Errorable.class);
   }

   /**
    * Unwrap the given handle to a {@link CommonErrorStatus} instance. If the
    * handle is not the ERROR-STATUS handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static CommonErrorStatus unwrapErrorStatus(handle h)
   {
      return unwrapImpl(h, CommonErrorStatus.class);
   }
   
   /**
    * Unwrap the given handle to a {@link CommonFile} instance. If the
    * handle is not the FILE-INFO or a procedure handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static CommonFile unwrapFile(handle h)
   {
      return unwrapImpl(h, CommonFile.class);
   }

   /**
    * Unwrap the given handle to a {@link CommonFileInfo} instance. If the
    * handle is not the FILE-INFO handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static CommonFileInfo unwrapFileInfo(handle h)
   {
      return unwrapImpl(h, CommonFileInfo.class);
   }
   
   /**
    * Unwrap the given handle to a {@link FirstLastProcedureAttribute} instance. If the handle is
    * not a FirstLastProcedureAttribute handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static FirstLastProcedureAttribute unwrapFirstLastProcedure(handle h)
   {
      return unwrapImpl(h, FirstLastProcedureAttribute.class);
   }
   
   /**
    * Unwrap the given handle to a {@link CommonLastEvent} instance. If the
    * handle is not the KEY-READER handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static CommonLastEvent unwrapLastEvent(handle h)
   {
      return unwrapImpl(h, CommonLastEvent.class);
   }

   /**
    * Unwrap the given handle to a {@link LocalName} instance. If the handle is 
    * not a LocalName handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static LocalName unwrapLocalName(handle h)
   {
      return unwrapImpl(h, LocalName.class);
   }

   /**
    * Unwrap the given handle to a {@link NamespaceURI} instance. If the handle is 
    * not a NamespaceURI handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static NamespaceURI unwrapNamespaceURI(handle h)
   {
      return unwrapImpl(h, NamespaceURI.class);
   }

   /**
    * Unwrap the given handle to a {@link PersistentProcedure} instance. If 
    * the handle is not a procedure handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static PersistentProcedure unwrapProcedure(handle h)
   {
      return unwrapImpl(h, PersistentProcedure.class);
   }

   /**
    * Unwrap the given handle to a {@link P2JQuery} instance. If the
    * handle is not a query handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static P2JQuery unwrapQuery(handle h)
   {
      return unwrapImpl(h, P2JQuery.class);
   }
   
   /**
    * Unwrap the given handle to a {@link Remotable} instance. If the
    * handle is not a <code>Remotable</code> handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static Remotable unwrapRemotable(handle h)
   {
      return unwrapImpl(h, Remotable.class);
   }
   
   /**
    * Unwrap the given handle to a {@link ResultListHandler} instance. If the
    * handle is not a <code>ResultListHandler</code> handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static ResultListHandler unwrapResultListHandler(handle h)
   {
      return unwrapImpl(h, ResultListHandler.class);
   }

   /**
    * Unwrap the given handle to a {@link SaxAttributes} instance. If the handle is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is
    * needed because for invalid handles, a stack of handles is maintained, to know the state of
    * the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static SaxAttributes unwrapSaxAttributes(handle h)
   {
      return unwrapImpl(h, SaxAttributes.class);
   }

   /**
    * Unwrap the given handle to a {@link SaxEntity} instance. If the handle is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is
    * needed because for invalid handles, a stack of handles is maintained, to know the state of
    * the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static SaxEntity unwrapSaxEntity(handle h)
   {
      return unwrapImpl(h, SaxEntity.class);
   }

   /**
    * Unwrap the given handle to a {@link SaxReader} instance. If the handle is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is
    * needed because for invalid handles, a stack of handles is maintained, to know the state of
    * the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static SaxReader unwrapSaxReader(handle h)
   {
      return unwrapImpl(h, SaxReader.class);
   }

   /**
    * Unwrap the given handle to a {@link SaxWriter} instance. If the handle is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is
    * needed because for invalid handles, a stack of handles is maintained, to know the state of
    * the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static SaxWriter unwrapSaxWriter(handle h)
   {
      return unwrapImpl(h, SaxWriter.class);
   }

   /**
    * Unwrap the given handle to a {@link Sensitive} instance. If the
    * handle is not a Sensitive handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static Sensitive unwrapSensitive(handle h)
   {
      return unwrapImpl(h, Sensitive.class);
   }

   /**
    * Unwrap the given handle to a {@link Server} instance. If the handle is not a Server handle 
    * or is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static Server unwrapServer(handle h)
   {
      return unwrapImpl(h, Server.class);
   }

   /**
    * Unwrap the given handle to a {@link CommonProcedure} instance. If the
    * handle is not the SESSION system handle or is invalid, 
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static CommonSession unwrapSession(handle h)
   {
      return unwrapImpl(h, CommonSession.class);
   }

   /**
    * Unwrap the given handle to a {@link Sizeable} instance. If the
    * handle is not a procedure, widget, window or frame handle or is invalid,
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @param    h
    *           The handle to be unwrapped.
    *
    * @return   See above.
    */
   public static Sizeable unwrapSizeable(handle h)
   {
      return unwrapImpl(h, Sizeable.class);
   }

   /**
    * Unwrap the given handle to a {@link SOAPEntity} instance. If the handle's resource is 
    * not implementing the {@link SOAPEntity} interface or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static SOAPEntity unwrapSOAPEntity(handle h)
   {
      return unwrapImpl(h, SOAPEntity.class);
   }

   /**
    * Unwrap the given handle to a {@link SOAPFault} instance. If the handle's resource is 
    * not implementing the {@link SOAPFault} interface or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static SOAPFault unwrapSOAPFault(handle h)
   {
      return unwrapImpl(h, SOAPFault.class);
   }

   /**
    * Unwrap the given handle to a {@link SOAPHeader} instance. If the handle's resource is 
    * not implementing the {@link SOAPHeader} interface or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static SOAPHeader unwrapSOAPHeader(handle h)
   {
      return unwrapImpl(h, SOAPHeader.class);
   }

   /**
    * Unwrap the given handle to a {@link SOAPHeaderEntry} instance. If the handle's resource is 
    * not implementing the {@link SOAPHeaderEntry} interface or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static SOAPHeaderEntry unwrapSOAPHeaderEntry(handle h)
   {
      return unwrapImpl(h, SOAPHeaderEntry.class);
   }

   /**
    * Unwrap the given handle to a {@link Socket} instance. If the
    * handle is not the SOCKET handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static Socket unwrapSocket(handle h)
   {
      return unwrapImpl(h, Socket.class);
   }

   /**
    * Unwrap the given handle to a {@link SocketListener} instance. 
    * If the handle is not the SERVER-SOCKET handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static SocketListener unwrapSocketListener(handle h)
   {
      return unwrapImpl(h, SocketListener.class);
   }

   /**
    * Unwrap the given handle to a {@link SslServerAttribute} instance. If the handle is not a 
    * SslServerAttribute handle or is invalid, an {@link #invalidAttrAccessProxy dynamic proxy}
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static SslServerAttribute unwrapSslServerAttribute(handle h)
   {
      return unwrapImpl(h, SslServerAttribute.class);
   }

   /**
    * Unwrap the given handle to a {@link SubTypeAttribute} instance. If the handle is not a 
    * SubTypeAttribute handle or is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} 
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static SubTypeAttribute unwrapSubType(handle h)
   {
      return unwrapImpl(h, SubTypeAttribute.class);
   }

   /**
    * Unwrap the given handle to a {@link TempTable} instance. If the
    * handle is not a temporary handle or is invalid, 
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static TempTable unwrapTempTable(handle h)
   {
      return unwrapImpl(h, TempTable.class);
   }

   /**
    * Unwrap the given handle to a {@link CommonHandle} instance. If the
    * handle is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} 
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static CommonHandle unwrapType(handle h)
   {
      return unwrapImpl(h, CommonHandle.class);
   }

   /**
    * Unwrap the given handle to a {@link UniqueID} instance. If the handle is 
    * not a UniqueID handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static UniqueID unwrapUniqueID(handle h)
   {
      return unwrapImpl(h, UniqueID.class);
   }

   /**
    * Unwrap the given handle to a {@link XCommon} instance. If the handle is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is
    * needed because for invalid handles, a stack of handles is maintained, to know the state of
    * the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static XCommon unwrapXCommon(handle h)
   {
      return unwrapImpl(h, XCommon.class);
   }

   /**
    * Unwrap the given handle to a {@link XEntity} instance. If the
    * handle is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} 
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static XEntity unwrapXEntity(handle h)
   {
      return unwrapImpl(h, XEntity.class);
   }

   /**
    * Unwrap the given handle to a {@link XDocument} instance. If the
    * handle is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} 
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static XDocument unwrapXDocument(handle h)
   {
      return unwrapImpl(h, XDocument.class);
   }

   /**
    * Unwrap the given handle to a {@link XNodeRef} instance. If the
    * handle is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} 
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static XNodeRef unwrapXNodeRef(handle h)
   {
      return unwrapImpl(h, XNodeRef.class);
   }

   /**
    * Unwrap the given handle to a {@link CommonWidget} instance. If the
    * handle is not a widget handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static CommonWidget unwrapWidget(handle h)
   {
      return unwrapImpl(h, CommonWidget.class);
   }

   /**
    * Unwrap the given handle to a {@link CommonWindow} instance. If the
    * handle is a window handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    *           
    * @return   See above.
    */
   public static CommonWindow unwrapWindow(handle h)
   {
      return unwrapImpl(h, CommonWindow.class);
   }
   
   /**
    * Remove the given resource from the context-local map of all resource.
    * <p>
    * This will be called after the resource is deleted.
    * 
    * @param    h
    *           The handle containing the resource to be removed.
    */
   static void removeResource(handle h)
   {
      work.get().map.remove(resourceId(h.value));
   }

   /**
    * Get the integer ID of the given resource, as used by the {@link #toString} and 
    * {@link #fromString} APIs.
    * 
    * @param    resource
    *           A resource instance.
    * 
    * @return   The integer ID of this resource.
    */
   static int resourceId(WrappedResource resource)
   {
      // this needs to register it with the map, to allow it to be restored from the map later
      int code = resource.hashCode();
      
      // this handle's code value is being requested, save off the reference
      // in a weak hash map so that later, it can be converted back via the
      // widgetHandle() method
      WorkArea wa = work.get();
      wa.map.put(new Integer(code), new handle(resource));

      return code; 
   }

   /**
    * Unwrap the given handle so that it returns an instance compatible with
    * the given interface, <code>ifc</code>. If the handle is invalid or is
    * not compatible with the given interface, a dynamic proxy is returned
    * (see {@link #invalidAttrAccessProxy}.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    * @param    ifc
    *           The expected interface.
    *            
    * @return   See above.
    */
   private static <T> T unwrapImpl(handle h, Class<T> ifc)
   {
      WrappedResource value = h.value;
      if (h._isValid() && ifc.isAssignableFrom(value.getClass()))
      {
         return (T) value;
      }

      work.get().invalidCalls.push(h);
      return (T) invalidAttrAccessProxy;
   }

   /**
    * Populate the {@link #apisToAttrOrMthd} map, using as key the converted
    * API name and as value the legacy attribute or method name, in uppercase.
    */
   private static void populateApis()
   {
      // all API names must be in uppercase
      apisToAttrOrMthd.put("fileInfoGetType", "FILE-TYPE");
      // TODO: finish this map

      apisToAttrOrMthd = Collections.unmodifiableMap(apisToAttrOrMthd);
   }
   
   /**
    * Get the legacy attribute or method name mapped by this method.
    * 
    * @param    method
    *           The method for which the legacy name is needed.
    * @param    args
    *           The method's arguments.
    * 
    * @return   The legacy name from the {@link #apisToAttrOrMthd} map or the
    *           converted API name, if not found in the map.
    */
   private static String getAttributeName(Method method, Object[] args)
   {
      String name = method.getName();
      if ("readOnlyError".equals(name) &&
          args.length == 1             &&
          args[0] != null              &&
          (args[0] instanceof String)  &&
          method.getDeclaringClass().equals(CommonHandle.class))
      {
         return ((String) args[0]).toUpperCase();
      }
      
      String api = apisToAttrOrMthd.get(name);
   
      return api == null ? name : api;
   }

   /**
    * Creates a new instance of the same type that represents the 
    * <code>unknown value</code>.
    *
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public BaseDataType instantiateUnknown()   
   {
      return new handle();
   }
   
   /**
    * Does the same as standard <code>clone()</code> method but returns an
    * instance of <code>BaseDataType</code> and doesn't throw the
    * <code>CloneNotSupportedException</code>.
    *
    * @return   A clone of this instance.  
    */
   public BaseDataType duplicate()
   {
      return new handle(this);
   }
   
   /**
    * Hash code implementation which is consistent with {@link
    * BaseDataType#equals}.
    *
    * @return  Hash code value for this object instance.
    */
   public int hashCode()
   {  
      return value == null ? ((Object) this).hashCode() : resourceId(value);
   }
   
   /**
    * Returns the value of this instance as an <code>Object</code>.
    * <p>
    * If this instance represents the <code>unknown value</code>, a
    * <code>null</code> will be returned.
    *
    * @return   The value as an <code>Object</code> or <code>null</code> if
    *           this is <code>unknown</code>.
    */
   public Object get()
   {
      Object result = value;
      
      // redirect to contained object if it is the special wrapper
      if (result != null && result instanceof TransparentWrapper)
      {
         TransparentWrapper wrap = (TransparentWrapper) result;
         result = wrap.get();
      }
      
      return result;
   }   
   
   /**
    * Set the value of this instance as a {@link WrappedResource}.
    *
    * @param    value
    *           The value to be represented in this instance.
    */
   public void set(WrappedResource value)
   {
      this.value = value;
   }
   
   /**
    * Determines if this instance is valid (to be used).
    *
    * @return   <code>true</code> if this instance can be used.
    */
   public logical isValid()
   {
      return new logical(_isValid());
   }   
   
   /**
    * Determines if this instance is valid (to be used).
    *
    * @return   boolean <code>true</code> if this instance can be used.
    */
   public boolean _isValid()
   {
      return (value == null) ? false : value.valid();
   }   
   
   /**
    * Determines if this instance represents the <code>unknown value</code>.
    * 
    * @return   <code>true</code> if this instance is set to the
    *           <code>unknown value</code>.
    */
   public boolean isUnknown()
   {
      return (value == null || value.unknown());
   }
   
   /**
    * Sets the state of this instance's <code>unknown value</code> flag to
    * <code>true</code>.
    * <p>
    * <b>Warning: a separate call is needed to ensure that the data of this
    * instance is set to the correct value.</b>
    */
   public void setUnknown()
   {
      value = null;
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    * <p>
    * If the value is not of type <code>handle</code>, the following automatic type
    * conversion will occur:
    * <p>
    * <pre>
    * TODO: implement and document this
    * </pre>
    * <p>
    * This conversion is meant to handle the cases of built-in functions and methods
    * in the 4GL which have polymorphic return types (e.g. DYNAMIC-FUNCTION()).
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(BaseDataType value, boolean force)
   {
      if (value instanceof handle)
      {
         assign((handle) value, force);
      }
      else
      {
         // TODO: implement type conversion here for _POLY cases 
      }
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(handle value, boolean force)
   {
      assign(value.value, force);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(WrappedResource value, boolean force)
   {
      if (force || !ErrorManager.isPendingError())
      {
         if (value instanceof CommonFrame)
         {
            value = ((CommonFrame) value).asWidget();
         }

         this.value = value;
      }
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    value
    *           The instance from which to copy state. This must be of
    *           type <code>handle</code> or <code>WrappedResource</code>.
    */
   public void assign(Object value)
   {
      if (value instanceof WrappedResource)
      {
         assign((WrappedResource) value);
      }
      else if (value instanceof handle)
      {
         assign((handle) value);
      }
      else
      {
         // unexpected failure
         throw new IllegalArgumentException("Invalid value type.");
      }
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(handle value)
   {
      assign(value.value, false);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(Undoable value)
   {
      assign((handle) value, false);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * given value.  Note that <code>null</code> may be passed in which case
    * this instance will be set to the <code>unknown value</code>.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(WrappedResource value)
   {
      assign(value, false);
   }
   
   /**
    * Compares this instance with the specified instance and returns a -1
    * if this instance is less than the specified, 0 if the two instances
    * are equal and 1 if this instance is greater than the specified
    * instance.  This is the implementation of the <code>Comparable</code>
    * interface.
    * <p>
    * The algorithm will fail to give meaningful results in the case where
    * one tries to sort against other objects that do not represent compatible
    * values.
    *
    * @param    obj
    *           The instance to compare against.
    *
    * @return   0 or 1 depending on if the passed instance is or is not
    *           <code>unknown</code>.
    */
   public int compareTo(Object obj)   
   {  
      handle h = null;
      
      if (obj instanceof handle)
      {
         h = (handle) obj;
      }
      else
      {
         return -1;
      }
      
      int result = 0;
      
      int h1 = hashCode();
      int h2 = h.hashCode();
      
      if (h1 < h2)
      {
         result = -1;
      }
      else if (h1 == h2)
      {
         // arbitrary but if the hashcodes are equal and the instances are
         // not actually equal, then make sure this is clear
         result = 1;

         if (value.equals(h.value))
         {
            result = 0;
         }
      }
      else
      {
         result = 1;
      }
         
      return result;
   }
   
   /**
    * Creates a string representation of the instance data using the default
    * Progress 4GL format of '>>>>>>>>>9'.  If the instance represents the
    * <code>unknown value</code>, a '?' will be returned.
    *
    * @return   The formatted string.
    */
   public String toString()   
   {
      return toString(null);
   }
   
   /**
    * Creates a string representation of the instance data using the user
    * specified format string (Progress 4GL compatible) or if no such
    * string is provided, the default format of '>>>>>>>>>9'.  If the
    * instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @param    fmt
    *           The Progress 4GL format string or <code>null</code> if the
    *           default format is to be used.
    *
    * @return   The formatted string.
    */
   public String toString(String fmt)   
   {
      if (isUnknown())
      {
         return "?";
      }
      
      int hc = hashCode();
      
      // this handle's text value is being requested, save off the reference
      // in a weak hash map so that later, it can be converted back via the
      // widgetHandle() method
      WorkArea wa = work.get();
      wa.map.put(new Integer(hc), this);
      
      integer i = new integer(hc);
      return i.toString((fmt == null) ? defaultFormat : fmt);
   }
   
   /**
    * Creates a string representation of the instance data in a form that is
    * compatible with the <code>MESSAGE</code> language statement.  If the
    * instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @return   The 'message' formatted string.
    */
   public String toStringMessage()   
   {
      return toString();
   }
   
   /**
    * Creates a string representation of the instance data using the 'export'
    * format.  If the instance represents the <code>unknown value</code>, a
    * '?' will be returned.
    *
    * @return   The 'export' formatted string.
    */
   public String toStringExport()   
   {
      return toString();
   }   
   
   /**
    * Return the default display format string for this type.
    * <p>
    * Warning: this is larger than the Progress version because the range of
    * handle values can be a 32-bit integer in this implementation.
    *
    * @return   The default format string.
    */
   public String defaultFormatString()   
   {
      return defaultFormat;
   }      
      
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      throw new NotSerializableException("handle");
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      throw new NotSerializableException("handle");
   }
   
   /**
    * Getter for the wrapped resource.
    * 
    * @return   The {@link #value resource}. 
    */
   public WrappedResource getResource()
   {
      return value;
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** 
       * Stores handles that have been converted to strings and may need
       * to be converted back into a handle.
       * 
       * TODO: this map gets cleaned only on client exit, memory problems may 
       * appear if the client is long-running. Should it be maintained by the
       * DELETE OBJECT/PROCEDURE statements too ?
       */
      private Map<Integer, handle> map = new HashMap<Integer, handle>();
     
      private ArrayDeque<handle> invalidCalls = new ArrayDeque<handle>();
   }
   
   /**
    * Handler for the {@link #invalidAttrAccessProxy} proxy.
    */
   private static class InvalidAttributeAccess
   implements InvocationHandler
   {
      
      /**
       * This proxy is used only for cases when the handle is invalid or not of the expected type.
       */
      @Override
      public Object invoke(Object proxy, Method method, Object[] args)
      throws Throwable
      {
         WorkArea wa = work.get();
         
         BaseDataType unkn = null;
         Class<?> ret = method.getReturnType();
         
         if (ret != null && BaseDataType.class.isAssignableFrom(ret))
         {
            unkn = (BaseDataType) ret.newInstance();
         }

         handle target = wa.invalidCalls.pop();
         String attr = getAttributeName(method, args);

         if (!target._isValid())
         {
            final String errMsg1 = 
               "Invalid handle.  Not initialized or points to a deleted object";
            final String errMsg2 = 
              "Cannot access the %s attribute because the widget does not exist";
            int[] errIds = { 3135, 3140 };
            String[] msgs = 
            {
               errMsg1,
               String.format(errMsg2, attr)
            };
            
            ErrorManager.recordOrThrowError(errIds, msgs, false);

            return unkn;
         }

         // valid handle, wrong API, show error
         final String errMsg = 
            "**%s is not a queryable attribute for %s widget";
         String type = unwrapType(target).getResourceType().toStringMessage();
         String msg = String.format(errMsg, attr, type);
         ErrorManager.displayError(4052, msg, false);

         return unkn;
      }
   }
}
